#!/usr/bin/perl

use vars qw( %bin $dumpDir $ptrArgs $pwd $workDir);
use strict;

#
# confixx global parameters
#
use vars qw($user_homeDir $installDir $stdShell $version
	    $dbType $dbServer $dbUser $dbPw $dbDB $db_address $dbPort
	    $mysqlUserServer $mysqlUserUser $mysqlUserPw $mysqlUserPort
	    $mysql_host $mysql_sock $stdDomain $majordomo_ldir
	    $maildrop $pop_homeDir $mailBoxName $mailSpool $backup_off);

#
# end confixx global parameters
#

my $VERSION='1.3.3';

#
# parse command line
#
my %arg_opts = ('--help|-h'=>'',
				'--config|-c'=>'s',
				'--debug|-d'=>'i?',
				'--get-status|-s'=>'',
				'--dump-accounts|-dc'=>'s',
				'--dump-domains|-dd'=>'s',
				'--dump-all|-da'=>'',
				'--get-content-list|-lc'=>'',
				'--no-content|-nc'=>'',
				'--no-compress|-nz'=>'',
				'--output|-o'=>'s',
				'--status-file|-sf'=>'s',
				'--version|-v'=>'',
	       );

%bin = ( 'tar'=>'tar', 'split'=>'split', 'cp'=>'cp', 'rm'=>'rm',
				 'echo'=>'echo',
				 'mysql'=>'mysql', 'mysqldump'=>'mysqldump', 'psql'=>'psql',
				 'uptime'=>'uptime', 'uname'=>'uname', 'cat'=>'cat', 
				 'pwd'=>'pwd',
				 'grep'=>'grep', 'find'=>'find', 'gzip'=>'gzip', 'wc'=>'wc',
				 'md5sum'=>'md5sum', 'locale'=>'locale',  'iconv'=>'iconv',
				 'crontab'=>'crontab' );

my %mapResLimits = ( 'max_subdom' => 'maxsubdomains',
										 'max_dom' => 'maxdomains',
										 'disk_space' => 'maxkb',
										 'max_traffic' => 'maxtransfer',
										 'max_db' => 'maxmysql',
										 'max_box' => 'maxpop',
										 'max_resp' => 'maxautoresponder',
										 'max_maillists' => 'maxmaillist',
									 );

my %mapResPermis = ('create_domains' => 'maxdomains',
										'manage_phosting' => 'maxkunden',
										'manage_subdomains' => 'maxsubdomains',
										'change_limits' => 'maxkunden',
										'manage_dns' => 'dns',
										'manage_log' => 'maxkunde',
										'manage_sh_access' => 'shell',
										'make_dumps' => 'maxkunden',
										'manage_maillists' => 'maxmaillist',
										'manage_crontab' => 'maxcronjobs',
									 );

my %mapResPinfo = ( 'company' => 'firma',
										'phone' => 'telefon',
										'fax' => 'fax',
										'address' => 'anschrift',
										'city' => 'plzort',
										'state' => '',
										'zip' => 'plz',
										'country' => 'land',
										'locale' => 'language',
										'email' => 'emailadresse' );

my %mapUserPinfo = ( 'company' => 'firma',
										 'phone' => 'telefon',
										 'fax' => 'fax',
										 'address' => 'anschrift',
										 'city' => 'plzort',
										 'state' => '',
										 'zip' => 'plz',
										 'country' => 'land',
										 'locale' => 'language',
										 'email' => 'emailadresse' );


my %mapCustLimits = ( 'max_subdom' => 'maxsubdomains',
											'disk_space' => 'maxkb',
											'max_traffic' => 'maxtransfer',
											'max_db' => 'maxmysql',
											'max_box' => 'maxpop',
											'max_resp' => 'maxautoresponder',
											'max_maillists' => 'maxmaillist',
										);

my %comulatLimits = ( 'max_db' => 1 );

my %mapCustPermis = ('manage_subdomains' => 'maxsubdomains',
										 'manage_log' => 1,
										 'manage_maillists' => 'maxmaillist',
										 'manage_ftp_password' => 1,
										 'manage_crontab' => 'maxcronjobs',
										);

my %mapCustScripting = ('ssi'=>'ssi',
												'php'=>'php',
												'cgi'=>'perl',
												'perl'=>0);

my %mapLocales = ('de' => 'de-DE',
									'dk' => 'dk',
									'en' => 'en-US',
									'es' => 'es-ES',
									'fr' => 'fr-FR',
									'gr' => 'gr',
									'it' => 'it-LT',
									'ru' => 'ru-RU');

#
# global variables
#
my (%preparedCustomers);
#
# end global variables
#

my $returnCode=0;


#==============================================================
#
# common subroutines
#

sub printToLog {
  my ($msg, $noeol) = @_;
  if ($msg) {
	print $msg;
  }
  unless ($noeol) {
	print "\n";
  }
}

sub printToError {
  my ($msg, $noeol, $errCode) = @_;
  if ($msg) {
	print STDERR $msg;
  }
  unless ($noeol) {
	print STDERR "\n";
  }
  if (defined($errCode)) {
    $returnCode = $errCode; ## translate an error code to the return code
  } else {
    $returnCode = 127;
  }
}

#BEGINNOTINRAQ2

sub makeFileReader
{
	my ($filename, $doNotFail) = @_;
	local (*INPUT);

	unless (-T $filename) {

		if (defined($doNotFail)) {
			&printToLog("Error: file '$filename' is not a text file");
		} else {
			&printToError("Error: file '$filename' is not a text file");
		}

		return undef;
	}

	unless (open(INPUT,"<$filename")) {

		binmode INPUT;
		if (defined($doNotFail)) {
			&printToLog("Error: unable to open file '$filename' to read");
		} else {
			&printToError("Error: unable to open file '$filename' to read");
		}

		return undef;
	}

	my $input = *INPUT;

	my $this = {
		'READ' => sub {
				my $a = <$input>;
				return $a;
			},

		'CLOSE' => sub { close($input); }

	};

	return $this;
}

sub makeArrayReader {
  my (@arr) = @_;
  my $size= 0;
  my $count = scalar(@arr);

  my $this={'READ'=>sub {
			  if($count == $size) {
				return undef;
			  }

			  return $arr[$count++];
			},
			'CLOSE'=>sub {
			}
		   };
  return $this;
}

sub makeParser {
  my $reader;
  my ($this,$ptrRows,$keySep,$valSep,$valKeySep,$comment,
      $typeRows,$maxIndex,$rowIndex);
  {
	$reader = shift;
	unless(defined($reader)) {
	  return undef;
	}

    my %args=(@_);
    $keySep = $args{'KEYSEPARATOR'};
    $valSep = $args{'VALUESEPARATOR'};
    $valKeySep = $args{'KEYVALUESEPARATOR'};
  }
  $this={'READER'=>sub {
		   if(@_){
			 $reader=shift;
		   }
		   return $reader;
		 },
		 'FILE'=>sub {
		   if(@_){
			 $reader = makeFileReader(@_);
		   }
		   return undef;
		 },
		 'KEYSEPARATOR'=>sub{
		   if(@_){
			 $keySep=shift;
		   }
		   return $keySep;
		 },
		 'VALUESEPARATOR'=>sub{
		   if(@_){
			 $valSep=shift;
		   }
		   return $valSep;
		 },
		 'KEYVALUESEPARATOR'=>sub{
		   if(@_){
			 $valKeySep=shift;
		   }
		   return $valKeySep;
		 },
		 'COMMENT'=>sub{
		   if(@_){
			 $comment=shift;
		   }
		   return $comment;
		 },
		 'PARSE'=>sub{
		   my %args=(@_);

		   $keySep=$args{'KEYSEPARATOR'} if exists $args{'KEYSEPARATOR'};
		   $valSep=$args{'VALUESEPARATOR'} if exists $args{'VALUESEPARATOR'};
		   $valKeySep=$args{'KEYVALUESEPARATOR'} if exists $args{'KEYVALUESEPARATOR'};
		   $comment=$args{'COMMENT'} if exists $args{'COMMENT'};

		   $ptrRows = &parseFile('reader'=>$reader,'keysep'=>$keySep,
								 'valsep'=>$valSep,'keyvalsep'=>$valKeySep,
								 'comment'=>$comment);
		   $typeRows = ref($ptrRows);
		   if($typeRows=~/ARRAY/){
			 $rowIndex=-1;
			 $maxIndex=scalar(@{$ptrRows})-1;
		   }
		   return $ptrRows;
		 },
		 'ROW'=>sub{
		   if($typeRows=~/HASH/){
			 my $key=shift;
			 return $ptrRows->{$key};
		   }elsif($typeRows=~/ARRAY/){
			 if($rowIndex<$maxIndex){
			   $rowIndex++;
			   return $ptrRows->[$rowIndex];
			 }else{
			   return undef;
			 }
		   }else{
			 return undef;
		   }
		 },
		};

  return $this;
}

sub makeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename), @_);
}

sub makeSafeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename, 'doNotFail'), @_);
}

sub makeStringParser {
  my $content = shift;
  return makeParser(makeArrayReader($content), @_);
}

sub parseFile {
  my %arg = (@_);
  my $reader = $arg{'reader'};
  my $keySep = $arg{'keysep'};
  my $valSep = $arg{'valsep'};
  my $valKeySep = $arg{'keyvalsep'};
  my $comment = $arg{'comment'};

  my ($ret,$key,$value,$reKey,$reValue,@values,$keyBlank,$valBlank,
     $vkey,$vvalue,$reKeyValue,$reComment);

  if($keySep){
    $ret = {};
    if(ref($keySep)=~/regexp/i || ($keySep eq ' ')){
      $reKey = $keySep;
    }else{
      $reKey = qr/$keySep/;
    }
  }else{
    $ret = [];
  }
  if($valSep){
    if(ref($valSep)=~/regexp/i || ($valSep eq ' ')){
      $reValue = $valSep;
    }else{
      $reValue = qr/$valSep/;
    }
  }
  if($valKeySep){
    if(ref($valKeySep)=~/regexp/i || ($valKeySep eq ' ')){
      $reKeyValue = $valKeySep;
    }else{
      $reKeyValue = qr/$valKeySep/;
    }
  }
  if($comment){
    if(ref($comment)=~/regexp/i){
      $reComment = $comment;
    }else{
      $reComment = qr/$comment/;
    }
  }

  while($_ = $reader->{'READ'}->()){
    chomp;
    next unless $_;
    if($comment && /$reComment/){  ## commentar
	  next;
    }
    if ($keySep){
      ($key,$value)=split($reKey,$_,2);
      if($key){
		$key=~s/^\s+//;
		$key=~s/\s+$//;
		if($valSep){

		  if ($valKeySep){
			$ret->{$key}={};
			foreach $value (split($reValue,$value)){
			  ($vkey,$vvalue)=split($reKeyValue,$value);
			  $vkey=~s/^\s+//;
			  $vkey=~s/\s+$//;
			  $ret->{$key}->{$value}=$vvalue;
			}
		  }else{
			push @{$ret->{$key}},split($reValue,$value);
		  }
		}else{
		  $value=~s/^\s+//;
		  $value=~s/\s+$//;
		  $ret->{$key}=$value;
		}
      }
    }else{
      if($valSep){
		push @{$ret},[split($valSep,$_)];
      }else{
		push @{$ret},$_;
      }
    }
  }

  $reader->{'CLOSE'}->();

  return $ret;
}

#ENDNOTINRAQ2

sub makePasswordNode {
  my $value= shift;
  my $type = shift||'encrypted';

  my $item = &makeXmlNode('password');
  $item->{'CONTENT'}->($value);  ## html encode
  $item->{'ATTRIBUTE'}->('type',$type);
  return $item;
}

sub mangleName {
  my $name = shift;

  if($name =~ /^\d/) {
    $name = 'user' . $name;
  }
  $name =~ s/\./_/g;

  return $name;
}

sub makeSysUserNode {
  my ($name, $password, $type) = @_;

  unless(defined($name) && defined($password) && defined($type)) {
	printToError("makeSysUserNode: Username, password or password type is not set.");
  }

  my $sysUserNode = makeXmlNode('sysuser');
  $sysUserNode->{'ATTRIBUTE'}->('name', mangleName($name));
  $sysUserNode->{'ADDCHILD'}->(makePasswordNode($password, $type));

  return $sysUserNode;
}

# Returns home directory for the system account $user
sub getHomeDir {
	my ($user) = @_;
	my $home;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$home = $pw[7];
	}
	unless ($home) {
		# how it is by default
		$home = '/home/$account';
	}
	return $home;
}

# Returns password for system account $user
sub getSystemPassword {
	my ($user) = @_;
	my $password;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$password = $pw[1];
	}
	unless ($password) {
		$password = '';
	}
	return $password;
}

sub createUserNameFromDomain {
  my ($name) = @_;

  if ($name =~ /^\d/) {
	$name = 'u' . $name;
  }

  $name =~ s/\./_/g;
  $name =~ s/^www\.//g;
  $name =~ s/[aoeui]//g if length($name) > 16;
  $name = substr($name, 0, 16) if length($name) > 16;

  return $name;
}

my $compress = 1;

sub setCompress {
  ($compress) = @_;
}

sub makeDumpDb
{
  my ($dbName, $dbType, $dbUser, $dbPasswd, $dbHost, $ptrDomainBin, $variables, $domainSocket) = @_;
  $dbType ||= 'mysql';

  my ($dumpFile,$dumpSql,$dumpName,$cmd,$gzip,$split,$mysql,$psql, $retCode,$out);

  $dumpName = "$dbName.$dbType.sql";
  $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  if (-x $gzip) {
	$dumpName .= ".gz";
  }
  $dumpFile="$dumpDir/$dumpName";

  printToLog(">>Checking database presence");

  if ($dbType eq 'mysql') {
	$mysql = $bin{'mysql'} || 'mysql';
	if (-x $mysql) {
	  $cmd  = "$mysql -u '$dbUser'";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -p'$dbPasswd'" if $dbPasswd;
	  $cmd .= " -e '' '$dbName' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: mysql datatbase '$dbName' not found. mysql has returned '$retCode'");
		return "";
	  }
	}
  } elsif ($dbType eq 'Pg') {
	$psql = $bin{'psql'} || 'psql';
	if (-x $psql) {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if $dbPasswd;
	  $cmd .= "; $psql '$dbName' ";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -c '' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: PostgreSQL datatbase '$dbName' not found. psql has returned '$retCode'");
		return "";
	  }
	}
  }

  printToLog("..done");

  unless (exists($ptrArgs->{'no-content'})) {
	$gzip = $bin{'gzip'} || 'gzip';
	$split = $bin{'split'} || 'split';

	if ($dbType eq 'mysql') {
	  my ($prefx, $postfix) = ("", "");
	  my $commonprefix = "LD_LIBRARY_PATH=$variables " if $variables;

	  my $commonopts;
	  $commonopts = " --socket='$domainSocket'" if $domainSocket;
	  $commonopts .= " -h $dbHost" if $dbHost;
	  $commonopts .= " -p'$dbPasswd'" if $dbPasswd;
	  $commonopts .= " -u '$dbUser' '$dbName'";

	  my $dumpcmd;
	  if ($ptrDomainBin) {
		$dumpcmd = $ptrDomainBin->{'mysqldump'};
	  } else {
		$dumpcmd = $bin{'mysqldump'} || 'mysqldump';
	  }

	  my $dumpopts = " --quick --quote-names --add-drop-table";

	  my $mysqlversion = `$commonprefix $bin{'mysql'} $commonopts --version`;
	  if ($mysqlversion =~ /Distrib (\d+)\.(\d+)\./) {
		if ($1 > 4 or ($1 == 4 and $2 > 0)) {
		  $dumpopts .= " --compatible=mysql323";
		}
	  }

	  $cmd = $commonprefix . $dumpcmd . $commonopts . $dumpopts;

	} elsif ($dbType eq 'Pg') {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if($dbPasswd);
	  $cmd .= ' ; ' . ($bin{'pg_dump'} || 'pg_dump') . ' -d -O';
	  $cmd .= " -h '$dbHost'" if $dbHost;
	  $cmd .= " '$dbName'";
	}

	if ($compress and $gzip and -x $gzip) {
	  $cmd .= " | $gzip";
	}
	if (-x $split) {
	  $cmd .= " | $split -b 2040m - $dumpFile.";
	} else {
	  $cmd .= " > $dumpFile";
	}

	local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

	unless (system($cmd) == 0) {
	  printToError("Error execute '$cmd'");
	  return undef;
	}
  }
  return $dumpName;
}

sub addDbUsers {
  my ($dbNode, $dbName, $wrapUserMysql) = @_;

  printToLog("Dupming $dbName database users");

  $dbName =~ s/_/\\_/g;

  my (@users, $dbUser, $ptrRow);
  my $sql = "SELECT DISTINCT User FROM db WHERE Db = '$dbName'";
  $wrapUserMysql->{'EXECUTE'}->($sql);
  while ($ptrRow = $wrapUserMysql->{'FETCHROW'}->()) {
		push @users, (@{$ptrRow})[0];
  }
  $wrapUserMysql->{'FINISH'}->();

  foreach $dbUser (@users) {
		addDbUser($dbNode, $dbUser, $wrapUserMysql);
  }
}

sub addDbUser {
  my ($dbNode, $user, $wrapUserMysql, $passwordNode) = @_;

  &printToLog("Dumping mysql user '$user'");

	my ( $item );

  if ( ref( $wrapUserMysql ) =~ /HASH/ &&
			 ref( $wrapUserMysql->{'EXECUTE'} ) =~ /CODE/ ) {

    my ($sql,$ptrRow,@accesshosts,$accesshost);

    $sql = "SELECT password,host FROM user WHERE user='$user'";

    if ( $wrapUserMysql->{'EXECUTE'}->( $sql ) ) {

			$item = makeXmlNode( 'dbuser' );
			$item->{'ATTRIBUTE'}->('name', $user);

      while ( $ptrRow = $wrapUserMysql->{'FETCHROW'}->() ) {
				unless ( ref($passwordNode) =~ /HASH/ ) {
					$passwordNode = makePasswordNode( $ptrRow->[0] );
				}
				push @accesshosts, $ptrRow->[1];
      }

			if ( $passwordNode ) {
				$item->{'ADDCHILD'}->( $passwordNode );
			}

			foreach $accesshost ( @accesshosts ) {
				$item->{'ADDCHILD'}->( makeXmlNode( 'accesshost', $accesshost ) );
			}

			$dbNode->{'ADDCHILD'}->( $item );

    }

    $wrapUserMysql->{'FINISH'}->();
  }

	return $item;
}

my $lastDumpId = 0;

sub generateDumpId {
  return $$ . "-" . $lastDumpId++;
}

#
# sub makeDumpFile - make a dump of dir tree (use 'tar')
#
# Arguments:
#          $destFile - name of destination file.
#                      If the file has not '.tgz' or '.tar.gz' extension
#                      '.tgz' will be added
#          $scrDir   - path to source dir
#
#          $ptrNames - pointer to an array with names of files and dirs to
#                      include into dump
#                      can be a single name
#                      by default '.' - include all entries
#
#          $ptrExcludes - pointer to an array with names of files and dirs to
#                      exclude from dump
#                      can be a single name
#
#          $extraParams - extra params to run tar
#                         ('-h' - don't dump symlinks; dump the files they point to)
#
sub makeDumpFile {
  my ($destFile, $srcDir, $ptrNames,$ptrExclude,$extraParams)=@_;

  my @destPath = split(/\//, $destFile);
  if (length($destPath[$#destPath]) > 32) {
	$destPath[$#destPath] = generateDumpId();
  }
  $destFile = join('/', @destPath);

  my $tar = $bin{'tar'}||'tar';
  my $split = $bin{'split'}||'split';
  my $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  my ($cmd,$ret,@names,@exclude);

  if(ref($ptrNames)=~/ARRAY/){
    @names = @{$ptrNames};
  }else{
    @names=($ptrNames) if $ptrNames;
  }

  if(ref($ptrExclude)=~/ARRAY/){
    @exclude = @{$ptrExclude};
  }else{
    @exclude=($ptrExclude) if $ptrExclude;
  }

  unless(-x $tar ){
    &printToError("Error: '$tar' is not executable");
    return undef;
  }

#
# check extension
#
  if(-x $gzip){
    unless($destFile=~/\.tgz$/ ||$destFile=~/\.tar\.gz$/){
      $destFile .= '.tar.gz';
    }
  }else{
    unless($destFile=~/\.tar$/){
      $destFile .= '.tar';
    }
  }
#
# end check extension
#

  unless( exists $ptrArgs->{'no-content'}){

    $cmd=$tar;
    if($srcDir){
      if( -d $srcDir){
		$cmd.=" -C $srcDir -f - -c --ignore-failed-read ";
      }else{
		return undef;
      }
    }


    if($extraParams){
      $cmd .= $extraParams.' ';
    }

    if(@names){
      if($srcDir){
		$cmd .= join( ' ', map {s/^\Q$srcDir\E\///;$_} @names); ## extract domain name
      }else{
		$cmd .= join( ' ', @names );
      }
    }else{
      $cmd .= '.';  ## dump all files & dirs
    }
    if(@exclude){
      $cmd .= ' '.join( ' ', map {s/^\Q$srcDir\E\///;'--exclude '.$_} @exclude); ## extract domain name
    }
    if(-x $gzip){
      $cmd .= " | $gzip";
    }
    if(-x $split){
      $cmd .= " | $split -b 2040m - $destFile.";
    }else{
      $cmd .= " >$destFile";
    }

		local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

    unless(system($cmd) == 0){
      &printToError("Error execute command: '$cmd': $!");
      $destFile='';
    }
  }

  $destFile=~s/^\Q$dumpDir\E\///;  ### remove dir name
  return $destFile;
}

sub getDumpDir {
  my ($subdir,$clean) = @_;
  my ($dumpdir);

  if($subdir=~/^\//){
    $dumpdir=$subdir;
  }else{
    my $util = $bin{'pwd'}||'pwd';
    $dumpdir = $pwd||`$util`;
    chomp $dumpdir;
    if($subdir){
      $dumpdir.='/'.$subdir;
    }
  }
  if(-d $dumpdir && $clean){
    unless(system("rm -rf $dumpdir")==0){
      &printToError("Error: can not clean temporary dir '$dumpdir': $!");
    }
  }
  unless(-d $dumpdir){
    mkdir ($dumpdir,0755) or
      die "Error: can not create temporary dir '$dumpdir': $!\n";
  }
  return $dumpdir;
}

sub printXml {
  my ($root,$outFh,$dtd) = @_;
  unless($outFh){
    $outFh = *STDOUT;
  }

  unless(ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/){
    print $outFh "Error: a xml root is not found\n";
    &printToError("Error: sub printXml: bad arguments");
    return 0;
  }
  print $outFh "<?xml version='1.0' encoding=\"UTF-8\"?>\n";

  my  $doctype = $root->{'NAME'}->();
  if($dtd){
    print $outFh "<!DOCTYPE $doctype SYSTEM '$dtd'>\n";
  }
  $root->{'PRINT'}->($outFh,'  ');
  return 1;
}

#
# class for handle with XML
#
sub makeXmlNode {
  my ($this,$parent,%attributes,@content);
  my ($name,$plain)=@_;

  $plain = toUtf8(encodeHtml($plain));

  unless($name||$plain){
    return undef;
  }
  $this = {
	   'ATTRIBUTE' => sub {
	     my $key = shift;
	     if(@_){
	       my $value=shift;
	       if(@_){
					 $attributes{$key} = $value;
	       }else{
					 $attributes{$key} = toUtf8(normalizeAttribute($value));
	       }
	     }
	     return $attributes{$key};
	   },
	   'PARENT' => sub {
	     if(@_){
	       my $arg = shift;
	       if(ref($arg)=~/HASH/){
					 $parent=shift;
	       }
	     }
	     return $parent;
	   },
	   'NAME' => sub {
	     if(@_){
	       $name = shift;
	     }
	     return $name;
	   },
	   'CONTENT' => sub {
	     if(@_){
	       $plain = toUtf8(encodeHtml(shift));
	     }
	     return $plain;
	   },
	   'ADDCHILD'=> sub {
	     my $kind = shift;
	     if(
					(ref($kind)=~/HASH/) &&
					(ref($kind->{'PARENT'})=~/CODE/)
	       ) {
	       push @content,$kind;
	       $kind->{'PARENT'}->($this);
	     }
	     return $#content;
	   },
	   'KIND'=> sub {
	     my $index = shift;
			 my ($ret);
			 if ( $index =~ /^\d+$/ ) {
				 if($index >= 0 && $index <= $#content ) {
					 $ret =  $content[$index];
				 }
			 } else {
				 foreach my $item (@content){
					 if ( $item->{'NAME'}->() eq $index ) {
						 $ret = $item;
						 last;
					 }
				 }
			 }
			 return $ret;
	   },
	   'KINDSREF'=> sub {
			 my $name = shift;
			 my $ret = \@content;
			 if ( $name ) {
				 $ret = [];
				 foreach my $item (@content) {
					 if ( $item->{'NAME'}->() eq $name ) {
						 push @{$ret}, $item;
					 }
				 }
			 }
	     return $ret;
	   },
	   'ATTRIBUTESREF' => sub {
	     return \%attributes;
	   },
	   'PRINT'=>sub {
	     my $fh = shift || *STDOUT;
	     my $addPrefix = shift || '';
	     my $prefix = shift || '';
	     if($name){
	       my ($out,$item);
	       if(@content || defined($plain)){

					 $out = $prefix . &getTag($name,\%attributes,'start');
					 print $fh $out;
					 if(@content){
						 print $fh "\n";
						 foreach $item (@content){
							 if(ref($item)=~/HASH/ && ref($item->{'PRINT'})=~/CODE/){
								 $item->{'PRINT'}->($fh,$addPrefix,$prefix.$addPrefix);
							 }
						 }
					 }
					 if(defined($plain)){
						 print $fh $plain;
						 if(@content){
							 print $fh "\n$prefix";
						 }
					 }else{
						 if($prefix){
							 print $fh $prefix;
						 }
					 }
					 $out = &getTag($name,\%attributes,'stop') . "\n";
					 print $fh $out;

	       }else{
					 $out = $prefix . &getTag($name,\%attributes) . "\n";
					 print $fh $out;
	       }
	     }else{
	       if(defined($plain)){
					 print $fh $plain;
	       }
	     }
	   }
	 };

  return $this;
}

sub getTag {
  my ($name,$ptrAttr,$mode) = @_;
  my ($out,$key,$value);
  $mode||='';
  if ($mode eq 'stop'){
    $out="</$name>";
  }else{
    $out="<$name";
    while (($key,$value)=each(%{$ptrAttr})){
      $out .= ' '.$key;
      if (defined($value)){
		$out .= '="'.$value.'"';
      }
    }
    if($mode eq 'start'){
      $out .= ">";
    }else{
      $out .= "/>";
    }
  }
  return $out;
}

#
# end class for handle with XML
#


#
# class for handle with database
#
sub getDbConnect {
  my $ptrBin = shift;
  my $myDbType = shift;
  my $myDbAddress = shift ;
  my $myDbUser = shift ;
  my $myDbPasswd = shift ;
  my $myDbDB = shift ;
  my $myDbHost = shift ;

  #added by Honig for Sphera
  my $myDbSocket = shift;
  my $variables = shift;

  my ($dbh,$this,$sth);
  my ($shell,$args);
  my $echo = $ptrBin->{'echo'}||'echo';
  my ($sql,@result,$resultIndex,$ptrResult);

  unless(defined($myDbPasswd) && defined($myDbUser)){
    if($myDbType eq 'mysql'){
      foreach my $myCnf ('/etc/my.cnf',$ENV{'HOME'}.'/.my.cnf'){
		next unless (-T $myCnf);
		next unless open(CNF,"<$myCnf");
		binmode CNF;
		my($key,$value);
		my $state=0;
		while (<CNF>){
		  next unless $_;
		  next if (/^#/);
		  if($state){
			if(/^\s*\[/){
			  last;
			}
			if(/^\s*(\S+)\s*=\s*(\S+)/){
			  $key=$1;
			  $value=$2;
			  if($key =~ /^user$/){
				$myDbUser = $value;
				# sometimes strings are quoted
				$myDbUser =~ s/^('|")(.*)\1$/$2/;
			  }elsif($key =~ /^pass$/){
				$myDbPasswd = $value;
				# sometimes strings are quoted
				$myDbPasswd =~ s/^('|")(.*)\1$/$2/;
			  }
			}
		  }else{
			if(/^\s*\[client\]/){
			  $state=1;
			}
		  }
		}
		close (CNF);
      }
    }elsif($myDbType eq 'Pg'){
      $myDbPasswd ||= $ENV{'PGPASSWORD'};
      $myDbUser ||= $ENV{'PGUSER'};
    }
  }

  if (eval "require DBI"){
    if (eval "require DBD::$myDbType"){
      if($myDbAddress){
		if($myDbUser && $myDbPasswd){
		  $dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}elsif($myDbUser){
		  $dbh = DBI->connect($myDbAddress,$myDbUser) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}else{
		  $dbh = DBI->connect($myDbAddress) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}
      }
    }
  }
#  unless($dbh){
#    if ($myDbType eq 'Pg'){
#      if(eval "require Pg"){
#        $myDbAddress="user=$myDbUser dbname=$myDbDB";
#        if($myDbPasswd){
#          $myDbAddress.=" password=$myDbPasswd";
#        }
#        $dbh = Pg::connectdb($myDbAddress);
#      }
#    }
# }
	unless ($dbh) {

		if ($myDbType eq 'mysql') {
			$shell = $ptrBin->{'mysql'} || 'mysql';
			$args  = " --socket='$myDbSocket'" if ($myDbSocket);
			if ($myDbPasswd) {
				$args .= " -B -u '$myDbUser' -p'$myDbPasswd' '$myDbDB' ";
			} else {
				$args .= " -B -u $myDbUser $myDbDB ";
			}
			$shell = "LD_LIBRARY_PATH=$variables $shell" if ($variables);
    }elsif($myDbType eq 'Pg'){
      $shell = $ptrBin->{'psql'}||'psql';
	  if($myDbPasswd) {
		$shell = "PGPASSWORD=$myDbPasswd $shell";
	  }
	  if($myDbUser) {
		$shell = "PGUSER=$myDbUser $shell";
	  }

      $args = " -d $myDbDB -c ";
    }
#
# check connection to Confixx databnase
#
    if($myDbDB=~/confixx/){
      $sql = "select hostname from admin";
      @result = `$echo "$sql"| $shell $args`;
      unless($result[0]=~/^\w+$/){
		$args = ' -h $myDbHost'.$args;
		@result = `$echo "$sql"| $shell $args`;
		unless($result[0]=~/^\s*hostname/){
		  die "Error: Can not connect to Confixx database (unable to execute sql: $sql)\n@result\n";
		}
      }
    }
#
# end check connection to Confixx databnase
#
    $shell .= $args;
  }
  $this = {
	   'RECONNECT'=>sub{
	     if($dbh){
	       $dbh->disconnect();
		   if($myDbPasswd){
			 $dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
			   die "Error: unable to connect to database: $DBI::errstr\n";
		   }else{
			 $dbh = DBI->connect($myDbAddress,$myDbUser) or
			   die "Error: unable to connect to database: $DBI::errstr\n";
		   }
	       return $dbh;
	     }else{
	       return 1;
	     }
	   },
	   'EXECUTE'=> sub {
	     $sql = shift;
	     if($dbh){
	       $sth = &executeDbh($dbh,$sql);
	       if($sth){
		 return $sth->rows;
	       }else{
		 return 0;
	       }
	     }else{
	       $resultIndex=0;
	       &executeDbShell($shell,\@result,$echo,$sql,$myDbType);
	       return scalar(@result);
	     }

	   },
	   'FETCHROW'=> sub {
	     if($sth){
	       if(@result = $sth->fetchrow){
		 return \@result;
	       }else{
		 return undef;
	       }
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
		 return undef;
	       }else{
		 return $result[$resultIndex];
	       }
	     }
	   },
	   'FETCHHASH'=>sub {
	     if($sth){
	       $ptrResult = $sth->fetchrow_hashref();
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
		 $ptrResult = undef;
	       }else{
			 my ($i,$key);
			 $ptrResult = {};

			 for($i=0; $i<@{$result[0]}; $i++){
			   $ptrResult->{$result[0]->[$i]}=$result[$resultIndex]->[$i];
			 }
	       }
	     }
	     return $ptrResult;
	   },
	   'FINISH'=>sub {
	     if($sth){
	       $sth->finish();
	     }
	     @result=();
	     $ptrResult = undef;
	   },
	   'PRINTERROR'=>sub {
	     if($dbh){
	       if(@_){
			 $dbh->{'PrintError'}=shift;
	       }
	       return $dbh->{'PrintError'};
	     }else{
	       return 0;
	     }
	   },
	   'PING'=>sub{
	     if($dbh){
	       return $dbh->ping();
	     }else{
	       return 1;
	     }
	   },
	   'DISCONNECT'=>sub{
	     if($dbh){
	       return $dbh->disconnect();
	     }else{
	       return 1;
	     }
	   },
	   'USER'=>sub{
	     return $myDbUser;
	   },
	   'PASSWORD'=>sub{
	     return $myDbPasswd;
	   },
	   'HOST'=>sub{
	     return $myDbHost;
	   },
	   'TYPE'=>sub{
	     return $myDbType;
	   },
	  };

  return $this;
}

sub executeDbh {
  my $dbh = shift;
  unless(ref($dbh)=~/DBI/){
    &printToError ("Error: DBI object is not found");
  }
  my $sql = shift;

  my $sth = $dbh->prepare($sql) or
    &printToError ("Error: prepare sql: $sql\n$DBI::errstr");
  if($sth->execute){
    return $sth;
  }else{
    &printToError ("Error: exesute sql: $sql\n$DBI::errstr");
  }
}

sub executeDbShell {
  my ($shell,$ptrRes,$echo,$sql, $type) = @_;
  @{$ptrRes}=();
  my (@out,@row,$ret);
  $ret=0;
  if($type eq 'Pg'){
    @out = `$shell "$sql"`;
  }else{
    @out = `$echo "$sql"|$shell`;
  }
  if ($out[0]=~/^error/i){
    &printToErrror ("Error: execute sql: $sql\n@out");
  }else{
    if($type eq 'Pg'){
      foreach my $line (@out){
	chomp $line;
	next if ($line=~/^-[-+]*$/); # skip delimiter
	if($line=~/^\(\s*\d+\s+\w+\)$/){
	    last;  ## (6 rows)
	}
	$line=~s/^\s+//;
	$line=~s/\s+$//;
	@row = split(/\s*\|\s*/,$line);
	push @{$ptrRes}, [@row];
	$ret++;
      }
    }else{
      if($out[0]=~/^\+[-+]*$/){
	foreach my $line (@out){
	  chomp $line;
	  next if ($line=~/^\+[-+]*$/); # skip delimiter
	  $line=~s/^\|\s*//;
	  $line=~s/\s+$//;
	  @row = split(/\s*\|\s*/,$line);
	  @row = map { $_ eq 'NULL' ? undef : $_ } @row;
	  push @{$ptrRes}, [@row];
	  $ret++;
	}
      }else{  ## batch mode
	foreach my $line (@out){
	  chomp $line;
	  @row = split(/\t/,$line);
	  @row = map { $_ eq 'NULL' ? undef : $_ } @row;
	  push @{$ptrRes}, [@row];
	  $ret++;
	}
      }
    }
  }
  return $ret;
}

#
# end class for handle with database
#

sub makeMIMEBase64 {
  my ($useModule,$this);
  if (eval "require MIME::Base64"){
    $useModule = 1;
  }else{
    $useModule = 0;
  }

  $this={
	 'ENCODE'=>sub{
	   my $src = shift;
	   my ($encoded);
	   if($useModule){
	     $encoded = MIME::Base64::encode($src);
		 chomp $encoded;
	   }else{
	     $encoded = &encodeBase64($src);
	   }
	   return $encoded;
	 },
	 'USEMODULE'=>sub{
	   return $useModule;
	 },
	};

  return $this;
}

#
# url decode (certificates in Plesk DB encoded by the url encoding)
#

sub urlDecode {
    my $url = $_[0];
    $url =~ tr/+/ /;
    $url =~ s/%([a-fA-F0-9]{2,2})/chr(hex($1))/eg;
    return $url;
}

sub encodeHtml {
  my(@ret,$src);
  if ($src=shift){
    if(ref $src){
      if(${$src}=~/[&"'><]/){
		${$src}=~s/&/&amp;/sg;
		${$src}=~s/>/&gt;/sg;
		${$src}=~s/</&lt;/sg;
		${$src}=~s/\'/&#39;/sg;
		${$src}=~s/\"/&quot;/sg;
      }
      return ${$src};
    }else{
      if($src=~/[&"'><]/){
		$src=~s/&/&amp;/sg;
		$src=~s/>/&gt;/sg;
		$src=~s/</&lt;/sg;
		$src=~s/\'/&#39;/sg;
		$src=~s/\"/&quot;/sg;
      }
    }
  }
  return $src;
}

# Normalization according to the XML 1.0 (Third Edition),
# clauses 2.3 "Common Syntactic Constructs"
# and 3.3.3 "Attribute-value normalization"

sub normalizeAttribute {
  $_ = $_[0];
  s/\r/&#xD;/g;
  s/\n/&#xA;/g;
  s/\t/&#x9;/g;
  s/\</&lt;/g;
  s/\&/&amp;/g;
  s/\"/&quot;/g;
  s/\'/&#39;/g;
  return $_;
}

############################
#
#	Non-ascii characters handling
{
	# static
	my $charset;

	sub setCharset {
	  $charset = shift;
	}

	# try to determine the local charmap
	sub _determineCharset()
	{
	  my $charset;
	  my $locale = $bin{'locale'} || 'locale';

	  if (exists $ENV{'LANG'}) {
		my @pieces = split /\./, $ENV{'LANG'};
		if (defined $pieces[1]) {
		  $charset = $pieces[1];
		  chomp $charset;
		}
	  }
	  if (not $charset && open(CM, "$locale charmap |")) {
	  	binmode CM;
		$charset = <CM>;
		close(CM);
		chomp $charset;
	  }
	  if (not $charset) {
		$charset = 'ISO-8859-1';
	  }

	  return $charset;
	}

	sub toUtf8
	{
		my ($subj, $myCharset) = @_;

		$charset = _determineCharset() if (not $charset);

		$myCharset = $charset unless ( $myCharset );

		if ( (not $subj) ||
				 (not $bin{'iconv'}) ||
				 ( $myCharset =~ /UTF-?8/i )
			 ) {
			return $subj;
		}

		if ($subj =~ /^[\x00-\x7f]*$/) { return $subj; }

		my $converted;
		my $escaped = $subj;
		$escaped =~ s/'/'"'"'/g;

		if (open(IC, "echo '$escaped' | " . $bin{'iconv'} . " -f '$myCharset' -t 'UTF-8' |")) {
			binmode IC;
			while (my $string = <IC>) { $converted .= $string; }
			close(IC);
			chomp $converted;
			if ($converted) {
			  return $converted;
			}
		}

		return "";
	}
}

#
# Based on the b64.c by Bob Trower (base64.sf.net)
#

my @cb64 = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
			'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
			'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
			'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
			'g', 'h', 'i', 'j',	'k', 'l', 'm', 'n',
			'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
			'w', 'x', 'y', 'z', '0', '1', '2', '3',
			'4', '5', '6', '7', '8', '9', '+', '/');

@cb64 = map { ord } @cb64;

sub encodeBlock {
  my ($len, @in) = @_;
  my @res;
  push @res, $cb64[$in[0] >> 2];
  push @res, $cb64[(($in[0] & 0x3)<<4)|(($in[1]&0xf0)>>4)];
  push @res, $len>1 ? $cb64[(($in[1]&0xf)<<2)|(($in[2]&0xc0)>>6)] : ord('=');
  push @res, $len>2 ? $cb64[$in[2]&0x3f] : ord('=');
  return @res;
}

sub encodeBase64 {
  my @in = unpack "C*", $_[0];

  my @out;
  while (scalar(@in)) {
	my @block = (0, 0, 0);
	my $length = 0;
	while (scalar(@in) && $length < 3) {
	  $block[$length++] = shift(@in);
	}
	push @out, encodeBlock($length, @block);
  }
  return pack "C*", @out;
}

sub getBins {
  my ($ptrBin,$path) = @_;
  my (%searchPath,$key,$prefix,$bin,$fullPath);
  my ($notFound,$i,@prefixes);

  $notFound=0;

  $i=0;
  foreach $prefix (split(/:/,$path)){
    next unless $prefix;
    unless(exists($searchPath{$prefix})){
      $searchPath{$prefix}=$i++;
      push @prefixes,$prefix;
    }
  }

  foreach $key (keys %{$ptrBin}){
	$ptrBin->{$key} = undef;
    foreach $prefix (@prefixes){
      $prefix.='/' unless $prefix=~/\/$/;
      $bin = $key;
      $bin =~ s/^\///;
      $fullPath = $prefix.$bin;
      if (-x $fullPath){
		$ptrBin->{$key}=$fullPath;
		last;
      }
    }
    $notFound++;
  }
  return $notFound;
}

sub getArguments {
  my ($ptrArgv,$ptrOpts) = @_;

  my (@keys,$firstKey,%prepKeys,$key,$value,$ptrArr,$arg,$state);
  my (%retHash,$pat,$found,$used,@rest,$fullArg,$prevKey);

  while (($key,$value)=each(%{$ptrOpts})){
    @keys = split(/\|/,$key);
    $firstKey = $keys[0];
    $firstKey =~s/^-*//;
    push @{$prepKeys{$firstKey}},[@keys];
    push @{$prepKeys{$firstKey}},$value;
  }

  $state =0;
  foreach $arg (@{$ptrArgv}){

#    print "arg: $arg\n";

    $used = 0;
    if($state==1){
      if  ($arg =~ /^-/){
	$state=0;
      }else{
	$used = 1;
	$retHash{$prevKey}=$arg;
	next;
      }
    }
    if ($state == 2){
      $retHash{$prevKey}=$arg;
      $state = 0;
      $used = 1;
    }else{
      $fullArg = $arg;
      if ($arg =~ /^(-\S+)=(\S+)/){
	$arg = $1;
	$value = $2;
      }else{
	$value = undef;
      }
      foreach $key (keys %prepKeys){

	$ptrArr = $prepKeys{$key};

#	print "key: $key, ",join('|',@{$ptrArr->[0]}),"\n";


	$found = 0;
	foreach $pat (@{$ptrArr->[0]}){
	  if ($pat eq $arg){
	    $found = 1;
	    last;
	  }
	}
	if($found){
	  $used = 1;
	  if(defined($value)){
	    $retHash{$key}=$value;
	  }else{
	    if($ptrArr->[1]){
	      if($ptrArr->[1] =~ /\?$/){
		$state = 1;
	      }else{
		$state = 2;
	      }
	      $prevKey = $key;
	    }else{
	      $retHash{$key}=undef;
	    }
	  }
	  last;
	}
      }
    }
    unless($used){
      if($value){
	push @rest,$fullArg;
      }else{
	push @rest,$arg;
      }
    }
  }
  @{$ptrArgv}=@rest;
  return \%retHash;

}

sub makeDumpStatus($) {
  my $statusFile = shift;
  my($pid,$currAccount,$currDomain,$leftAccounts,$leftDomains,$this);
  $pid=$$;
  $this = {'FILE'=>sub {
	     if(@_){
	       $statusFile = shift;
	     }
	     return $statusFile;
	   },
	   'PID'=>sub {
	     if(@_){
	       $pid = shift;
	     }
	     return $pid;
	   },
	   'ACCOUNT'=>sub{
	     if(@_){
	       $currAccount = shift;
	       if($leftAccounts){
		 $leftAccounts--;
	       }
	     }
	     return $currAccount;
	   },
	   'DOMAIN'=>sub{
	     if(@_){
	       $currDomain = shift;
	       if($leftDomains){
		 $leftDomains--;
	       }
	     }
	     return $currDomain;
	   },
	   'COUNTS'=>sub{
	     if(@_){
	       my $newValue = shift;
	       if (defined $newValue){
		 $leftAccounts=$newValue;
	       }
	       $newValue = shift;
	       if (defined $newValue){
		 $leftDomains=$newValue;
	       }
	     }
	     return ($leftAccounts,$leftDomains);
	   },
	   'PRINT'=>sub{
	     if(@_){
	       $statusFile=shift;
	     }
	     open(STATUS,">$statusFile.tmp");

		 binmode STATUS;
	     my $root = &makeXmlNode('agent-dump-status');
	     my $item = &makeXmlNode('pid',$pid);
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('account',$currAccount||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domain',$currDomain||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('accounts-left',
				  defined($leftAccounts)?$leftAccounts:'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domains-left',
				  defined($leftDomains)?$leftDomains:'');
	     $root->{'ADDCHILD'}->($item);

	     &printXml($root,*STATUS,'agent-dump-status.dtd');

	     close(STATUS);

		 rename("$statusFile.tmp", $statusFile);
	   }

	  };
  return $this;
}

sub makeContentList {
  my ($dirDump,$xmlDump) = @_;
  my ($name,$fullPath,%files,%fromXML,$size,@cids,$root,$item,$key,$cmd,
      $gzip,$wc,$md5sum);

  if(opendir(DUMP,$dirDump)){
    while ($name=readdir(DUMP)){
      $fullPath="$dirDump/$name";
      unless(-f $fullPath && $name=~/\.\w\w$/){
	next;
      }
      $key = $fullPath;
      $key =~ s/\.\w\w$//;
      $size=(stat($fullPath))[7];
      if(defined $size){
	push @{$files{$key}},[$fullPath,$size];
      }
    }
    closedir(DUMP);
  }

	if (-T $xmlDump) {
		if (open(DUMP, "<$xmlDump")) {
			binmode DUMP;
			while (<DUMP>) {
				if (@cids = /cid(?:_\w+)?=\"([^"]+)\"/g) {
					foreach $fullPath (@cids) {
						unless ($fullPath =~ /^\//) {
							$fullPath = "$dirDump/$fullPath";
						}
						$fromXML{$fullPath} = undef;
					}
				}
			}
			close(DUMP);
		}

		foreach $key (keys %files) {
			unless (exists($fromXML{$key})) {
				delete $files{$key};
			}
		}
	}

  $gzip = $bin{'gzip'}||'gzip';
  $wc = $bin{'wc'}||'wc';
  $md5sum = $bin{'md5sum'}||'md5sum';

  $root=&makeXmlNode('contentlist');

  my($compressed,$uncompressed,$list,$count,$ptrList,@sortedList,
     $ptrItem,$out);
  my($totalUncomp,$totalSize)=(0,0);
  foreach $key (keys %files){

#
# sort by extension  .aa .ab ...
#
    @sortedList =  sort {&getExt($a->[0]) cmp &getExt($b->[0])} @{$files{$key}};

#
# output list of files
#
    $compressed=0;
    foreach $ptrItem (@sortedList){
      ($fullPath,$size)=@{$ptrItem};
      $item=&makeXmlNode('file',$fullPath);
      $item->{'ATTRIBUTE'}->('size',$size,'noencode');
      $root->{'ADDCHILD'}->($item);

#
# get md5sum
#
      if(-x $md5sum){
	$out=`$md5sum $fullPath`;
	if($out=~/\s*([0-9a-fA-F]+)\s+/){
	  $item->{'ATTRIBUTE'}->('md5sum',$1,'noencode');
	}
      }

      $compressed += $size;

    }

    $uncompressed = undef;
#
# get unpacked size of the archive
#
    if ($key=~/\.tgz$|\.gz$/){
      if(-x $gzip){
	if(scalar(@sortedList)==1){
	  ($fullPath,$size)=@{$sortedList[0]};
	  ($uncompressed,$compressed)=&getZipSize($fullPath,$gzip);
	}else{
	  $list = join(' ',map {$_->[0]} @sortedList);
	  if(-x $wc){
	    $cmd = "cat $list | $gzip | $wc -c";
	    $uncompressed = `$cmd`;
	    chomp $uncompressed;
	    if ($uncompressed =~ /^\s*(\d+)\s*$/){
	      $uncompressed = $1;
	    }else{
	      $uncompressed = undef;
	    }
	  }else{
	    $cmd = "cat $list >$key";
	    if(system($cmd)==0){
	      ($uncompressed,$compressed)=&getZipSize($key,$gzip);
	      unlink $key;
	    }
	  }
	}
      }
    }
#
# calc total size
#
    $totalSize += $compressed;
    $totalUncomp += $uncompressed if $uncompressed;
  }

  $root->{'ATTRIBUTE'}->('size',$totalSize,'noencode');
  if ($totalUncomp && $totalUncomp != $totalSize){
    $root->{'ATTRIBUTE'}->('unpacked',$totalUncomp,'noencode');
  }

  return $root;
}

sub getExt {
  my $path = shift;
  if($path=~/\.([^.]*)$/){
    return $1;
  }
  return $path;
}

sub getZipSize {
  my ($fullPath,$gzip) = @_;
  $gzip ||= 'gzip';
  my $cmd = "$gzip -l $fullPath";
  my ($uncompressed,$compressed,@out);
  if (open(OUTPUT,"$cmd|")){
    ($compressed,$uncompressed)=(0,0);
    while(<OUTPUT>){
      @out = split(' ',$_);
      if($out[0]=~/^\d+$/ && $out[1]=~/^\d+$/){
	$uncompressed+=$out[1];
	$compressed+=$out[0];
      }
    }
    close (OUTPUT);
  }
  return $uncompressed,$compressed;
}

my $outputIsFile;

sub openOutput
{
	my ($outPath) = @_;
	my ($outFh);

	if ($outPath) {
		open(OUTPUT, ">>$outPath") or die "Error: can not open '$outPath' for write: $!\n";
		binmode OUTPUT;
		$outFh = *OUTPUT;
		$outputIsFile = 1;
	} else {
		$outFh = *STDOUT;
		$outputIsFile = 0;
	}

	return $outFh;
}

sub closeOutput{
  if($outputIsFile){
    close (OUTPUT) or
      &printToError("Error: unable to close OUTPUT: $!\n");
    $outputIsFile=0;
  }
}

#=========================================================================
#
# dtd-specific (but common) subroutines
#
sub createPasswordNode {
  my ($parent, $password, $type) = @_;

  my $item = makeXmlNode('password', $password);
  $item->{'ATTRIBUTE'}->('type', $type);
  $parent->{'ADDCHILD'}->($item);
}

sub createSysuserNode {
  my ($parent, $name, $password, $type) = @_;

  my $item = makeXmlNode('sysuser');
  $item->{'ATTRIBUTE'}->('name', $name);
  createPasswordNode($item, $password, $type);
  $parent->{'ADDCHILD'}->($item);
}

sub readFile {
  my ($fileName) = @_;

  unless (-f $fileName) {
	return undef;
  }

  open(INPUT, "< $fileName");
  binmode INPUT;
  my $content = join('', <INPUT>);
  close(INPUT);

  return $content;
}

1;

#
# end common subroutines
#
#=========================================================================


$ptrArgs = &getArguments (\@ARGV,\%arg_opts);

#
# search binaries
#

&getBins(\%bin,'/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:/usr/local/sbin:'.$ENV{'PATH'});

my ($outPath,$outFh,$outHas,$dumpDir,$dumpFile,$statusFile,$app,$pwd,$workDir);

$app = $0;
if ($app=~/(.+)\/([^\/]+)$/){
  $workDir = $1;
  $app = $2;
}
$app=~s/\.pl$//;

$pwd=`$bin{'pwd'}`||'.';

chop $pwd;

if($workDir eq '.'){
  $workDir = $pwd;
}elsif($workDir !~ /^\//){
  $workDir = $pwd.'/'.$workDir;
}

$dumpFile = $workDir.'/ConfixxX.xml';
$statusFile = $workDir.'/ConfixxX.status';
$dumpDir = $workDir.'/pma';

my $objDumpStatus = &makeDumpStatus($ptrArgs->{'status-file'}||$statusFile);


if ( !$ptrArgs->{'output'} &&
   ( exists $ptrArgs->{'dump-all'} ||
     exists $ptrArgs->{'dump-accounts'}||
     exists $ptrArgs->{'dump-domains'}
   )
  ) {
  $ptrArgs->{'output'} = $dumpFile;
}

if ( $outPath = $ptrArgs->{'output'} ) {

  unless ( open( OUTPUT,">$outPath" ) ) {
    die "Error: can not open '$outPath' for write: $!\n";
  }
  $dumpFile = $outPath;
  close(OUTPUT);
}

if ( exists $ptrArgs->{'version'} ) {
  print $outFh $VERSION,"\n";
  exit 0;
}

#
# find & load main config
#

my $rootName = 'plesk7dump';
my $dtdName = 'ConfixxX.dtd';

unless ( &loadMainConfig( $ptrArgs->{'config'} ) ) {

  if ( exists $ptrArgs->{'get-status'} ) {
    my $outFh = &openOutput( $ptrArgs->{'output'} );

    &printAgentStatus( $outPath=$ptrArgs->{'output'}||$dumpFile,
		      $objDumpStatus->{'FILE'}->(),
		      $outFh);
    &closeOutput();

    exit 0;

  }else{
    if ( exists( $ptrArgs->{'dump-all'} ) ||
         exists( $ptrArgs->{'dump-accounts'} ) ||
         exists( $ptrArgs->{'dump-domains'} ) ) {

      my $root = &makeEmptyDump( $rootName );

      if ( ref( $root ) =~ /HASH/ && ref( $root->{'PRINT'} ) =~ /CODE/ ) {
				my $outFh = &openOutput( $ptrArgs->{'output'} );
				&printXml( $root,$outFh,$dtdName );
				&closeOutput();
      }
    }

    die "Error: confixx config file 'confixx_main.conf' is not found\n";
  }
}

my($majorVersion,$minorVersion,$patchLevel) = &getConfixxVersion();


#
# get db connect
#
my $wrapDbh = getDbConnect(\%bin,$dbType,$db_address,
			   $dbUser, $dbPw, $dbDB, $dbServer);
unless($wrapDbh){
  die "Error: can not connect to confixx database\n";
}

my ($mysqlUserDbAddress,$wrapUserMysql);
$mysqlUserDbAddress = "DBI:mysql:mysql:$mysqlUserServer";
$mysqlUserPort ||= $mysql_sock;
if ( $mysqlUserPort ) {
	if ( -S $mysqlUserPort ) {
		$mysqlUserDbAddress .= ";mysql_socket=$mysqlUserPort";
	} elsif ( $mysqlUserPort =~ /^\d+$/ ) {
		$mysqlUserDbAddress .= ";port=$mysqlUserPort";
	}
}

if ( $mysqlUserUser ){
	$mysqlUserServer ||= 'localhost';
	$wrapUserMysql = getDbConnect(\%bin,'mysql',$mysqlUserDbAddress,
																	 $mysqlUserUser, $mysqlUserPw, 'mysql',$mysqlUserServer);
}

#
# get MIME Base64 encoder
#
my $wrapBase64 = makeMIMEBase64();

my $ID=0;

if(exists $ptrArgs->{'get-status'}){

  my $outFh = &openOutput($ptrArgs->{'output'});

  &printAgentStatus($outPath=$ptrArgs->{'output'}||$dumpFile,
		    $objDumpStatus->{'FILE'}->(),
		    $outFh);

  &closeOutput();

}elsif(exists ($ptrArgs->{'dump-all'}) ||
       exists ($ptrArgs->{'dump-accounts'}) ||
       exists ($ptrArgs->{'dump-domains'})){

  my ($root,@accounts,@domains,$ptrAccounts,$ptrDomains, $value);
  if(exists $ptrArgs->{'no-content'}){
    $::dumpDir = &getDumpDir("$workDir/pma");
  }else{
    $::dumpDir = &getDumpDir("$workDir/pma",'clean');
  }

  if (exists $ptrArgs->{'no-compress'}) {
	setCompress(0);
  }

  &printToLog("Work dir: $workDir");
  &printToLog("Dump file: $dumpFile");
  &printToLog("Status file: ".$objDumpStatus->{'FILE'}->());


  if ($value = $ptrArgs->{'dump-accounts'}){
  	if ($value eq "-") {
		$value = <STDIN>;
		chomp $value;
	}
    @accounts = split(/\s*,\s*/,$value);
    $ptrAccounts = \@accounts;
  }

  if ($value = $ptrArgs->{'dump-domains'}){
  	if ($value eq "-") {
		$value = <STDIN>;
		chomp $value;
	}
    @domains = split(/\s*,\s*/,$value);
    $ptrDomains = \@domains;
  }
#
# generate a xml dump
#

  $root = &getConfixx2PleskDump($dumpDir, $ptrAccounts, $ptrDomains, $rootName);

#
# print dump to output
#
  &printToLog('Print dump file ...',1);


  if(ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/){
    my $outFh = &openOutput($ptrArgs->{'output'});

    &printXml($root,$outFh,$dtdName);

    &closeOutput();
  }
  &printToLog(' OK');

}elsif(exists $ptrArgs->{'get-content-list'}){

  $::dumpDir = &getDumpDir("$workDir/pma");

  my $root = &makeContentList($::dumpDir,$dumpFile);

  if(ref($root)=~/HASH/){
    my $outFh = &openOutput($ptrArgs->{'output'});

    &printXml($root,$outFh);

    &closeOutput();
  }
}else{
  &printHelp();
}

exit $returnCode;

#
# end main
#
#==============================================================

#==============================================================
#
# confixx subroutines
#

my (%admin); ## hash with admin settings

sub getConfixx2PleskDump{
  my($dumpDir,$ptrAccounts,$ptrDomains,$rootName) = @_;
  $rootName = 'plesk7dump' unless $rootName;
  my ($root,$item,%ipMap,$reseller,%resellers,%customers,$sql,$ptrRow,
     $ptrResellers,$full);

  $root = &makeXmlNode($rootName);

  unless(ref($wrapDbh)=~/HASH/ && ref($wrapDbh->{'EXECUTE'})=~/CODE/){
    $root->{'CONTENT'}->("Error: there is no connect to database");
    &printToError("Error: unable to connect to SQL database");
    return $root;
  }

#
# get admin settings
#
	$sql = "SELECT * FROM admin";
	if ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {
		my ( $ptrHash );
		if ( $ptrHash = $wrapDbh->{'FETCHHASH'}->() ) {
			%admin = %{$ptrHash};
		}
	}
#
# end get admin settings
#

#
# prepare ip
#
  &printToLog('Dump ip map ...',1);

  $item = &makeIpNode('map',\%ipMap);
  $root->{'ADDCHILD'}->($item);

  &printToLog(' OK');
#
# end prepare ip
#


#
# prepare reseller
#
  if(
     (ref($ptrAccounts)=~/ARRAY/ && (@{$ptrAccounts}>0)) ||
     (ref($ptrDomains)=~/ARRAY/ && (@{$ptrDomains}>0))
    ){
    my (@res,@cust,$acc,$list);
    if (ref($ptrAccounts)=~/ARRAY/){
      foreach $acc (@{$ptrAccounts}){
		if ($acc=~/^res\d+$/){
		  $resellers{$acc}=1;
		}else{
		  $customers{$acc}=1;
		}
      }
    }
    $list = join(',',keys %customers);
    if($list){
      $sql = "select distinct anbieter from kunden where (kunde in ($list))";
      if($wrapDbh->{'EXECUTE'}->($sql)){
				while( $ptrRow = $wrapDbh->{'FETCHROW'}->()){
					$acc = $ptrRow->[0];
					unless(exists $resellers{$acc}){
						$resellers{$acc} = 0;
					}
				}
			}
      $wrapDbh->{'FINISH'}->();
    }

    if (ref($ptrDomains)=~/ARRAY/){
      $list = join(',',map {"'$_'"} @{$ptrDomains});
      if($list){
				$sql = "select distinct anbieter,kunde from domains where (domain in ($list))";
				if($wrapDbh->{'EXECUTE'}->($sql)){
					while( $ptrRow = $wrapDbh->{'FETCHROW'}->()){
						$acc = $ptrRow->[0];
						unless(exists $resellers{$acc}){
							$resellers{$acc} = 0;
						}
						$acc = $ptrRow->[1];
						unless(exists $customers{$acc}){
							$customers{$acc} = 0;
						}
					}
				}
				$wrapDbh->{'FINISH'}->();
      }
    }
  }else{
    if($wrapDbh->{'EXECUTE'}->("select anbieter from anbieter")){
      while($ptrRow = $wrapDbh->{'FETCHROW'}->()){
				$resellers{$ptrRow->[0]}=1;
      }
    }
    $wrapDbh->{'FINISH'}->();
  }
  $ptrResellers = \%resellers;

  my $domainsCount = 0;
  if ($ptrDomains) {
		$domainsCount = scalar(@{$ptrDomains});
  }

  $objDumpStatus->{'COUNTS'}->(scalar (keys %resellers), $domainsCount);

  my %leftResellers = %resellers;
  my ($list);

  while (($reseller,$full)=each(%resellers)){
    &printToLog("Client '$reseller' is started");
	#
	# work with status of dump
	#
    $objDumpStatus->{'ACCOUNT'}->($reseller);
    $objDumpStatus->{'PRINT'}->();
    $list = join(',',map {"'$_'"} keys %leftResellers);
    if($list){
      $sql = "SELECT count(domain) FROM domains WHERE (richtigedomain in (0,1)) ".
				" AND (anbieter in ($list))";
      if($wrapDbh->{'EXECUTE'}->($sql)){
				if($ptrRow = $wrapDbh->{'FETCHROW'}->()){
					$domainsCount = $ptrRow->[0];
				}
      }
      $wrapDbh->{'FINISH'}->();
    }else{
      $domainsCount=0;
    }

    $objDumpStatus->{'PRINT'}->();

	#
	# end work with status of dump
	#

    $item = &makeResellerNode($reseller,\%ipMap,$full,\%customers,$ptrDomains);
    if(ref($item)=~/HASH/ && ref($item->{'PRINT'})=~/CODE/){
      $root->{'ADDCHILD'}->($item);
      &printToLog("Client '$reseller' is successfully dumped");
    }else{
      &printToLog("Client '$reseller' is not dumped");
    }

    delete $leftResellers{$reseller};
  }

  $objDumpStatus->{'COUNTS'}->(0,0);
  $objDumpStatus->{'PRINT'}->();

#
# end prepare reseller
#

  return $root;
}


#
# sub makeResellerNode
#            - create a 'client'-node
# Arguments:
#            $nameRes   - name of reseller
#            $refIpHash - reference to hash of ip-info
#            $full      - 1 dump all domains of the reseller
#            $$ptrCustomersHash
#                       - reference to hash of customer-info to dump
#            $ptrDomainsArr
#                       - reference to array of domains to dump

sub makeResellerNode {
  my $nameRes = shift;
  my $refIpHash = shift || {};
  my $full = shift;
  $full = 1 unless defined($full);
  my $ptrCustomersHash = shift || {};
  my $ptrDomainsArr = shift || [];

  my ($item,$root,$sql,$ptrHash,$value,%ipPool);

	my $nocontent = (exists $ptrArgs->{'no-content'})?1:0;

  $sql = "select * from anbieter where anbieter ='$nameRes'";

  unless($wrapDbh->{'EXECUTE'}->($sql)){
    $wrapDbh->{'FINISH'}->();
    return undef;
  }
  unless($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
    $wrapDbh->{'FINISH'}->();
    return undef;
  }

  $root = &makeXmlNode('client');

  $value = $ptrHash->{'emailadresse'};

  $root->{'ATTRIBUTE'}->('name',$nameRes);

	unless ( $nocontent ) {

		$root->{'ATTRIBUTE'}->('email',$value) if $value;
#  $value = $ptrHash->{'firma'};
#  $root->{'ATTRIBUTE'}->('contact',$value) if $value;
		$root->{'ATTRIBUTE'}->('contact',$nameRes);

	}	## no content

	$item = &makePasswordNode($ptrHash->{'longpw'});
	$root->{'ADDCHILD'}->($item);

	unless ( $nocontent ) {
		if($ptrHash->{'gesperrt'}){
#
# locked
#
			$item = &makeXmlNode('status','false');
			$root->{'ADDCHILD'}->($item);
		}

		&addPinfo($root,$ptrHash,\%mapResPinfo);

		&addLimits($root,$ptrHash,\%mapResLimits);

		&addPermissions($root,$ptrHash,\%mapResPermis);
	}

  $wrapDbh->{'FINISH'}->();

#
# IP-pool
#

	$item = &makeIpPoolNode($nameRes,$refIpHash,\%ipPool);
	if ( ref($item) =~ /HASH/ ) {
		$root->{'ADDCHILD'}->($item);
	}
#
# end IP-pool
#
  my (@domains,$domain,@subdomains,$subdomain,@addSubDomain,%allSubDomains,
		 $ptrDomains,$ptrSubdomains,$ptrRow, %do2ku);

  if($full){
    $sql = "SELECT domain, kunde FROM domains WHERE (richtigedomain=1) AND (anbieter='$nameRes')";
    $wrapDbh->{'EXECUTE'}->($sql);
    while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
      push @domains, $ptrRow->[0];
			$do2ku{$ptrRow->[0]} = $ptrRow->[1];
    }
    $wrapDbh->{'FINISH'}->();

  }else{
    while (my($cust,$full)=each (%{$ptrCustomersHash})){
      next unless $full;
      $sql = "SELECT domain, kunde FROM domains WHERE (richtigedomain=1) AND (anbieter='$nameRes') AND (kunde='$cust')";
      $wrapDbh->{'EXECUTE'}->($sql);
      while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
				push @domains,$ptrRow->[0];
				$do2ku{$ptrRow->[0]} = $ptrRow->[1];
      }
    }
    my $list = join(',',map {"'$_'"} @{$ptrDomainsArr});
    if($list){
      $sql = "SELECT domain, kunde FROM domains WHERE richtigedomain IN (0,1) ".
				"AND (anbieter='$nameRes') AND (domain IN ($list))";
      $wrapDbh->{'EXECUTE'}->($sql);
      while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
				push @domains,$ptrRow->[0];
				$do2ku{$ptrRow->[0]} = $ptrRow->[1];
      }
      $wrapDbh->{'FINISH'}->();
    }
  }

#
# loop by domains
#

	my ($initFactor,%limitFactor,$cnt,$newFactor,$countExtraDomains,$kunde);

	%limitFactor = map {$_ => 1} @domains;

  while (@domains){

		$domain = shift @domains;
		$kunde = $do2ku{$domain};

    &printToLog("Domain '$domain' is started");

    $objDumpStatus->{'DOMAIN'}->($domain);
    $objDumpStatus->{'PRINT'}->();

		%allSubDomains = ($domain => 1); ## init hash

    ($ptrDomains,$ptrSubdomains) = &addDomainNode($root,$domain,$refIpHash,\%ipPool);

		map { $allSubDomains{$_} = 1 } @{$ptrSubdomains}; ## add prepared subdomains

		if ( $cnt = @subdomains = @{$ptrDomains} ) {

#
# calculate adjustment factor
#
			$limitFactor{$domain} ||= 1;
			$newFactor = $limitFactor{$domain}*($cnt+1); ## subdomains + $domain
			$limitFactor{$domain} = $newFactor;
			map { $limitFactor{$_} = $newFactor } @subdomains;

#
# translate subdomain to domains
#
			while ( @subdomains ) {
				$subdomain = shift(@subdomains);
				&printToLog ( "transfer subdomain '$subdomain' to domain is started" );

				$objDumpStatus->{'DOMAIN'}->($subdomain);
				$objDumpStatus->{'PRINT'}->();

				$allSubDomains{$subdomain} = 1; ## add transfered to domain subdomain
				$do2ku{$subdomain} = $kunde;

				($ptrDomains,$ptrSubdomains) = &addDomainNode($root,$subdomain,$refIpHash);
				$countExtraDomains++;

				map { $allSubDomains{$_} = 1 } @{$ptrSubdomains}; ## add prepared subdomains
				push @subdomains,@{$ptrDomains};

#
# calculate adjustment factor
#
				if ( $cnt = @{$ptrDomains} ) {
					$limitFactor{$subdomain} ||= 1;
					$newFactor = $limitFactor{$subdomain}*($cnt+1); ## subdomains + $domain
					$limitFactor{$subdomain} = $newFactor;
					map { $limitFactor{$_} = $newFactor } @{$ptrDomains};
				}

				&printToLog ( "sbdomain '$subdomain' is transfered" );
			}
		}

#
# add to domain subdomains with yet not processed
#
		my $list = join(',',map {"'$_'"} keys %allSubDomains);
		if ( $list ) {
			my (%subdom,$cnt,@pnts);
			$sql = "SELECT domain FROM domains WHERE richtigedomain = 0 AND ".
				" ( domain like '\%.$domain' ) AND NOT ( domain IN ($list) )";
      $wrapDbh->{'EXECUTE'}->($sql);
      while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
				$cnt = @pnts = ($ptrRow->[0] =~ /\./g);  ## get level of subdomain (count of dot)
				if ( $ptrRow->[0] =~ /\*/ ) { ## skip wildcard-subdomains
					$allSubDomains{$ptrRow->[0]} = 1;
				} else {
					push @{$subdom{$cnt}},$ptrRow->[0];
				}
      }
      $wrapDbh->{'FINISH'}->();
			if ( %subdom ) {
#
# claculate minimum of subdomain's levels
#
				my $minCnt = (sort { $a <=> $b } keys %subdom)[0];

#
# transfer subdomain to domain
#

				map {
					push( @domains, $_ );
					$do2ku{$_} = $kunde;	
					$countExtraDomains++;
				} @{$subdom{$minCnt}};

#
# recalc limit factor
#
				my ($newFactor,$delta,$reDomain,$oldFactor);

				$limitFactor{$domain} ||= 1;
				$oldFactor = $limitFactor{$domain};
				$newFactor = @{$subdom{$minCnt}}+$oldFactor;

## correct old factor for currect domain
				$delta = $newFactor/$oldFactor;
				$reDomain = qr/\Q$domain\E$/;
				map { $limitFactor{$_} *= $delta if /$reDomain/ } keys %limitFactor;

## add new factor
				map { $limitFactor{$_} = $newFactor } @{$subdom{$minCnt}};

			}
		}

    &printToLog("Domain '$domain' is dumped");

  }

	my $ptrKundeLimits =  &getKundeLimits ( $root, \%do2ku );

	&correctLimits( $root, \%limitFactor );

	&correctClientsLimits ( $root, $countExtraDomains, $ptrKundeLimits, \%do2ku );


#
# end loop by domains
#
  return $root;
}

sub correctLimits {
	my ($root, $ptrFactor, $cntExtraDomains ) = @_;
	unless( ref($root)=~/HASH/ && ref($ptrFactor)=~/HASH/ ){
		&printToError("Error: sub correctLimits: wrong parameters");
		return;
	}

	my($ptrKinds,$item,$domain,$domainItem,$factor,$ptrLimits,
		 $value,$ret,$limitItem,$name,%limits, $itemName, $limitMaxDomains);
	$ret = 0;

#
# clean summators
#
	foreach $name (keys %comulatLimits) {
		$limits{$name} = undef;
	}
#
# end clean summators
#

	$ptrKinds = $root->{'KINDSREF'}->();
	foreach $domainItem ( @{$ptrKinds} ) { ## loop by domains
		$itemName = $domainItem->{'NAME'}->();
		if ( $itemName eq 'domain' ) {

			$domain = $domainItem->{'ATTRIBUTE'}->('name');

			$factor = $ptrFactor->{$domain};
			next unless $factor > 1; ## There is a factor which more than 1

			&printToLog ("update limits of '$domain' with factor $factor");

			$ptrLimits = $domainItem->{'KINDSREF'}->();
			foreach $limitItem ( @{$ptrLimits} ) { ## loop by limits
				next unless $limitItem->{'NAME'}->() eq 'limit';

				$name = $limitItem->{'ATTRIBUTE'}->('name');
				$value = $limitItem->{'CONTENT'}->();
				if ( $value>0 ) { ## a positive value
					
					if ( exists( $limits{$name} ) ) { ## comulative limit
						if ( $limits{$name} ) {
							$limits{$name} += $value;
						} else {
							$limits{$name} = $value;
						}
						
					} else {
						$value /= $factor;
						$limitItem->{'CONTENT'}->( int($value) );
						$ret++;
					}
				}
			}
		} elsif ( $itemName eq 'limit' ) {
			$name = $domainItem->{'ATTRIBUTE'}->('name');
			if ( $name eq 'max_dom' ) {
				$limitMaxDomains = $domainItem; ## store reference
			}
		}
	}

#
# correct maximum of domains
#	
	if ( $limitMaxDomains && $cntExtraDomains ) {
		$value = $limitMaxDomains->{'CONTENT'}->();
		if ( $value > 0 ) {
			$value += $cntExtraDomains;
			$limitMaxDomains->{'CONTENT'}->($value);
		}
	}
#
# /correct maximum of domains
#


#
# correct comulative limits
#
	foreach $limitItem ( @{$ptrKinds} ) { ## loop by limits
		next unless $limitItem->{'NAME'}->() eq 'limit';

		$name = $limitItem->{'ATTRIBUTE'}->('name');
		next unless exists $limits{$name}; ## check if name is suit

		$value = $limitItem->{'CONTENT'}->();
		if ( ( $value > 0 ) && ( $value < $limits{$name} ) ) {
			$limitItem->{'CONTENT'}->( $limits{$name} ); ## increase limit
		}
	}
#
# end correct comulative limits
#

	return $ret;
}
#
#
#

sub getKundeLimits {
	my ( $root, $ptrDo2Ku ) = @_;
	unless( ref($root) =~ /HASH/ && 
					ref($ptrDo2Ku) =~ /HASH/ 
				){
		&printToError("Error: sub getKundeLimits: wrong parameters");
		return undef;
	}

	my ( $ptrKinds, $domainItem, $itemName, $domain, $kunde, $name, $value,
			 $ptrLimits, $limitItem);

	my $ret = {}; ## make an empty hash

	$ptrKinds = $root->{'KINDSREF'}->();
	foreach $domainItem ( @{$ptrKinds} ) { ## loop by domains
		$itemName = $domainItem->{'NAME'}->();

		if ( $itemName eq 'domain' ) {
			$domain = $domainItem->{'ATTRIBUTE'}->('name');
			$kunde = $ptrDo2Ku->{$domain};
			next unless $kunde;
			next if exists $ret->{$kunde};

			$ret->{$kunde} = {};

			$ptrLimits = $domainItem->{'KINDSREF'}->();

			foreach $limitItem ( @{$ptrLimits} ) { ## loop by limits
				next unless $limitItem->{'NAME'}->() eq 'limit';

				$name = $limitItem->{'ATTRIBUTE'}->('name');
				$value = $limitItem->{'CONTENT'}->();
				if ( $value > 0 ) {
					$ret->{$kunde}->{$name} = $value;
				}
			}
		}
	}
	return $ret;
}

sub correctClientsLimits {
	my ($root, $cntExtraDomains, $ptrKu2Lim,  $ptrDo2Ku ) = @_;

	unless( ref($root) =~ /HASH/ && 
					ref($ptrDo2Ku) =~ /HASH/ &&
					ref($ptrKu2Lim) =~ /HASH/
				){
		&printToError("Error: sub correctClientLimits: wrong parameters");
		return;
	}

	my ($ptrKinds, $domainItem, $itemName, $domain, $ptrLimits, %domLimits, $limitItem, %limit2dom,
			$name, $value, $limitMaxDomains, %cliLimits, $ptrArr, $kunde, $limName, %ku2do);


	while (($domain,$kunde) = each (%{$ptrDo2Ku})){
		push @{$ku2do{$kunde}}, $domain;
	}

	foreach $kunde (keys %ku2do){
		unless ( @{$ku2do{$kunde}} > 1 ) {
			delete $ku2do{$kunde};
		}
	}

	$ptrKinds = $root->{'KINDSREF'}->();
	foreach $domainItem ( @{$ptrKinds} ) { ## loop by domains
		$itemName = $domainItem->{'NAME'}->();

		if ( $itemName eq 'domain' ) {
			$domain = $domainItem->{'ATTRIBUTE'}->('name');
			$ptrLimits = $domainItem->{'KINDSREF'}->();

			$domLimits{$domain} = [ {}, $domainItem ];

			foreach $limitItem ( @{$ptrLimits} ) { ## loop by limits
				next unless $limitItem->{'NAME'}->() eq 'limit';

				$name = $limitItem->{'ATTRIBUTE'}->('name');
				$value = $limitItem->{'CONTENT'}->();
				if ( $value > 0 ) { ## a positive value
					$domLimits{$domain}->[0]->{$name} = [ $value, $limitItem ];
					unless ( exists( $limit2dom{$name} ) ) {
						$limit2dom{$name}={};
					}
					$limit2dom{$name}->{$domain} = $value;
				}
			}

		} elsif ( $itemName eq 'limit' ) {
			$name = $domainItem->{'ATTRIBUTE'}->('name');
			if ( $name eq 'max_dom' ) {
				$limitMaxDomains = $domainItem; ## store reference
			}
			$value = $domainItem->{'CONTENT'}->();
			if ( $value > 0 ) { ## a positive value
				$cliLimits{$name} = [ $value, $domainItem ];
			}
		}
	}

#
# correct maximum of domains
#	
	if ( $limitMaxDomains && $cntExtraDomains ) {
		$value = $limitMaxDomains->{'CONTENT'}->();
		if ( $value > 0 ) {
			$value += $cntExtraDomains;
			$limitMaxDomains->{'CONTENT'}->($value);
			$cliLimits{'max_dom'}->[0] = $value;
		}
	}
#
# /correct maximum of domains
#

	unless ( keys %ku2do ) {
		return 0;
	}

#
# check overdraft
#
	foreach $name ( keys %cliLimits )  {
		$value = 0;
		map { $value += $_ } grep { $_ > 0 } values %{$limit2dom{$name}};

		if ( $cliLimits{$name}->[0] <= 0 || $value <= $cliLimits{$name}->[0] ) {

			delete $cliLimits{$name};
			delete $limit2dom{$name};

			foreach $domain (%domLimits) {
				foreach $limName ( keys ( %{$domLimits{$domain}->[0]} ) ) {
					if ( $limName eq $name ) {
						delete $domLimits{$domain}->[0]->{$name};
					}
				}
			}

		}
	}

#
# main loop by limits
#
	my ( %limit, %fact, $limVal, $factVal, $factSum, @dom2work, $kf, $oldLimit, $newLimit,
		 $dom, $val, $newValue);

	foreach $name ( keys %cliLimits )  {

		print "correct '$name' limit ...";

#
# loop by kunde
#
		while ( ($kunde, $ptrArr) = each( %ku2do ) ) {
			%limit = ();
			%fact = ();
			@dom2work = ();
			$factSum = 0;
			$oldLimit = 0;
#
# get limits & facts
#
			foreach $dom (@{$ptrArr}) { ## get limits & facts
				$limVal = $domLimits{$dom}->[0]->{$name}->[0];
				next unless $limVal > 0;
				push @dom2work, $dom;
				$limit{$dom} = $limVal;
				$oldLimit += $limVal;

				$factVal = &getFactUsage( $name, $dom, $domLimits{$dom}->[1]);

				$fact{$dom} = defined( $factVal )? $factVal: $limVal;
				if ( $fact{$dom} > 0 && $factSum >= 0 ) {
					$factSum += $fact{$dom};
				} elsif ( $fact{$dom}< 0 ) {
					$factSum = -1;
				}
			}

#
# correct limits
#
			if ( $factSum <= 0 ) {
				if ( $oldLimit > 0 ) {
					$kf = $ptrKu2Lim->{$kunde}->{$name} / $oldLimit;
					$newLimit = 0;
					foreach $dom ( @dom2work ) {
						$val  = int( $limit{$dom} * $kf );
						$limit{$dom} = $val;
						$newLimit += $val;
					}
				}
			} else {
				if ( $factSum <= $ptrKu2Lim->{$kunde}->{$name} &&
						 $ptrKu2Lim->{$kunde}->{$name} > 0
					 ) {
					$kf = $ptrKu2Lim->{$kunde}->{$name} / $factSum;
				} else {
					$kf = 1;
				}
				$newLimit = 0;
				foreach $dom ( @dom2work ) {
					$val  = int( $fact{$dom} * $kf );
					$limit{$dom} = $val;
					$newLimit += $val;
				}
			}
#
# save new domain's limit
#
			while ( ($dom,$val) = each(%limit) ) {
				$domLimits{$dom}->[0]->{$name}->[0] = $val;
			}
#
# save new client's limit
#
			if ( $newLimit > $oldLimit ) {
				$cliLimits{$name}->[0] += ($newLimit - $oldLimit);			
			}
		}

		print "\t  done\n";

	}
#
# save to xml
#
	my $ret = 0;
	foreach $name (keys (%cliLimits)) {
		$domainItem = $cliLimits{$name}->[1];
		$value = $domainItem->{'CONTENT'}->();
		if ( $value < $cliLimits{$name}->[0] ) {
			$domainItem->{'CONTENT'}->( $cliLimits{$name}->[0] );
			$ret++;
		}
	}

	foreach $dom ( keys (%domLimits) ) {
		foreach $name ( keys %{$domLimits{$dom}->[0]} ) {
			$limitItem = $domLimits{$dom}->[0]->{$name}->[1];
			$value = $limitItem->{'CONTENT'}->();
			$newValue = $domLimits{$dom}->[0]->{$name}->[0];
			unless ( $value == $newValue ) {
				$limitItem->{'CONTENT'}->( $newValue );
				$ret++;
			}
		}
	}

	return $ret;

}

sub getFactUsage {
	my ( $limit, $domain, $xmlDomainItem) = @_;
	my ( $retValue, $item, $ret );

	if ( $limit eq 'max_subdomain'){
		$retValue = 0;
		$item = $xmlDomainItem->{'KIND'}->('phosting');
		if ( $item ) {
			$ret = $item->{'KINDSREF'}->('subdomain');
			if ( $ret ) {
				$retValue = @{$ret};
			}
		}

	} elsif ( $limit eq 'max_dom' ) {
# nothing to do

	} elsif ( $limit eq 'disk_space' ) {
		$retValue = 0;
		$item = $xmlDomainItem->{'KIND'}->('phosting');
		if ( $item ) {
			my ($path, $cmd);
			my $tar ||= $bin{'tar'} || 'tar';
			my $cat ||= $bin{'cat'} || 'cat';
			$ret = $item->{'ATTRIBUTESREF'}->();
			while ( my( $key, $value) = each( %{$ret} ) ){
				next unless ($key =~ /^cid(_|$)/);
				$path = "$dumpDir/$value";
				next unless ( -f "$path.aa" );
				$cmd = "cat $path.* | $tar -z -t -vv -f - ";
				if ( open( IN, "$cmd |") ) {
					while ( <IN> ) {
#
#-rw-r--r-- apache/wws      256 2005-02-15 09:28:21 ./_vti_cnf/.htaccess
#                      --> size <---
#
						$retValue += ( split( ' ', $_, 4 ) )[2]; 
					}
					close( IN );
				}
			}
			$retValue /= 1024;
		}		

  } elsif ( $limit eq 'max_traffic' ) {
		$retValue = -1;

	} elsif ( $limit eq 'max_wu' ) {

	} elsif ( $limit eq 'max_db' ) {
		$retValue = 0;
		$ret = $xmlDomainItem->{'KINDSREF'}->('database');
		if ( $ret ) {
			$retValue = @{$ret};
		}

	} elsif ( $limit eq 'max_box' ) {
		$retValue = 0;
		$item = $xmlDomainItem->{'KIND'}->('mailsystem');
		if ( $item ) {
			$ret = $item->{'KINDSREF'}->('mailuser');
			if ( $ret ) {
				foreach $item (@{$ret}) {
					if ( $item->{'KIND'}->('mailbox') ) {
						$retValue++;
					}
				}
			}
		}

	} elsif ( $limit eq 'mbox_quota' ) {
		$retValue = -1;

	} elsif ( $limit eq 'max_redir' ) {
		$retValue = 0;
		$item = $xmlDomainItem->{'KIND'}->('mailsystem');
		if ( $item ) {
			$ret = $item->{'KINDSREF'}->('mailuser');
			if ( $ret ) {
				foreach $item (@{$ret}) {
					if ( $item->{'KIND'}->('redirect') ) {
						$retValue++;
					}
				}
			}
		}

	} elsif ( $limit eq 'max_mg' ) {

	} elsif ( $limit eq 'max_resp' ) {

	} elsif ( $limit eq 'max_maillists' ) {
		$ret = $xmlDomainItem->{'KINDSREF'}->('maillist');
		$retValue = $ret? @{$ret}: 0;
		return $retValue;

	} elsif ( $limit eq 'max_webapps' ) {

	} elsif ( $limit eq 'ans_freq' ) {

	} elsif ( $limit eq 'ans_count' ) {

	}

	return $retValue;
} 


#
# ip pool
#
sub makeIpPoolNode {
	my ($nameRes,$refIpHash,$refRetHash) = @_;

  my (%ips,$nodePool,$ptrRow,$ip,$id,$sql,$item);

  $sql = "SELECT standardip FROM anbieter WHERE anbieter='$nameRes'";
  if ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {

    ( $ip ) = @{$wrapDbh->{'FETCHROW'}->()};
		$ip =~ s/^\s+//;
		$ip =~ s/\s+$//;

  }
  $wrapDbh->{'FINISH'}->();

	if ( $ip ) { ## check IP
		unless ( $ip eq $admin{'standardip'} ) {
			$sql = "SELECT ip FROM ipadressen WHERE ip='$ip'";
			unless ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {
				$ip = $admin{'standardip'};
			}
		}
	} else {	 ## get standard ip from admin
		$ip = $admin{'standardip'};
	}

	if ( exists $refIpHash->{$ip} ) {
		$ips{$ip} = $refIpHash->{$ip}->[0];
	}


  $sql = "SELECT ip FROM ipadressen WHERE anbieter='$nameRes'";
  if ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {
    while ( $ptrRow = $wrapDbh->{'FETCHROW'}->() ) {

      $ip = $ptrRow->[0];
			$ip =~ s/^\s+//;
			$ip =~ s/\s+$//;

      if ( exists $refIpHash->{$ip} ) {
				$ips{$ip} = $refIpHash->{$ip}->[0];
      }
	if ( ref($refRetHash) =~ /HASH/ ) {
		%{$refRetHash} = %ips;
	}

    }
  }

  $wrapDbh->{'FINISH'}->();
  $nodePool = &makeXmlNode('ip_pool');

  foreach $ip (sort keys %ips){
    $id = $ips{$ip};
    $item = &makeXmlNode( 'client_ip' );
    $item->{'ATTRIBUTE'}->( 'ip_ref',$id );
    $nodePool->{'ADDCHILD'}->( $item );
  }

	return $nodePool;
}
#
# end ip pool
#




#
# addDomainNode
#            - create and add a domain node into the xml tree
# Arguments:
#
#    $root      - the node in which will be added a domain node
#    $domain    - name of domain
#    $refIpHash - refernce to hash with ip-info
#


sub addDomainNode {
  my ($root,$domain,$refIpHash,$refResIp) = @_;

  my ($rootDomain,$sql,$ptrRow,$item,$ip,$id,$idnName,
      $ptrHash,%custHash,$customer,$ptrSubdomains,$ptrDomains);

	my $nocontent = (exists $ptrArgs->{'no-content'})?1:0;

## make anonymous array. need for return
	$ptrDomains = [];
	$ptrSubdomains = [];


  $rootDomain = &makeXmlNode('domain');
  $rootDomain->{'ATTRIBUTE'}->('name',$domain);
  $root->{'ADDCHILD'}->($rootDomain);

  $sql = "SELECT k.* FROM kunden k, domains d WHERE (d.kunde=k.kunde) AND (d.domain='$domain')";
  unless ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {
    $wrapDbh->{'FINISH'}->();
    &printToError( "Error: domain '$domain' is not found" );
    return 0;
  }
  unless($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
    $wrapDbh->{'FINISH'}->();
    return 0;
  }
  %custHash = %{$ptrHash};
  $wrapDbh->{'FINISH'}->();

#
# IDN
#
	if ( $majorVersion >= 3 ) {
		my ($idnCharset,$domainId);
		$sql = "SELECT i.idn_name, i.charset, d.id FROM idn_aliases i, domains d ".
			"WHERE i.domain_id = d.id AND d.domain = '$domain'";
		if ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {
			($idnName,$idnCharset,$domainId) = @{$wrapDbh->{'FETCHROW'}->()};

			if( $idnCharset && ( $idnCharset !~ /UTF-?8/i ) ) {
## translate to UTF8
				$idnName = &toUtf8($idnName, $idnCharset);
			}

			## use localized-name instead IDN-name (bug #43508)
			$rootDomain->{'ATTRIBUTE'}->('name',$idnName);
			$rootDomain->{'ATTRIBUTE'}->('idn','false');

			$wrapDbh->{'FINISH'}->();
		}
	}
	unless( $idnName ) {
		if ( $domain =~ /(^|\.)xn--[^.]+/ ){
			$rootDomain->{'ATTRIBUTE'}->('idn','true');
		}
	}
#
# /IDN
#

#
# get ip
#

	$ip = $custHash{'ip'};
	$ip =~ s/^\s+//;
	$ip =~ s/\s+$//;

	unless ( $ip ) {
		my $reseller = $custHash{'anbieter'};
		$sql = "SELECT standardip FROM anbieter WHERE anbieter = '$reseller'";
		if ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {

			( $ip ) = @{$wrapDbh->{'FETCHROW'}->()};
			$ip =~ s/^\s+//;
			$ip =~ s/\s+$//;

			$wrapDbh->{'FINISH'}->();
		}
	}

	if ( $ip ) { ## check if it is in client's IP-pool
		if ( ref($refResIp) =~ /HASH/ ) {
			unless ( exists( $refResIp->{$ip} ) ) {
				$ip = $admin{'standardip'};
			}
		}
	} else {
		$ip = $admin{'standardip'};
	}

	if ( ( $ptrRow = $refIpHash->{$ip} ) && ( ref($ptrRow) =~ /ARRAY/ ) ) {
		$item = &makeXmlNode('domain_ip');
		$id = $ptrRow->[0];
		$item->{'ATTRIBUTE'}->('ip_ref',$id);
		$rootDomain->{'ADDCHILD'}->($item);
	}

#
# end get ip
#

	$customer = $custHash{'kunde'};

#
# increase database's limit  by number of webShop's limit
#
	unless ( $nocontent ) {
		if ( $admin{'shops'} && $custHash{'maxshops'} ) {
			if ( $custHash{'maxshops'} > 0 ) {
				$custHash{'maxmysql'} += $custHash{'maxshops'}*2;
			} else {
				$custHash{'maxmysql'} = -1;
			}
		}

		$customer = $custHash{'kunde'};

		if ( $custHash{'maxmysql'} > 0 ) {

			$sql = "SELECT COUNT(dbname) FROM mysql_datenbanken WHERE kunde='$customer'";
			if ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {
				my ( $count ) = @{$wrapDbh->{'FETCHROW'}->()};
				$wrapDbh->{'FINISH'}->();

				if ( $admin{'shops'} ) {  ## include webShop's datases
					$sql  = "SELECT COUNT(shop_name) FROM shops WHERE kunde='$customer'";
					if ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {
						$count += $wrapDbh->{'FETCHROW'}->()->[0]*2; ## 2 databases for 1 webShop
						$wrapDbh->{'FINISH'}->();
					}
				}

				if ( $custHash{'maxmysql'} < $count ) {
					$custHash{'maxmysql'} = $count;
				}
			}
		}



		&addLimits($rootDomain,\%custHash,\%mapCustLimits);

		$item = &makeMailSystem($domain);
		if ( ref($item) =~ /HASH/ ) {
			$rootDomain->{'ADDCHILD'}->($item);
		}

#
# add DNS records
#
		my (@ns); ## default value for NS-records
		$sql = "SELECT pns,sns FROM anbieter WHERE anbieter='".$custHash{'anbieter'}."'";
		if ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {
			@ns = @{$wrapDbh->{'FETCHROW'}->()};
			$wrapDbh->{'FINISH'}->();
		}
		$ns[0] ||= $admin{'ip_pns'};
		$ns[1] ||= $admin{'ip_sns'};

		&addDNS ( $rootDomain, $domain, @ns );

#
# user's databases
#
		&addUsersDatabase($rootDomain,$customer);
#
# end user's databases
#
	}

	my (%domainHash);

	$sql="SELECT * FROM domains WHERE domain='$domain'";
	if ($wrapDbh->{'EXECUTE'}->($sql)){
		if($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
			%domainHash=%{$ptrHash};
		}
  }
  $wrapDbh->{'FINISH'}->();


	unless ( $nocontent ) {

#
# maillists
#
		&addMaillist($rootDomain,$domain,\%domainHash);

#
# end maillists
#


		if($custHash{'gesperrt'}){
#
# locked
#
			$item = &makeXmlNode('status','false');
			$rootDomain->{'ADDCHILD'}->($item);
		}



#
# ssl
#
		if ($domainHash{'cssl'}){
			$sql="SELECT crt,privatekey FROM cssl WHERE kunde='$customer'";
			if ($wrapDbh->{'EXECUTE'}->($sql)){
				if($ptrRow = $wrapDbh->{'FETCHROW'}->()){
					my($sslNode);
					$sslNode = makeXmlNode('certificate');
					$sslNode->{'ADDCHILD'}->(makeXmlNode('certificate-data', $ptrRow->[0]));
					$sslNode->{'ADDCHILD'}->(makeXmlNode('private-key', $ptrRow->[1]));
					$rootDomain->{'ADDCHILD'}->($sslNode);
				}
			}
			$wrapDbh->{'FINISH'}->();
		}
#
# end ssl
#

#
# domain user
#
		$item = &makeDomainUserNode( \%custHash );
		if ( ref($item) =~ /HASH/ ) {
			$rootDomain->{'ADDCHILD'}->($item);
		}

#
# end domain user
#

	} ## no content
#
# hosting
#
	my($hostingNode);

	if( $domainHash{'richtigedomain'} == 1 ||                                  ## domain
			( $domainHash{'richtigedomain'} == 0 && $domainHash{'pfad'} =~ /^\//)){  ## subdomain & no redirect

		my($plesk,$confixx,%attr);
		my($pathPrefix,$rePathPrefix,@excludeFiles);

		unless ( $nocontent ) {

			$hostingNode = &makeXmlNode('phosting');

#
# sysuser
#
			$pathPrefix = $domainHash{'pfad'};
			$pathPrefix =~ s/^\///;

			$rePathPrefix = qr/^\/\Q$pathPrefix\E\// if $pathPrefix;

#
# check if there is ftp-account for this domain
#
			if ( $pathPrefix ) {
				$sql="SELECT * FROM ftp WHERE kunde='$customer' AND pfad='$pathPrefix'";
				if ($wrapDbh->{'EXECUTE'}->($sql)){
					my $ptrFtp = $wrapDbh->{'FETCHHASH'}->();
					%attr=('NAME'=>$ptrFtp->{'account'},
								 'PASSWORD'=>$ptrFtp->{'longpw'},
								);
				}
				$wrapDbh->{'FINISH'}->();
			}
			unless(%attr){
				%attr = ( 'NAME' => $domainHash{'kunde'},
									'PASSWORD' => $custHash{'longpw'},
							);
				if( $custHash{'maxkb'} > 0 ) {
					$attr{'QUOTA'} = $custHash{'maxkb'} * 1024; ## translate to bytes
				}
				if( $custHash{'shell'} ) {
					$attr{'SHELL'} = $stdShell || '/bin/sh';
				}
			}
			$item = &makeSysuserNode(\%attr);
			$hostingNode->{'ADDCHILD'}->($item);
#
# end sysuser
#

#
# scripting
#
			$item = &makeScriptingNode( \%mapCustScripting, \%custHash );
			$hostingNode->{'ADDCHILD'}->($item);
#
# end scripting
#

#
# FronPage user
#
			if($domainHash{'frontpage'}){
				$sql = "SELECT benutzer,passwort FROM frontpage WHERE domain='$domain'";
				if($wrapDbh->{'EXECUTE'}->($sql)){
					if($ptrRow = $wrapDbh->{'FETCHROW'}->()){
						$item = &makeXmlNode('fpuser');
						$hostingNode->{'ADDCHILD'}->($item);
						$item->{'ATTRIBUTE'}->('name',$ptrRow->[0]);
						$item->{'ADDCHILD'}->(&makePasswordNode($ptrRow->[1],'plain'));
					}
				}
				$wrapDbh->{'FINISH'}->();
			}
#
# end FronPage user
#

#
# .htpasswd
#

			if($custHash{'pwschutz'}){
				my (%pdirs,$path,$parentID,$pdirNode,$ptrArr,$title);

				$sql="SELECT ident,pfad,bereich FROM pwschutz WHERE kunde='$customer'";
				if ( $wrapDbh->{'EXECUTE'}->($sql) ) {
					while ( $ptrRow = $wrapDbh->{'FETCHROW'}->() ) {
						$path = $ptrRow->[1];
						if($pathPrefix){
							unless( $path =~ s/$rePathPrefix/\//){
								next;   ## not subdir
							}
						}
						$pdirs{$path} = [ $ptrRow->[0], $ptrRow->[2] ];
					}
					$wrapDbh->{'FINISH'}->();
				}

				while( ($path,$ptrArr) = each(%pdirs) ) {
					( $parentID, $title ) = @{$ptrArr};
					if($pathPrefix){
						unless ( $path =~ s/$rePathPrefix// ) {
							next;  ## not subdir
						}
					}
					$pdirNode = &makeXmlNode('pdir');
					$pdirNode->{'ATTRIBUTE'}->('name',$path);
					$pdirNode->{'ATTRIBUTE'}->('title',$title);
					$hostingNode->{'ADDCHILD'}->($pdirNode);
					$sql="SELECT login,longpw FROM users WHERE parent=$parentID";
					if($wrapDbh->{'EXECUTE'}->($sql)){
						while( $ptrRow = $wrapDbh->{'FETCHROW'}->() ) {
							$item = &makeXmlNode('pduser');
							$item->{'ATTRIBUTE'}->('name',$ptrRow->[0]);
							$item->{'ADDCHILD'}->(&makePasswordNode($ptrRow->[1]));
							$pdirNode->{'ADDCHILD'}->($item);
						}
					}
					$path =~ s/^\///;
					$path .= '/' if $path;
					push @excludeFiles,($path.'.htaccess',$path.'.htpasswd');

					$wrapDbh->{'FINISH'}->();
				}
			}
#
# end .htaccess
#
		}

		my ($ptrAccounts);
		($ptrSubdomains,$ptrAccounts) = &getSubdomains($domain,\%domainHash);


		unless ( $nocontent ) {

#
# translate ftp-accounts to webuser of main domain
#
			if($domainHash{'richtigedomain'} && ref($ptrAccounts)=~/HASH/){
				my ($account,$path);
				$sql="SELECT account,longpw,pfad FROM ftp WHERE kunde='$customer'";
				if($wrapDbh->{'EXECUTE'}->($sql)){
					my ($webuserNode,$ftpDump);
					while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
						$account = $ptrRow->[0];
						$path = $ptrRow->[2];
						if($pathPrefix){
							unless($path =~ s/$rePathPrefix//){
								next;   ## not subdir
							}
						}
						next if exists $ptrAccounts->{$account}; ## already there is a domain with the ftp-account
						$webuserNode = &makeXmlNode('webuser');
#
# removed
#
#	$webuserNode->{'ATTRIBUTE'}->('name',$path);
#
# end removed. Plesk does not support this attribute
#
						$item = &makeSysuserNode($account,$ptrRow->[1]);
						$webuserNode->{'ADDCHILD'}->($item);

						$ftpDump = &makeDumpFile("$dumpDir/$account","$user_homeDir/$customer/html/".$ptrRow->[2],'.');

						$webuserNode->{'ATTRIBUTE'}->('cid',$ftpDump);
						$hostingNode->{'ADDCHILD'}->($webuserNode);

						$item = &makeScriptingNode(\%mapCustScripting,\%custHash);
						$webuserNode->{'ADDCHILD'}->($item);
					}
				}
				$wrapDbh->{'FINISH'}->();
			}
#
# end translate ftp-accountys to webuser
#
		}

#
# add subdomains
#
		my ($reSimpleSubDomain,$name);
		$reSimpleSubDomain = qr/^([^.]+)\.\Q$domain\E$/; ## reg expression for the first level subdomain

    while (my($subdomain,$ptrAttrHash) = each(%{$ptrSubdomains})){

      if ( $subdomain =~ /$reSimpleSubDomain/ ) {
				$name = $1;
				unless ( $nocontent ) {
					unless ( $name =~ /\*/ ) {  ## skip wildcard-domain

						$item = &makeSubdomainNode( $name, $domain, $ptrAttrHash, \%custHash );

						if ( ref($item) =~ /HASH/ ) {
							$hostingNode->{'ADDCHILD'}->($item);
						}
					}
				}

      } else {
				delete ( $ptrSubdomains->{$subdomain} );
			}
    }
#
#  end add subdomains
#


		unless ( $nocontent ) {
#
# attributes of physical hosting
#
			if($pathPrefix){
				my $rePrefix = qr/^\Q$pathPrefix\E\//;
				my @excludeEntry = grep{/$rePrefix/} @excludeFiles; ## select files only from $pathPrefix
				$dumpFile = &makeDumpFile("$dumpDir/$domain.docroot",
																	"$user_homeDir/$customer/html/$pathPrefix",
																	['.'],
																	\@excludeEntry
																 );
				$hostingNode->{'ATTRIBUTE'}->($domainHash{'cssl'}?'cid_docroot_ssl':'cid_docroot',$dumpFile) if $dumpFile;

			}else{
				push @excludeFiles,'cgi-bin';
				$dumpFile = &makeDumpFile("$dumpDir/$domain.docroot",
																	"$user_homeDir/$customer/html",
																	['.'],
																	\@excludeFiles
																 );
				$hostingNode->{'ATTRIBUTE'}->($domainHash{'cssl'}?'cid_docroot_ssl':'cid_docroot',$dumpFile) if $dumpFile;

				$dumpFile = &makeDumpFile("$dumpDir/$domain.cgi","$user_homeDir/$customer/html/cgi-bin");
				$hostingNode->{'ATTRIBUTE'}->('cid_cgi',$dumpFile) if $dumpFile;
			}

			$hostingNode->{'ATTRIBUTE'}->('fp',$domainHash{'frontpage'}?'true':'false');
			$hostingNode->{'ATTRIBUTE'}->('webstat',$custHash{'statistik'}?'true':'false');
			$hostingNode->{'ATTRIBUTE'}->('errdocs',$custHash{'fehlerseiten'}?'true':'false');
#
# end attributes of physical hosting
#
		} ## no content

#
# translate subdomain to domain
#
		my ($list);

		$ptrSubdomains = [keys %{$ptrSubdomains}];  ## make anonymous array. need for return
    $list=join(',',map {"'$_'"} @{$ptrSubdomains}); ## list of prepared subdomains

    $sql="SELECT domain FROM domains WHERE (richtigedomain=0) AND (kunde='$customer')".
      " AND (domain like '\%.$domain')";
    if($list){
      $sql.=" AND NOT (domain in ($list))";
    }
    if($wrapDbh->{'EXECUTE'}->($sql)){
			my ($subdomain,$name);
      while($ptrRow = $wrapDbh->{'FETCHROW'}->()){
				$subdomain = $ptrRow->[0];
				if ( $subdomain =~ /$reSimpleSubDomain/ ){  ## subdomain of first level
					$name = $1;
					unless ( $name =~ /\*/ ) { ## exclude wildcard-subdomains
						push @{$ptrDomains}, $subdomain;
					}
				}
      }
    }
    $wrapDbh->{'FINISH'}->();

#
# end translate subdomain to domain
#

  }elsif( ( $domainHash{'richtigedomain'} == 0 ) &&
		 !( $domainHash{'pfad'} =~ /^\// ) ){
#
# redirect
#
		unless ( $nocontent ) {
			$hostingNode = &makeXmlNode('shosting',$domainHash{'pfad'});
		}
  }

	unless ( $nocontent ) {
		if(ref($hostingNode)=~/HASH/){
			$rootDomain->{'ADDCHILD'}->($hostingNode);
		}
	} ## no content
#
# end hosting
#

#
# return list of subdoamin where are translated to domain & list of prepared subdomains
#
  return ($ptrDomains,$ptrSubdomains);

}

#
# end addDomainNode
#

sub makeDomainUserNode {

	my ($custHash) = @_;

	my $root = &makeXmlNode('domainuser');

	if ( ref($custHash) =~ /HASH/ ) {

		$root->{'ATTRIBUTE'}->( 'contact', $custHash->{'name'});
		$root->{'ATTRIBUTE'}->( 'email', $custHash->{'emailadresse'});



		$root->{'ATTRIBUTE'}->('status',$custHash->{'gesperrt'}?'false':'true');
		#  $root->{'ATTRIBUTE'}->('account',$custHash->{'kunde'});
		$root->{'ADDCHILD'}->(&makePasswordNode($custHash->{'longpw'}));

		&addPinfo( $root, $custHash, \%mapUserPinfo);

		unless ( exists ( $mapCustPermis{'make_dumps'} ) ) {
			$mapCustPermis{'make_dumps'} = $backup_off? 0: 1;
		}

		&addPermissions( $root, $custHash, \%mapCustPermis );

	}

	return $root;
}

#
# dns records
#
sub addDNS {
	my ($root, $domain, @ns) = @_;

	my ($sql, $ptrRow, $cnt);

	return undef unless exists $admin{'dns'}; ## Premium 1.0.* has not dns

	return -1 unless $admin{'dns'};

	$cnt = 0;

	$sql = "SELECT zonefile FROM dns WHERE domain='$domain'";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
	  my ($line, @fields, $type, $i, $dst, $src, $opt, $soa);
	  $ptrRow = $wrapDbh->{'FETCHROW'}->();

	  foreach $line (split(/\n/, $ptrRow->[0])) {
			next unless $line;
			$cnt++;
			@fields = split(' ', $line);
			if (@fields > 2) {
				$src = $fields[0];
				$dst = undef;
				$opt = undef;
				for($i = 3; $i > 0; $i--) {
					$type = $fields[$i];
					if($type =~ /^(A|NS|CNAME|PTR)$/i) {
						$dst = $fields[$i+1];
						last;
					} elsif ($type =~ /^MX$/i) {
						$dst = $fields[$i+2];
						$opt = $fields[$i+1];
						last;
					}
				}
				
				if ($dst) {

					if ( $type eq 'NS' ) {
						if ( $dst eq '.' ) {
							$dst = shift( @ns);
						}
			# Some Confixxes are buggy and have IP followed by the
			# dot in the NS records
						if ( $dst =~ /$(\d+\.\d+\.\d+\.\d+)\.?^/ ) {  ## is an IP
							$dst =~ $1; #strip dot from the end of IP
							makeDnsRecord($root, 'A', "ns.$src", $dst);
							$dst = "ns.$src";
						}
						if ( $dst ) {
							makeDnsRecord($root, 'NS', $src, $dst );
						}

					} else { ## all other types
						makeDnsRecord($root, $type, $src, $dst, $opt);
					}

				}
			}
		}
	}

	$wrapDbh->{'FINISH'}->();
	return $cnt;
}

sub makeDnsRecord {
  my ($root, $type, $src, $dst, $opt) = @_;

  my $item = makeXmlNode('dnsrec');
  $item->{'ATTRIBUTE'}->('type', $type);
  $item->{'ATTRIBUTE'}->('src', $src);
  $item->{'ATTRIBUTE'}->('dst', $dst);
  $item->{'ATTRIBUTE'}->('opt', $opt) if ($opt);
  $root->{'ADDCHILD'}->($item);
}

#
# end dns records
#




#
# maillist
#

sub addMaillist {
	my ($root,$domain,$ptrDomainHash) = @_;
	unless(ref($root)=~/HASH/){
		&printToError("Error: sub addMaillist: wrong paraneters");
		return;
	}
	if ( $majorVersion < 3 ){
		return 0;
	}

	my ($sql,$ptrHash,$domainID,$ret,@ids,$maillistID,$ptrRow);

	$ret = 0;

	unless ( ref ($ptrDomainHash) =~ /HASH/ ){
		$sql="SELECT * FROM domains WHERE domain='$domain'";
		if ($wrapDbh->{'EXECUTE'}->($sql)){
			if($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
				$ptrDomainHash = { %{$ptrHash} };
			}
		}
		$wrapDbh->{'FINISH'}->();
	}

	unless ( $domainID = $ptrDomainHash->{'id'} ) {
		&printToError("Error: sub addMaillist: domain id is not found");
		return;
	}

	$sql = "SELECT id FROM maillist WHERE domain_id=$domainID ORDER BY id";
	if ($wrapDbh->{'EXECUTE'}->($sql)){
		while ( $ptrRow = $wrapDbh->{'FETCHROW'}->() ){
			push @ids,$ptrRow->[0];
		}
	}
	$wrapDbh->{'FINISH'}->();

	my ($ptrMailList,$rootMaillist,$recipientsFile,$item,$maillistName);

	foreach $maillistID (@ids) {
		$sql = "SELECT * FROM maillist WHERE id=$maillistID";
		if ($wrapDbh->{'EXECUTE'}->($sql)){
			$ptrMailList = $wrapDbh->{'FETCHHASH'}->();

			$rootMaillist = &makeXmlNode('maillist');
			$root->{'ADDCHILD'}->($rootMaillist);

			$maillistName = $ptrMailList->{'name'};
			$rootMaillist->{'ATTRIBUTE'}->('name',$maillistName);
			if ( $ptrMailList->{'resperrt'} ) {
				$rootMaillist->{'ATTRIBUTE'}->('state','off');
			}

			$item = &makeXmlNode('owner',$ptrMailList->{'owner_mail'});
			$rootMaillist->{'ADDCHILD'}->($item);

			if ( $ptrMailList->{'pwd'} ) {
				$item = &makePasswordNode($ptrMailList->{'pwd'},'plain');
				$rootMaillist->{'ADDCHILD'}->($item);
			}

			if ( $majordomo_ldir && ( -d $majordomo_ldir ) ) {
				$recipientsFile = "$majordomo_ldir/$domainID/lists/$maillistName";
				if ( -T $recipientsFile ){
					if ( open(LIST,"<$recipientsFile") ) {
						while ( <LIST> ) {
							chomp;
							next unless $_;
							$item = &makeXmlNode('recipient',$_);
							$rootMaillist->{'ADDCHILD'}->($item);
						}
						close(LIST);
					} else {
						&printToError("Error: sub addMaillist: unable to open '$recipientsFile' for read: $!\n");
					}
				}
			}
			$wrapDbh->{'FINISH'}->();
		}
	}

	return $ret;
}

#
# end maillist
#

#
# database
#

sub addUsersDatabase {

	my ($root,$customer) = @_;

	unless(ref($root)=~/HASH/){
		&printToError("Error: sub addUsersDatabase: wrong paraneters");
		return;
	}

	my ( $itemDbUser, $nameUser );
	my $ret = 0;

  unless(exists ($preparedCustomers{$customer})){
    $preparedCustomers{$customer}={};
  }

  unless(exists ($preparedCustomers{$customer}->{'db'})){
    my ($nameDb,$nodeDb,$sql);

    $sql = "SELECT dbname FROM mysql_datenbanken WHERE kunde='$customer'";

    if ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {

      my ($ptrRow,$nodeDB,$dumpDb,$password,$nodeUser);

      while($ptrRow = $wrapDbh->{'FETCHROW'}->()){
				$nameDb = $ptrRow->[0];

				&printToLog("Dump user's mysql '$nameDb'");

				$nodeDb = &makeXmlNode('database');
				$nodeDb->{'ATTRIBUTE'}->('name',$nameDb);
				$nodeDb->{'ATTRIBUTE'}->('type','mysql');

				$dumpDb = &makeDumpDb($nameDb,'mysql',
															$mysqlUserUser,$mysqlUserPw,$mysqlUserServer);

				$nodeDb->{'ATTRIBUTE'}->('cid',$dumpDb);

				$root->{'ADDCHILD'}->($nodeDb);

				$itemDbUser = &addDbUserConfixx( $nodeDb,$customer );

				&checkUniqueDbUser( $customer,$itemDbUser,$nameDb );

				$ret++;
      }
    }
    $wrapDbh->{'FINISH'}->();
    $preparedCustomers{$customer}->{'db'} = 1;
  }
#
# Premium 1.* - webShop
#
	if ( exists( $admin{'shops'}) && $admin{'shops'} ) {
		unless( exists ( $preparedCustomers{$customer}->{'shop'} ) ) {
			my ($nameDb,$nodeDb,$sql);

			$sql = "SELECT shop_name, password FROM shops WHERE kunde='$customer'";

			if ( $wrapDbh->{'EXECUTE'}->( $sql ) ) {

				my ($ptrRow,$nodeDB,$dumpDb,$password,$passwordNode,$nodeUser,$shopUser);

				while($ptrRow = $wrapDbh->{'FETCHROW'}->()){
					$shopUser = $nameDb = $ptrRow->[0];
					$password = $ptrRow->[1];
					$passwordNode = &makePasswordNode( $password,'plain' );


					&printToLog("Dump user's webShop '$nameDb'");

#
# shop_<customer_<N>
#
					$nodeDb = &makeXmlNode('database');
					$nodeDb->{'ATTRIBUTE'}->('name',$nameDb);
					$nodeDb->{'ATTRIBUTE'}->('type','mysql');

					$dumpDb = &makeDumpDb($nameDb,'mysql',
																$shopUser,$password,$mysqlUserServer);

					$nodeDb->{'ATTRIBUTE'}->('cid',$dumpDb);

					$root->{'ADDCHILD'}->( $nodeDb );

					$itemDbUser = &addDbUserConfixx( $nodeDb,$shopUser,$passwordNode );

					&checkUniqueDbUser( $shopUser,$itemDbUser,$nameDb );

					$ret++;

#
# design_<customer>_<N>
#
					$nameDb =~ s/^shop_/design_/;

					$nodeDb = &makeXmlNode('database');
					$nodeDb->{'ATTRIBUTE'}->('name',$nameDb);
					$nodeDb->{'ATTRIBUTE'}->('type','mysql');

					$dumpDb = &makeDumpDb($nameDb,'mysql',
																$shopUser,$password,$mysqlUserServer);

					$nodeDb->{'ATTRIBUTE'}->('cid',$dumpDb);

					$root->{'ADDCHILD'}->( $nodeDb );

					$itemDbUser = &addDbUserConfixx( $nodeDb,$shopUser,$passwordNode );

					&checkUniqueDbUser( $shopUser,$itemDbUser,$nameDb );


					$ret++;
				}
      }

			$wrapDbh->{'FINISH'}->();

			$preparedCustomers{$customer}->{'shop'} = 1;
		}
	}
#
# end webShop
#
	return $ret;
}

my %dbUsers;

sub checkUniqueDbUser {
	my ($nameUser,$itemDbUser,$nameDb) = @_;

	my ( $i );

	if ( exists( $dbUsers{$nameUser} ) ) {
#
# make unique name
#
		$i = $dbUsers{$nameUser}+1;
		my $newName = $nameUser.'_'.$i;
		while ( exists( $dbUsers{$newName} ) ) {
			$i++;
			$newName = $nameUser.'_'.$i;
		}
		$nameUser = $newName;

		if ( ref($itemDbUser) =~ /HASH/ ) {
			$itemDbUser->{'ATTRIBUTE'}->( 'name', $nameUser );
		}

	}

	$dbUsers{$nameUser} = $i;

	return $nameUser;

}


#
# end databse
#



#
# make list of subdomains
#
sub getSubdomains {
	my($domain,$ptrDomainHash) = @_;

	my ($subdomain,$ptrAttrHash,$name,$reSimpleSubDomain,$sql,$subsubdomainLevel);
	my ($pathPrefix,$rePathPrefix,$ptrSubdomains,$ptrAccounts,$customer,@tmp);
	my ($account,$path,@subsubdomains,$level,$ptrRow,$subdomainLevel);

	$pathPrefix = $ptrDomainHash->{'pfad'};
	$pathPrefix =~ s/^\///;

	$rePathPrefix = qr/^\/\Q$pathPrefix\E\// if $pathPrefix;

	$reSimpleSubDomain = qr/^([^.]+)\.\Q$domain\E$/; ## reg expression for the first level subdomain

	$subdomainLevel = @tmp = ($domain =~ /\./g); ## calculate level of domain
	$subdomainLevel++;
	$subsubdomainLevel = $subdomainLevel + 1;

	$ptrSubdomains = {};
	$ptrAccounts = {};
	$customer = $ptrDomainHash->{'kunde'};

	$sql = "SELECT f.account,f.longpw,d.pfad,d.domain FROM domains d ".
		" LEFT JOIN ftp f ON (f.kunde=d.kunde) AND (f.pfad = SUBSTRING(d.pfad,2)) ".
			" WHERE (d.kunde='$customer') AND (d.domain like '\%.$domain') AND (d.richtigedomain = 0)";
#
# 'domains.pfad' begins with '/', 'ftp.pfad' does not
#

	if($wrapDbh->{'EXECUTE'}->($sql)){
		while ( $ptrRow = $wrapDbh->{'FETCHROW'}->() ) {
			$path = $ptrRow->[2];
			$path =~ s/^\///;  ## remove first slash

			if ( $pathPrefix ) {
				unless ( $path =~ /$rePathPrefix/ ) {
					next;   ## not subdir
				}
			}
			$subdomain = $ptrRow->[3];

			$level = @tmp = ($subdomain =~ /\./g); ## calculate level of subdomain
			if ( $level == $subsubdomainLevel ) {
				push @subsubdomains,$subdomain;
			}elsif ( ( $level == $subdomainLevel ) &&
							 ( $subdomain =~ /$reSimpleSubDomain/ ) ) {
				$ptrSubdomains->{$subdomain} = {
																				'NAME' => $1,
																				'PATH' => "$user_homeDir/$customer/html/$path",
																			 };
			}
			if ( $account = $ptrRow->[0] ) {
				$ptrAccounts->{$account} = 1; ## all ftp-accounts where matchs with subdomain
				if ( $level == $subdomainLevel ) {
					$ptrSubdomains->{$subdomain}->{'ACCOUNT'} = $account;
					$ptrSubdomains->{$subdomain}->{'PASSWORD'} = $ptrRow->[1];
					$ptrSubdomains->{$subdomain}->{'PATH'} = "$user_homeDir/$customer/html/$path";
				}
			}
		}
	}
	$wrapDbh->{'FINISH'}->();

#
# translate subdomain having subdomains to domain
#
	my ($reSubDomain,$subsubdomain);
	foreach $subdomain (keys %{$ptrSubdomains}) {
		$reSubDomain = qr/\.\Q$subdomain\E$/;
		foreach $subsubdomain ( @subsubdomains ) {
			if ( $subsubdomain =~ /$reSubDomain/ ) { ## there is subdomain
				delete $ptrSubdomains->{$subdomain};
				last;
			}
		}
	}

#
# translate subdomain having mail to domain
#
	my $list = join(',',map {"'$_'"} keys %{$ptrSubdomains});
	if ( $list ) {
		$sql = "SELECT domain FROM email WHERE domain in ($list)";
		if ( $wrapDbh->{'EXECUTE'}->($sql) ) {
			my ($emailDomain);
			while ($ptrRow=$wrapDbh->{'FETCHROW'}->()){
				$emailDomain = $ptrRow->[0];
				delete ($ptrSubdomains->{$emailDomain});  ## translate subdomain to domain if there is an email
			}
		}
		$wrapDbh->{'FINISH'}->();
	}

#
# translate subdomain with redirect to domain
#
	$list = join(',',map {"'$_'"} keys %{$ptrSubdomains});
	if($list){
		$sql = "SELECT domain, pfad FROM domains WHERE (richtigedomain=0) AND ".
			"NOT (pfad like '/\%') AND (domain in ($list))";
		my ($redirectDomain);
		if($wrapDbh->{'EXECUTE'}->($sql)){
			while ($ptrRow=$wrapDbh->{'FETCHROW'}->()){
				$redirectDomain = $ptrRow->[0];
				delete ($ptrSubdomains->{$redirectDomain}); ## translate subdomain to domain if it is a redirect
			}
		}
		$wrapDbh->{'FINISH'}->();
	}

	return ($ptrSubdomains,$ptrAccounts);

}

#
# end make list of subdomains
#


#
# mail system
#

sub makeMailSystem {

	my ($domain) = @_;
	my ($sql,$rootMail,$prefix,%box,%redir,@dest,$pop,$rootUser,
			$count,$i,$ptrPrefixes,$ptrHash,$ptrRow,$item,$redirect,$account,
			@prefixes,%group,$catch_all,%alias);

  $sql="SELECT * FROM email WHERE domain='$domain' ORDER BY prefix";
  if ( $wrapDbh->{'EXECUTE'}->($sql) ) {

    $rootMail = &makeXmlNode('mailsystem');

    while ( $ptrHash = $wrapDbh->{'FETCHHASH'}->() ) {
      $prefix = $ptrHash->{'prefix'};

#
# exclude catch-all from mailusers
#
			unless ( $prefix =~ /\*/ ) {
				push @prefixes,$prefix;
			} else {
				$catch_all = $prefix;
			}

      @dest = ( $ptrHash->{'pop3'} );
      push @dest, split( /:/,$ptrHash->{'pop3x'} );
      foreach $pop ( @dest ) {

				if ($pop =~ /^[^@]+\d+p\d+$/ ) { ## <user_prefix><N>p<M>  - web2p1
					if ( exists($alias{$prefix} ) ) {
						push @{$redir{$prefix}}, $pop.'@'.$domain;
					} else {
						push @{$box{$pop}},$prefix; ## all emails for this mail box
						$alias{$prefix} = $pop; ## register alias
					}

				} elsif ( $pop =~ /\@/ ){
					push @{$redir{$prefix}}, $pop; ## prefix is unique in the domain

				}
      }
    }
    $wrapDbh->{'FINISH'}->();
	}

	foreach $account (sort keys %box) {

		next if ( $account =~ /\*/ ); ## skip catch_all

		$rootUser = &makeXmlNode('mailuser');

		$rootUser->{'ATTRIBUTE'}->('name',$account);
		$rootUser->{'ATTRIBUTE'}->('mailgroup-enabled','false');
		$rootMail->{'ADDCHILD'}->($rootUser);

		$sql="SELECT longpw FROM pop3 WHERE account='$account'";
		if ($wrapDbh->{'EXECUTE'}->($sql)){
			$ptrRow = $wrapDbh->{'FETCHROW'}->();
			$item = &makePasswordNode($ptrRow->[0]);
			$rootUser->{'ADDCHILD'}->($item);
		}
		$wrapDbh->{'FINISH'}->();

		if ( $majorVersion > 2 ){  ## for Confixx 3 and later
			$item = &makeXmlNode('mailuser-permission');
			$item->{'ATTRIBUTE'}->('name','multiple-sessions');
			$rootUser->{'ADDCHILD'}->($item);
		}

		$item = &makeMailboxNode($account);
		$rootUser->{'ADDCHILD'}->($item);

		foreach $prefix ( @{$box{$account}} ) {

			next if ( $prefix =~ /\*/ ); ## skip catch_all

			$item = &makeXmlNode( 'alias', $prefix );
			$rootUser->{'ADDCHILD'}->($item);
		}
	}

	foreach $prefix (sort keys %redir ) {

		next if ( $prefix =~ /\*/ ); ## skip catch_all

		foreach $redirect ( @{$redir{$prefix}} ) {
			next if ( $redirect =~ /\*/ ); ## skip catch_all
			push @{$group{$prefix}},split( /\s*,\s*/,$redirect );
		}

		if ( 1 == @{$group{$prefix}} ) { ## only one redirect - alias
			($redirect) = @{$group{$prefix}};

		} else { ## more thet one redirect - mailgroup
			$redirect = undef;
		}

		$rootUser = &makeXmlNode('mailuser');

		$rootUser->{'ATTRIBUTE'}->('name',$prefix);
		$rootUser->{'ATTRIBUTE'}->('mailgroup-enabled', $redirect?'false':'true');
		$rootMail->{'ADDCHILD'}->($rootUser);

		if ($redirect) {
			$item = &makeXmlNode('redirect',$redirect);
			$rootUser->{'ADDCHILD'}->($item);

		} else {
			foreach $pop ( @{$group{$prefix}} ) {
				$item = &makeXmlNode('mailgroup-member',$pop);
				$rootUser->{'ADDCHILD'}->($item);
			}
		}

	}

	if ( $catch_all ) {
		if ( ref($redir{$catch_all}) =~ /ARRAY/ ) {
			foreach my $rdr ( @{$redir{$catch_all}} ) {
				if ( $rdr ) {
					$redirect = $rdr;
					last;
				}
			}
		} else {
			$redirect = $redir{$catch_all};
		}
		unless( $redirect ) {
			if ( $redirect = $box{$catch_all} ) {
				$redirect .= '@'.$domain;
			}
		}
		if ( $redirect ) {
			$item = &makeXmlNode( 'catch-all',$redirect );
			$rootMail->{'ADDCHILD'}->($item);
		}
	}

	return $rootMail;
}
#
# end mail system
#



sub addDbUserConfixx {
  my ($rootNode, $customer, $passwordNode) = @_;

  unless((ref($wrapUserMysql)=~/HASH/) &&
	 (ref($wrapUserMysql->{'EXECUTE'})=~/CODE/)){

    $wrapUserMysql = getDbConnect(\%bin,'mysql',$mysqlUserDbAddress,
								  $mysqlUserUser, $mysqlUserPw, 'mysql',
								  $mysqlUserServer);

  }

  unless((ref($wrapUserMysql)=~/HASH/) &&
		 (ref($wrapUserMysql->{'EXECUTE'})=~/CODE/)){

    &printToError("Error: unable to connect to 'mysql' database\n",
				  "host: $mysqlUserServer, user: $mysqlUserUser, pass: $mysqlUserUser");
    return;
  }
  unless($wrapUserMysql->{'PING'}->()){
    unless($wrapUserMysql->{'RECONNECT'}->()){
      &printToError("Error: unable to connect to 'mysql' database\n",
					"host: $mysqlUserServer, user: $mysqlUserUser, pass: $mysqlUserUser");
      return;
    }
  }

  my $userItem = &addDbUser($rootNode, $customer, $wrapUserMysql, $passwordNode);

	return $userItem;

}

#
# makeSubdomainNode  - create XML-node for subdomain
#
# argumets:
#      $name        - name of subdomain (without '.')
#      $domain      - domain
#      $ptrAttrHash - pointer to hash with attributes of subdomain
#      $ptrScriptingHash - pointer to hash with parameters of scripting
#
# return:
#      $subdomNode  - pointer to new XML-node
#

sub makeSubdomainNode {
  my( $name, $domain, $ptrAttrHash, $ptrScriptingHash ) = @_;
  my ($subdomNode,$item,$path,$dumpFile);
  unless( ref($ptrAttrHash) =~ /HASH/ ) {
    $ptrAttrHash = {};
  }
  unless( ref($ptrScriptingHash) =~ /HASH/ ) {
    $ptrScriptingHash = {};
  }

  $subdomNode = &makeXmlNode( 'subdomain' );

  $subdomNode->{'ATTRIBUTE'}->( 'name', $name );

  if ( $path = $ptrAttrHash->{'PATH'}){
    if($dumpFile = &makeDumpFile("$dumpDir/$name.$domain",$path,'.')){
      $subdomNode->{'ATTRIBUTE'}->( 'cid_docroot', $dumpFile );
    }
  }

	if ( keys %{$ptrAttrHash} ) {
		$item = &makeSysuserNode( $ptrAttrHash );
		$subdomNode->{'ADDCHILD'}->( $item );
	}

  $item = &makeScriptingNode(\%mapCustScripting,$ptrScriptingHash);
  $subdomNode->{'ADDCHILD'}->( $item );

  return $subdomNode;
}

#
# create system user node
# check if there was already this account
# & add a suffix to make an unique account
#
my %sysUsers;

sub makeSysuserNode {

  my ($account,$password,$pswdType,$quota,$shell)=@_;

  my ($ptrHash,$item);

  if ( ref($account) =~ /HASH/ ) {
    $ptrHash = $account;
    $account = $ptrHash->{'ACCOUNT'}||$ptrHash->{'NAME'};
    $password = $ptrHash->{'PASSWORD'};
    $pswdType = $ptrHash->{'PSWDTYPE'};
    $shell = $ptrHash->{'SHELL'};
    $quota = $ptrHash->{'QUOTA'};
  }
	if ( $account ) {
	  if ( exists( $sysUsers{$account} ) ) {
			my $clon = 1;
			while(exists($sysUsers{$account})){
				if($account =~ /(.*)c(\d+)$/){
					$account = $1;
					$clon = $2;
					$clon++;
				}
				$account .= "c$clon";
			}
		}
		$sysUsers{$account} = 1;

		$item = &makeXmlNode('sysuser');
		$item->{'ATTRIBUTE'}->('name',$account);
		if($quota){
			$item->{'ATTRIBUTE'}->('quota',$quota);
		}
		if($shell){
			$item->{'ATTRIBUTE'}->('shell',$shell);
		}

		$item->{'ADDCHILD'}->( &makePasswordNode($password,$pswdType) );

	}

  return $item;
}

sub makeScriptingNode {
  my $ptrMapHash = shift;
  my $ptrValueHash = shift;

  my ($plesk,$confixx,$value);

  my $item = &makeXmlNode('scripting');
  if(ref($ptrValueHash)=~/HASH/ && ref($ptrMapHash)=~/HASH/){
    while (($plesk,$confixx)=each(%{$ptrMapHash})){
			if ( $confixx =~ /^\d+$/ ) {
				$value = $confixx;
			}	elsif( exists ( $ptrValueHash->{$confixx} ) ) {
				$value = $ptrValueHash->{$confixx} || 0;
			} else {
				$value = undef;
			}
			if ( defined ( $value ) ) {
				$item->{'ATTRIBUTE'}->($plesk,$value?'true':'false');
      }
    }
  }
  return $item;
}

sub addAutoresponders {
  my ($root,$domain,$ptrBoxes)=@_;
  my($list,$sql,$ptrHash,$item,$name,$value);
  $list = join(',',map {"'$_\@$domain'"} @{$ptrBoxes});
  if($list){
    $sql = "SELECT * FROM autoresponder WHERE absenderemail in ($list)";
    if ($wrapDbh->{'EXECUTE'}->($sql)){
      while ($ptrHash=$wrapDbh->{'FETCHHASH'}->()){
		$value = $wrapBase64->{'ENCODE'}->($ptrHash->{'emailtext'});
		$item = &makeXmlNode('autoresponder',$value);
		$root->{'ADDCHILD'}->($item);

		$name = (split(/\@/,$ptrHash->{'absenderemail'}))[0];
		$item->{'ATTRIBUTE'}->('name',$name);

#
# bug #33493
#
#	$item->{'ATTRIBUTE'}->('replyto',$ptrHash->{'absenderemail'});
#
#
		$value = $ptrHash->{'emailbetreff'};
		$value = $wrapBase64->{'ENCODE'}->($value);
		chomp $value;
		$item->{'ATTRIBUTE'}->('subject',$value);
      }
    }
    $wrapDbh->{'FINISH'}->();
  }
}

sub makeMailboxNode {
  my($account)=@_;

  my($type,$dir,@names,$tarFile,$item);

  if($maildrop eq "HOMEDIR/Mailbox"  ){
    $type='mbox';
    $dir = "$pop_homeDir/$account";
    @names=($mailBoxName);
  }elsif($maildrop eq "HOMEDIR/Maildir/"){
    $type='mdir';
    $dir = "$pop_homeDir/$account";
    @names=($mailBoxName);
  }else{
    $type='mbox';
    $dir = $mailSpool;
    if(opendir(SPOOL,$dir)){
      @names = grep {/^\Q$account\E/} readdir(SPOOL);
      closedir(SPOOL);
    }
  }

	if ( @names ) {
		$tarFile = &makeDumpFile("$dumpDir/$account.$type",$dir,\@names);
	}

  $item = &makeXmlNode('mailbox');
  $item->{'ATTRIBUTE'}->('type',$type);
  if($tarFile){
    $item->{'ATTRIBUTE'}->('cid',$tarFile);
  }
#  $item->{'ATTRIBUTE'}->('quota',$mboxQuota);
  return $item;
}


sub addPinfo {
  my ($root, $ptrValue, $ptrMap) = @_;
  my ($item, $plesk, $confixx, $value);

  while (($plesk,$confixx) = each %{$ptrMap}) {

		if ( $value = $ptrValue->{$confixx} )  {

			if ( $plesk eq 'locale' ) { ## map name of locale
				if ( exists $mapLocales{$value} ) {
					$value = $mapLocales{$value};
					next unless $value;
				}
			}

			$item = &makeXmlNode('pinfo', $value);
			$item->{'ATTRIBUTE'}->('name', $plesk);
			$root->{'ADDCHILD'}->($item);
		}
  }
}


sub addLimits {
  my ($root, $ptrValue, $ptrMap) = @_;
  my ($item, $plesk, $confixx, $value);
  while (($plesk,$confixx) = each %{$ptrMap}) {

		if (($value = $ptrValue->{$confixx}) > 0) {
			if ($plesk eq 'disk_space' || $plesk eq 'max_traffic') {
				$value *= 1024;
			}

			$item = &makeXmlNode('limit', $value);
			$item->{'ATTRIBUTE'}->('name', $plesk);
			$root->{'ADDCHILD'}->($item);
		}
  }
}

sub addPermissions {
  my ($root,$ptrValue,$ptrMap) = @_;
  my ($item,$plesk,$confixx,$value);
  while(($plesk,$confixx) = each %{$ptrMap}){

		if ( $confixx=~/^\d+$/ ) {
			$value = $confixx;
		}else {
			$value = $ptrValue->{$confixx};
		}

		if ( $value ) {
      $item = &makeXmlNode('permission');
      $item->{'ATTRIBUTE'}->('name',$plesk);
      $root->{'ADDCHILD'}->($item);
    }
  }
}

sub makeIpNode {
  my $nodeName = shift||'map';
  my $refIpHash = shift || {};
  my ($root,$item,$ptrRow,$id,$shared,$exclusive,$ip,$ipNode,$sql);
  my (%sharedIps,%exlusiveIps);

  $root = &makeXmlNode($nodeName);

  $shared = &makeXmlNode('shared');
  $exclusive = &makeXmlNode('exclusive');
  $root->{'ADDCHILD'}->($shared);
  $root->{'ADDCHILD'}->($exclusive);

#
# standard ip is shared
#
  $sql = "SELECT standardip FROM admin";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    ($ip)=@{$wrapDbh->{'FETCHROW'}->()};

    $id = 'ip_'.join('',unpack('n2',pack('C4', split(/\./,$ip))));

    $refIpHash->{$ip}=[$id,'',''];

    $ipNode = &makeXmlNode('ip');
    $ipNode->{'ATTRIBUTE'}->('id',$id);
    $item = &makeXmlNode('old',$ip);
    $ipNode->{'ADDCHILD'}->($item);
    $item = &makeXmlNode('new');
    $ipNode->{'ADDCHILD'}->($item);
    $shared->{'ADDCHILD'}->($ipNode);
  }
#
# end standard ip
#

  $wrapDbh->{'EXECUTE'}->("select ip,anbieter,kunde from ipadressen");
  while($ptrRow = $wrapDbh->{'FETCHROW'}->()){
    $ip = $ptrRow->[0];
# get ID
    $id = 'ip_'.join('',unpack('n2',pack('C4', split(/\./,$ip))));

    $refIpHash->{$ip}=[$id,$ptrRow->[1],$ptrRow->[2]];

    $ipNode = &makeXmlNode('ip');
    $ipNode->{'ATTRIBUTE'}->('id',$id);
    $item = &makeXmlNode('old',$ip);
    $ipNode->{'ADDCHILD'}->($item);
    $item = &makeXmlNode('new');
    $ipNode->{'ADDCHILD'}->($item);

# is ip exclusive or share ?
    if($ptrRow->[1] || $ptrRow->[2]){
      $exclusive->{'ADDCHILD'}->($ipNode);
    }else{
      $shared->{'ADDCHILD'}->($ipNode);
    }
  }
  $wrapDbh->{'FINISH'}->();

  return $root;
}

sub getConfixxVersion {
  my $util = "$installDir/confixx_updatescript.pl";
  my ($out, $majorVersion, $minorVersion, $pathLevel);

  unless (-x $util) {
		return;
  }

  $out = `$util VERSION`;
  chomp $out;
  if ($out =~ /^(?:premium\s+)?(\d+)\.(\d+)\.?(\d+)?/i) {
		$majorVersion = $1;
		$minorVersion = $2;
		$patchLevel = $3 || 0;
  } else {
		$out = `$util --version`;
		chomp $out;
		if ($out =~ /^(\d+)\.(\d+)\.?(\d+)?/) {
			$majorVersion = $1;
			$minorVersion = $2;
			$patchLevel = $3 || 0;
		}
  }

  if ( $majorVersion ) {
		if (	wantarray()	) {
		  return(	$majorVersion, $minorVersion, $patchLevel	);
		} else {
	  	return 'Confixx ' . $majorVersion;
		}
  }
}

sub printAgentStatus {
  my $dumpFile = shift;
  my $statusFile = shift;
  my $outFh = shift || *STDOUT;
  my $rootName = shift || 'agentstatus';

  my ($root,$util,$out,$item);

  $root = &makeXmlNode($rootName);

  $out = getConfixxVersion();
  unless ($out && defined $bin{'iconv'}) {
	if (defined $bin{'iconv'}) {
	  $item = &makeXmlNode('wrong-platform', '');
	} else {
	  $item = &makeXmlNode('wrong-platform', 'no iconv found on the source host');
	}

	$root->{'ADDCHILD'}->($item);
	&printXml($root,$outFh,$rootName.'.dtd');
	return;
  }

  $item = &makeXmlNode('dumpfile',$dumpFile);
  $root->{'ADDCHILD'}->($item);
  $item = &makeXmlNode('statusfile',$statusFile);
  $root->{'ADDCHILD'}->($item);

  &printXml($root,$outFh,$rootName.'.dtd');
}

sub loadMainConfig {
  my $pathToConf = shift||'';

  unless(-T $pathToConf){
		if ( exists( $ENV{'CONFIXX_ROOT'} ) && -d $ENV{'CONFIXX_ROOT'} ) {
			$pathToConf = $ENV{'CONFIXX_ROOT'}.'/confixx_main.conf';
		}
	}

  unless(-T $pathToConf){
    foreach my $prefix ('./','/root/confixx/','/usr/local/confixx/','/confixx/'){
      $pathToConf = $prefix.'confixx_main.conf';
      last if -T $pathToConf;
    }
  }

#
# get path to confixx from ctontab
#
  unless(-T $pathToConf){
    my(@out,$line,$run);

		@out = `$bin{'grep'} -hr -e confixx /etc/crontab /etc/cron.d`;

		foreach $line ( @out ){
			chomp $line;
			$run = (split(' ',$line))[6];
			$run=~s/\/[^\/]*$//;
      $pathToConf = $run.'/confixx_main.conf';
			last if -T $pathToConf;
		}
  }

  unless(-T $pathToConf){
		my (@out,$line,$run);

		@out = `$bin{'crontab'} -l`;

		foreach $line ( @out ) {
			chomp $line;
			next if ( $line =~ /^#/ ); ## comment
			$run = (split(' ',$line))[6];
			$run=~s/\/[^\/]*$//;
			$pathToConf = $run.'/confixx_main.conf';
			last if -T $pathToConf;
		}
	}

#
# end get path to confixx from crontab
#
  unless(-T $pathToConf){
    my (@out,$line,$dir);

		foreach $dir ('/root','/usr/local') {

    	@out=`$bin{'find'} $dir -type f -name confixx_main.conf`;

	    foreach $line ( @out ) {
		  chop $line;
    	  if ( -T $line ) {
					$pathToConf=$line;
					last;
				}
			}
			if ( -T $pathToConf ) {
				last;
			}
    }
  }

  if(-T $pathToConf){
    do $pathToConf;
  }else{
    $pathToConf='';
  }

  return $pathToConf;
}

sub makeEmptyDump {
  my($rootName)=@_;
  my $root=&makeXmlNode($rootName);

  my $mapRoot = &makeXmlNode('map');
  $root->{'ADDCHILD'}->($mapRoot);

  my $item = &makeXmlNode('shared');
  $mapRoot->{'ADDCHILD'}->($item);

  $item = &makeXmlNode('exclusive');
  $mapRoot->{'ADDCHILD'}->($item);

  return $root;
}


sub printHelp {

  print <<"HELP";

Agent of migration from 'Confix 2' to 'Plesk 7'

Usage:
  $0 <options>

Options:
  -c |--config=<path>        path to confixx_main.conf

  -s |--get-status           get status of the agent
  -dc|--dump-accounts=<list> a coma separated list of resellers to dump
  -dd|--dump-domains=<list>  a coma separated list of customers to dump
  -da|--dump-all             make a full dump

  -lc|--get-content-list     get list of content files
  -nc|--no-content           do not make content files
  -nz|--no-compress          do not compress content files

  -o |--output=<path>        redirect output to file

  -h |--help                 this help
  -d |--debug=<level>        level of debug messages
  -v |--version

HELP
}

#
# end confixx subroutines
#




