#!/usr/bin/perl

# use for dump of user's mysql databases
use vars qw($mysqlUserUser $mysqlUserServer $mysqlUserPw);

use vars qw( %bin
             $dumpDir
             $ptrArgs
             $pwd
             $pleskVersion
             $pleskMajorVersion
             $pleskMinorVersion
             $pleskMinorMinorVersion
             $workDir);

use strict;

my $VERSION='1.0.1';

#
# parameters for connection to database
#
my $dbUser = 'admin';
my $dbName = 'psa';
my $dbHost = 'localhost';
my $dbType = 'mysql';
#
# end parameters for connection to database
#

#
# maps to translate xmlName => fieldName
#

my %xmlPermissions = ('create_domains'=>1,
		      'create_domains'=>1,
		      'manage_phosting'=>1,
		      'manage_quota'=>1,
		      'manage_subdomains'=>1,
		      'change_limits'=>1,
		      'manage_dns'=>1,
		      'manage_log'=>1,
		      'manage_crontab'=>1,
		      'manage_anonftp'=>1,
		      'manage_webapps'=>1,
		      'manage_sh_access'=>1,
		      'manage_maillists'=>1,
		      'make_dumps'=>1,
		      'manage_drweb'=>1,
		      'site_builder'=>1,
		      'make_phosting'=>1,
		      'manage_quota'=>1,
			  'manage_not_chroot_shell'=>1,
			  'manage_ftp_password'=>1,
		     );

my %commonLimits = ('max_wu' => 1,
					'max_db' => 1,
					'max_pbox' => 1,
					'mbox_quota' => 1,
					'max_redir' => 1,
					'max_mg' => 1,
					'max_resp' => 1,
					'disk_space' => 1,
					'max_traffic' => 1
				   );

my %domainLimits25 = %commonLimits;
$domainLimits25{'max_pbox'} = 1;

my %domainLimits = %commonLimits;
$domainLimits{'max_box'} = 1;
$domainLimits{'expiration'} = 1;
$domainLimits{'max_subdom'} = 1;
$domainLimits{'max_maillists'} = 1;
$domainLimits{'max_webapps'} = 1;
$domainLimits{'ans_freq'} = 1;
$domainLimits{'ans_count'} = 1;

my %clientLimits25 = %domainLimits25;
$clientLimits25{'max_dom'} = 1;

my %clientLimits = %domainLimits;
$clientLimits{'max_dom'} = 1;

my %typeOfField = ('expiration'=>'timestamp');

my %clientsAttribute = ('name'=>'login',
						'email'=>'email',
						'contact'=>'pname',
					   );

my %clientsInfo = ('company'=>'cname',
				   'phone'=>'phone',
				   'fax'=>'fax',
				   'address'=>'address',
				   'city'=>'city',
				   'state'=>'state',
				   'zip'=>'pcode',
				   'country'=>'country',
				   #		  'locale'=>'locale',
				  );

my %domainUserAttribute = ('contact'=>'personalName',
						   'email'=>'email');

my %domainUserInfo = ('company'=>'companyName',
					  'phone'=>'phone',
					  'fax'=>'fax',
					  'address'=>'address',
					  'city'=>'city',
					  'state'=>'state',
					  'zip'=>'zip',
					  'country'=>'country',);

my %hostingAttribute = ('https'=>'ssl',
						'fp'=>'fp',
						'fpssl'=>'fp_ssl',
						'fpauth'=>'fp_enable',
						'webstat'=>'webstat',
						'at-domains'=>'at_domains',
					   );

my %hostingScripting = ('ssi'=>'ssi',
						'php'=>'php',
						'cgi'=>'cgi',
						'perl'=>'perl',
						'asp'=>'asp',
						'python'=>'python',
						'coldfusion'=>'coldfusion',
						'asp_dot_net'=>'asp_dot_net',
					   );

my %webUserScripting = ('ssi'=>'ssi',
						'php'=>'php',
						'cgi'=>'cgi',
						'perl'=>'perl',
						'asp'=>'asp',
						'python'=>'python',
						'asp_dot_net'=>'asp_dot_net',
					   );

my %subDomainScripting = ('ssi'=>'ssi',
						  'php'=>'php',
						  'cgi'=>'cgi',
						  'perl'=>'perl',
						  'asp'=>'asp',
						  'python'=>'python',
						  'coldfusion'=>'coldfusion',
						  'asp_dot_net'=>'asp_dot_net',
						 );

#
# end maps
#

#
# GLOBAL
#

my $defaultIpId;

#
# parse command line
#
my %arg_opts = ('--help|-h'=>'',
		'--config|-c'=>'s',
		'--debug|-d'=>'i?',
		'--get-status|-s'=>'',
		'--dump-accounts|-dc'=>'s',
		'--dump-domains|-dd'=>'s',
		'--dump-all|-da'=>'',
		'--get-content-list|-lc'=>'',
		'--no-content|-nc'=>'',
		'--no-compress|-nz'=>'',
		'--output|-o'=>'s',
		'--status-file|-sf'=>'s',
		'--version|-v'=>'',
	       );

%bin = ('tar'=>'tar','split'=>'split','cp'=>'cp','rm'=>'rm','echo'=>'echo',
	   'mysql'=>'mysql','mysqldump'=>'mysqldump',
	   'psql'=>'psql','pg_dump'=>'pg_dump','uptime'=>'uptime','uname'=>'uname',
	   'cat'=>'cat','pwd'=>'pwd','md5sum'=>'md5sum',
	   'grep'=>'grep','find'=>'find','gzip'=>'gzip','wc'=>'wc',
	   'locale'=>'locale', 'iconv'=>'iconv', 'python'=>'python');

my $returnCode = 0;


#==============================================================
#
# common subroutines
#

sub printToLog {
  my ($msg, $noeol) = @_;
  if ($msg) {
	print $msg;
  }
  unless ($noeol) {
	print "\n";
  }
}

sub printToError {
  my ($msg, $noeol, $errCode) = @_;
  if ($msg) {
	print STDERR $msg;
  }
  unless ($noeol) {
	print STDERR "\n";
  }
  if (defined($errCode)) {
    $returnCode = $errCode; ## translate an error code to the return code
  } else {
    $returnCode = 127;
  }
}

#BEGINNOTINRAQ2

sub makeFileReader
{
	my ($filename, $doNotFail) = @_;
	local (*INPUT);

	unless (-T $filename) {

		if (defined($doNotFail)) {
			&printToLog("Error: file '$filename' is not a text file");
		} else {
			&printToError("Error: file '$filename' is not a text file");
		}

		return undef;
	}

	unless (open(INPUT,"<$filename")) {

		binmode INPUT;
		if (defined($doNotFail)) {
			&printToLog("Error: unable to open file '$filename' to read");
		} else {
			&printToError("Error: unable to open file '$filename' to read");
		}

		return undef;
	}

	my $input = *INPUT;

	my $this = {
		'READ' => sub {
				my $a = <$input>;
				return $a;
			},

		'CLOSE' => sub { close($input); }

	};

	return $this;
}

sub makeArrayReader {
  my (@arr) = @_;
  my $size= 0;
  my $count = scalar(@arr);

  my $this={'READ'=>sub {
			  if($count == $size) {
				return undef;
			  }

			  return $arr[$count++];
			},
			'CLOSE'=>sub {
			}
		   };
  return $this;
}

sub makeParser {
  my $reader;
  my ($this,$ptrRows,$keySep,$valSep,$valKeySep,$comment,
      $typeRows,$maxIndex,$rowIndex);
  {
	$reader = shift;
	unless(defined($reader)) {
	  return undef;
	}

    my %args=(@_);
    $keySep = $args{'KEYSEPARATOR'};
    $valSep = $args{'VALUESEPARATOR'};
    $valKeySep = $args{'KEYVALUESEPARATOR'};
  }
  $this={'READER'=>sub {
		   if(@_){
			 $reader=shift;
		   }
		   return $reader;
		 },
		 'FILE'=>sub {
		   if(@_){
			 $reader = makeFileReader(@_);
		   }
		   return undef;
		 },
		 'KEYSEPARATOR'=>sub{
		   if(@_){
			 $keySep=shift;
		   }
		   return $keySep;
		 },
		 'VALUESEPARATOR'=>sub{
		   if(@_){
			 $valSep=shift;
		   }
		   return $valSep;
		 },
		 'KEYVALUESEPARATOR'=>sub{
		   if(@_){
			 $valKeySep=shift;
		   }
		   return $valKeySep;
		 },
		 'COMMENT'=>sub{
		   if(@_){
			 $comment=shift;
		   }
		   return $comment;
		 },
		 'PARSE'=>sub{
		   my %args=(@_);

		   $keySep=$args{'KEYSEPARATOR'} if exists $args{'KEYSEPARATOR'};
		   $valSep=$args{'VALUESEPARATOR'} if exists $args{'VALUESEPARATOR'};
		   $valKeySep=$args{'KEYVALUESEPARATOR'} if exists $args{'KEYVALUESEPARATOR'};
		   $comment=$args{'COMMENT'} if exists $args{'COMMENT'};

		   $ptrRows = &parseFile('reader'=>$reader,'keysep'=>$keySep,
								 'valsep'=>$valSep,'keyvalsep'=>$valKeySep,
								 'comment'=>$comment);
		   $typeRows = ref($ptrRows);
		   if($typeRows=~/ARRAY/){
			 $rowIndex=-1;
			 $maxIndex=scalar(@{$ptrRows})-1;
		   }
		   return $ptrRows;
		 },
		 'ROW'=>sub{
		   if($typeRows=~/HASH/){
			 my $key=shift;
			 return $ptrRows->{$key};
		   }elsif($typeRows=~/ARRAY/){
			 if($rowIndex<$maxIndex){
			   $rowIndex++;
			   return $ptrRows->[$rowIndex];
			 }else{
			   return undef;
			 }
		   }else{
			 return undef;
		   }
		 },
		};

  return $this;
}

sub makeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename), @_);
}

sub makeSafeFileParser {
  my $filename = shift;
  return makeParser(makeFileReader($filename, 'doNotFail'), @_);
}

sub makeStringParser {
  my $content = shift;
  return makeParser(makeArrayReader($content), @_);
}

sub parseFile {
  my %arg = (@_);
  my $reader = $arg{'reader'};
  my $keySep = $arg{'keysep'};
  my $valSep = $arg{'valsep'};
  my $valKeySep = $arg{'keyvalsep'};
  my $comment = $arg{'comment'};

  my ($ret,$key,$value,$reKey,$reValue,@values,$keyBlank,$valBlank,
     $vkey,$vvalue,$reKeyValue,$reComment);

  if($keySep){
    $ret = {};
    if(ref($keySep)=~/regexp/i || ($keySep eq ' ')){
      $reKey = $keySep;
    }else{
      $reKey = qr/$keySep/;
    }
  }else{
    $ret = [];
  }
  if($valSep){
    if(ref($valSep)=~/regexp/i || ($valSep eq ' ')){
      $reValue = $valSep;
    }else{
      $reValue = qr/$valSep/;
    }
  }
  if($valKeySep){
    if(ref($valKeySep)=~/regexp/i || ($valKeySep eq ' ')){
      $reKeyValue = $valKeySep;
    }else{
      $reKeyValue = qr/$valKeySep/;
    }
  }
  if($comment){
    if(ref($comment)=~/regexp/i){
      $reComment = $comment;
    }else{
      $reComment = qr/$comment/;
    }
  }

  while($_ = $reader->{'READ'}->()){
    chomp;
    next unless $_;
    if($comment && /$reComment/){  ## commentar
	  next;
    }
    if ($keySep){
      ($key,$value)=split($reKey,$_,2);
      if($key){
		$key=~s/^\s+//;
		$key=~s/\s+$//;
		if($valSep){

		  if ($valKeySep){
			$ret->{$key}={};
			foreach $value (split($reValue,$value)){
			  ($vkey,$vvalue)=split($reKeyValue,$value);
			  $vkey=~s/^\s+//;
			  $vkey=~s/\s+$//;
			  $ret->{$key}->{$value}=$vvalue;
			}
		  }else{
			push @{$ret->{$key}},split($reValue,$value);
		  }
		}else{
		  $value=~s/^\s+//;
		  $value=~s/\s+$//;
		  $ret->{$key}=$value;
		}
      }
    }else{
      if($valSep){
		push @{$ret},[split($valSep,$_)];
      }else{
		push @{$ret},$_;
      }
    }
  }

  $reader->{'CLOSE'}->();

  return $ret;
}

#ENDNOTINRAQ2

sub makePasswordNode {
  my $value= shift;
  my $type = shift||'encrypted';

  my $item = &makeXmlNode('password');
  $item->{'CONTENT'}->($value);  ## html encode
  $item->{'ATTRIBUTE'}->('type',$type);
  return $item;
}

sub mangleName {
  my $name = shift;

  if($name =~ /^\d/) {
    $name = 'user' . $name;
  }
  $name =~ s/\./_/g;

  return $name;
}

sub makeSysUserNode {
  my ($name, $password, $type) = @_;

  unless(defined($name) && defined($password) && defined($type)) {
	printToError("makeSysUserNode: Username, password or password type is not set.");
  }

  my $sysUserNode = makeXmlNode('sysuser');
  $sysUserNode->{'ATTRIBUTE'}->('name', mangleName($name));
  $sysUserNode->{'ADDCHILD'}->(makePasswordNode($password, $type));

  return $sysUserNode;
}

# Returns home directory for the system account $user
sub getHomeDir {
	my ($user) = @_;
	my $home;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$home = $pw[7];
	}
	unless ($home) {
		# how it is by default
		$home = '/home/$account';
	}
	return $home;
}

# Returns password for system account $user
sub getSystemPassword {
	my ($user) = @_;
	my $password;
	my @pw = getpwnam($user);
	if (-1 != $#pw) {
		$password = $pw[1];
	}
	unless ($password) {
		$password = '';
	}
	return $password;
}

sub createUserNameFromDomain {
  my ($name) = @_;

  if ($name =~ /^\d/) {
	$name = 'u' . $name;
  }

  $name =~ s/\./_/g;
  $name =~ s/^www\.//g;
  $name =~ s/[aoeui]//g if length($name) > 16;
  $name = substr($name, 0, 16) if length($name) > 16;

  return $name;
}

my $compress = 1;

sub setCompress {
  ($compress) = @_;
}

sub makeDumpDb
{
  my ($dbName, $dbType, $dbUser, $dbPasswd, $dbHost, $ptrDomainBin, $variables, $domainSocket) = @_;
  $dbType ||= 'mysql';

  my ($dumpFile,$dumpSql,$dumpName,$cmd,$gzip,$split,$mysql,$psql, $retCode,$out);

  $dumpName = "$dbName.$dbType.sql";
  $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  if (-x $gzip) {
	$dumpName .= ".gz";
  }
  $dumpFile="$dumpDir/$dumpName";

  printToLog(">>Checking database presence");

  if ($dbType eq 'mysql') {
	$mysql = $bin{'mysql'} || 'mysql';
	if (-x $mysql) {
	  $cmd  = "$mysql -u '$dbUser'";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -p'$dbPasswd'" if $dbPasswd;
	  $cmd .= " -e '' '$dbName' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: mysql datatbase '$dbName' not found. mysql has returned '$retCode'");
		return "";
	  }
	}
  } elsif ($dbType eq 'Pg') {
	$psql = $bin{'psql'} || 'psql';
	if (-x $psql) {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if $dbPasswd;
	  $cmd .= "; $psql '$dbName' ";
	  $cmd .= " -h $dbHost" if $dbHost;
	  $cmd .= " -c '' 2>&1";

	  $out = `$cmd`;
	  chomp $out;
	  if ($out) {
		$retCode = $?>>8;
		printToLog("warning: PostgreSQL datatbase '$dbName' not found. psql has returned '$retCode'");
		return "";
	  }
	}
  }

  printToLog("..done");

  unless (exists($ptrArgs->{'no-content'})) {
	$gzip = $bin{'gzip'} || 'gzip';
	$split = $bin{'split'} || 'split';

	if ($dbType eq 'mysql') {
	  my ($prefx, $postfix) = ("", "");
	  my $commonprefix = "LD_LIBRARY_PATH=$variables " if $variables;

	  my $commonopts;
	  $commonopts = " --socket='$domainSocket'" if $domainSocket;
	  $commonopts .= " -h $dbHost" if $dbHost;
	  $commonopts .= " -p'$dbPasswd'" if $dbPasswd;
	  $commonopts .= " -u '$dbUser' '$dbName'";

	  my $dumpcmd;
	  if ($ptrDomainBin) {
		$dumpcmd = $ptrDomainBin->{'mysqldump'};
	  } else {
		$dumpcmd = $bin{'mysqldump'} || 'mysqldump';
	  }

	  my $dumpopts = " --quick --quote-names --add-drop-table";

	  my $mysqlversion = `$commonprefix $bin{'mysql'} $commonopts --version`;
	  if ($mysqlversion =~ /Distrib (\d+)\.(\d+)\./) {
		if ($1 > 4 or ($1 == 4 and $2 > 0)) {
		  $dumpopts .= " --compatible=mysql323";
		}
	  }

	  $cmd = $commonprefix . $dumpcmd . $commonopts . $dumpopts;

	} elsif ($dbType eq 'Pg') {
	  $cmd  = "export PGUSER='$dbUser'";
	  $cmd .= " PGPASSWORD='$dbPasswd'" if($dbPasswd);
	  $cmd .= ' ; ' . ($bin{'pg_dump'} || 'pg_dump') . ' -d -O';
	  $cmd .= " -h '$dbHost'" if $dbHost;
	  $cmd .= " '$dbName'";
	}

	if ($compress and $gzip and -x $gzip) {
	  $cmd .= " | $gzip";
	}
	if (-x $split) {
	  $cmd .= " | $split -b 2040m - $dumpFile.";
	} else {
	  $cmd .= " > $dumpFile";
	}

	local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

	unless (system($cmd) == 0) {
	  printToError("Error execute '$cmd'");
	  return undef;
	}
  }
  return $dumpName;
}

sub addDbUsers {
  my ($dbNode, $dbName, $wrapUserMysql) = @_;

  printToLog("Dupming $dbName database users");

  $dbName =~ s/_/\\_/g;

  my (@users, $dbUser, $ptrRow);
  my $sql = "SELECT DISTINCT User FROM db WHERE Db = '$dbName'";
  $wrapUserMysql->{'EXECUTE'}->($sql);
  while ($ptrRow = $wrapUserMysql->{'FETCHROW'}->()) {
		push @users, (@{$ptrRow})[0];
  }
  $wrapUserMysql->{'FINISH'}->();

  foreach $dbUser (@users) {
		addDbUser($dbNode, $dbUser, $wrapUserMysql);
  }
}

sub addDbUser {
  my ($dbNode, $user, $wrapUserMysql, $passwordNode) = @_;

  &printToLog("Dumping mysql user '$user'");

	my ( $item );

  if ( ref( $wrapUserMysql ) =~ /HASH/ &&
			 ref( $wrapUserMysql->{'EXECUTE'} ) =~ /CODE/ ) {

    my ($sql,$ptrRow,@accesshosts,$accesshost);

    $sql = "SELECT password,host FROM user WHERE user='$user'";

    if ( $wrapUserMysql->{'EXECUTE'}->( $sql ) ) {

			$item = makeXmlNode( 'dbuser' );
			$item->{'ATTRIBUTE'}->('name', $user);

      while ( $ptrRow = $wrapUserMysql->{'FETCHROW'}->() ) {
				unless ( ref($passwordNode) =~ /HASH/ ) {
					$passwordNode = makePasswordNode( $ptrRow->[0] );
				}
				push @accesshosts, $ptrRow->[1];
      }

			if ( $passwordNode ) {
				$item->{'ADDCHILD'}->( $passwordNode );
			}

			foreach $accesshost ( @accesshosts ) {
				$item->{'ADDCHILD'}->( makeXmlNode( 'accesshost', $accesshost ) );
			}

			$dbNode->{'ADDCHILD'}->( $item );

    }

    $wrapUserMysql->{'FINISH'}->();
  }

	return $item;
}

my $lastDumpId = 0;

sub generateDumpId {
  return $$ . "-" . $lastDumpId++;
}

#
# sub makeDumpFile - make a dump of dir tree (use 'tar')
#
# Arguments:
#          $destFile - name of destination file.
#                      If the file has not '.tgz' or '.tar.gz' extension
#                      '.tgz' will be added
#          $scrDir   - path to source dir
#
#          $ptrNames - pointer to an array with names of files and dirs to
#                      include into dump
#                      can be a single name
#                      by default '.' - include all entries
#
#          $ptrExcludes - pointer to an array with names of files and dirs to
#                      exclude from dump
#                      can be a single name
#
#          $extraParams - extra params to run tar
#                         ('-h' - don't dump symlinks; dump the files they point to)
#
sub makeDumpFile {
  my ($destFile, $srcDir, $ptrNames,$ptrExclude,$extraParams)=@_;

  my @destPath = split(/\//, $destFile);
  if (length($destPath[$#destPath]) > 32) {
	$destPath[$#destPath] = generateDumpId();
  }
  $destFile = join('/', @destPath);

  my $tar = $bin{'tar'}||'tar';
  my $split = $bin{'split'}||'split';
  my $gzip = $compress == 1 ? ($bin{'gzip'} || 'gzip') : undef;

  my ($cmd,$ret,@names,@exclude);

  if(ref($ptrNames)=~/ARRAY/){
    @names = @{$ptrNames};
  }else{
    @names=($ptrNames) if $ptrNames;
  }

  if(ref($ptrExclude)=~/ARRAY/){
    @exclude = @{$ptrExclude};
  }else{
    @exclude=($ptrExclude) if $ptrExclude;
  }

  unless(-x $tar ){
    &printToError("Error: '$tar' is not executable");
    return undef;
  }

#
# check extension
#
  if(-x $gzip){
    unless($destFile=~/\.tgz$/ ||$destFile=~/\.tar\.gz$/){
      $destFile .= '.tar.gz';
    }
  }else{
    unless($destFile=~/\.tar$/){
      $destFile .= '.tar';
    }
  }
#
# end check extension
#

  unless( exists $ptrArgs->{'no-content'}){

    $cmd=$tar;
    if($srcDir){
      if( -d $srcDir){
		$cmd.=" -C $srcDir -f - -c --ignore-failed-read ";
      }else{
		return undef;
      }
    }


    if($extraParams){
      $cmd .= $extraParams.' ';
    }

    if(@names){
      if($srcDir){
		$cmd .= join( ' ', map {s/^\Q$srcDir\E\///;$_} @names); ## extract domain name
      }else{
		$cmd .= join( ' ', @names );
      }
    }else{
      $cmd .= '.';  ## dump all files & dirs
    }
    if(@exclude){
      $cmd .= ' '.join( ' ', map {s/^\Q$srcDir\E\///;'--exclude '.$_} @exclude); ## extract domain name
    }
    if(-x $gzip){
      $cmd .= " | $gzip";
    }
    if(-x $split){
      $cmd .= " | $split -b 2040m - $destFile.";
    }else{
      $cmd .= " >$destFile";
    }

		local $SIG{'PIPE'} = 'DEFAULT' if $SIG{'PIPE'} eq 'IGNORE';

    unless(system($cmd) == 0){
      &printToError("Error execute command: '$cmd': $!");
      $destFile='';
    }
  }

  $destFile=~s/^\Q$dumpDir\E\///;  ### remove dir name
  return $destFile;
}

sub getDumpDir {
  my ($subdir,$clean) = @_;
  my ($dumpdir);

  if($subdir=~/^\//){
    $dumpdir=$subdir;
  }else{
    my $util = $bin{'pwd'}||'pwd';
    $dumpdir = $pwd||`$util`;
    chomp $dumpdir;
    if($subdir){
      $dumpdir.='/'.$subdir;
    }
  }
  if(-d $dumpdir && $clean){
    unless(system("rm -rf $dumpdir")==0){
      &printToError("Error: can not clean temporary dir '$dumpdir': $!");
    }
  }
  unless(-d $dumpdir){
    mkdir ($dumpdir,0755) or
      die "Error: can not create temporary dir '$dumpdir': $!\n";
  }
  return $dumpdir;
}

sub printXml {
  my ($root,$outFh,$dtd) = @_;
  unless($outFh){
    $outFh = *STDOUT;
  }

  unless(ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/){
    print $outFh "Error: a xml root is not found\n";
    &printToError("Error: sub printXml: bad arguments");
    return 0;
  }
  print $outFh "<?xml version='1.0' encoding=\"UTF-8\"?>\n";

  my  $doctype = $root->{'NAME'}->();
  if($dtd){
    print $outFh "<!DOCTYPE $doctype SYSTEM '$dtd'>\n";
  }
  $root->{'PRINT'}->($outFh,'  ');
  return 1;
}

#
# class for handle with XML
#
sub makeXmlNode {
  my ($this,$parent,%attributes,@content);
  my ($name,$plain)=@_;

  $plain = toUtf8(encodeHtml($plain));

  unless($name||$plain){
    return undef;
  }
  $this = {
	   'ATTRIBUTE' => sub {
	     my $key = shift;
	     if(@_){
	       my $value=shift;
	       if(@_){
					 $attributes{$key} = $value;
	       }else{
					 $attributes{$key} = toUtf8(normalizeAttribute($value));
	       }
	     }
	     return $attributes{$key};
	   },
	   'PARENT' => sub {
	     if(@_){
	       my $arg = shift;
	       if(ref($arg)=~/HASH/){
					 $parent=shift;
	       }
	     }
	     return $parent;
	   },
	   'NAME' => sub {
	     if(@_){
	       $name = shift;
	     }
	     return $name;
	   },
	   'CONTENT' => sub {
	     if(@_){
	       $plain = toUtf8(encodeHtml(shift));
	     }
	     return $plain;
	   },
	   'ADDCHILD'=> sub {
	     my $kind = shift;
	     if(
					(ref($kind)=~/HASH/) &&
					(ref($kind->{'PARENT'})=~/CODE/)
	       ) {
	       push @content,$kind;
	       $kind->{'PARENT'}->($this);
	     }
	     return $#content;
	   },
	   'KIND'=> sub {
	     my $index = shift;
			 my ($ret);
			 if ( $index =~ /^\d+$/ ) {
				 if($index >= 0 && $index <= $#content ) {
					 $ret =  $content[$index];
				 }
			 } else {
				 foreach my $item (@content){
					 if ( $item->{'NAME'}->() eq $index ) {
						 $ret = $item;
						 last;
					 }
				 }
			 }
			 return $ret;
	   },
	   'KINDSREF'=> sub {
			 my $name = shift;
			 my $ret = \@content;
			 if ( $name ) {
				 $ret = [];
				 foreach my $item (@content) {
					 if ( $item->{'NAME'}->() eq $name ) {
						 push @{$ret}, $item;
					 }
				 }
			 }
	     return $ret;
	   },
	   'ATTRIBUTESREF' => sub {
	     return \%attributes;
	   },
	   'PRINT'=>sub {
	     my $fh = shift || *STDOUT;
	     my $addPrefix = shift || '';
	     my $prefix = shift || '';
	     if($name){
	       my ($out,$item);
	       if(@content || defined($plain)){

					 $out = $prefix . &getTag($name,\%attributes,'start');
					 print $fh $out;
					 if(@content){
						 print $fh "\n";
						 foreach $item (@content){
							 if(ref($item)=~/HASH/ && ref($item->{'PRINT'})=~/CODE/){
								 $item->{'PRINT'}->($fh,$addPrefix,$prefix.$addPrefix);
							 }
						 }
					 }
					 if(defined($plain)){
						 print $fh $plain;
						 if(@content){
							 print $fh "\n$prefix";
						 }
					 }else{
						 if($prefix){
							 print $fh $prefix;
						 }
					 }
					 $out = &getTag($name,\%attributes,'stop') . "\n";
					 print $fh $out;

	       }else{
					 $out = $prefix . &getTag($name,\%attributes) . "\n";
					 print $fh $out;
	       }
	     }else{
	       if(defined($plain)){
					 print $fh $plain;
	       }
	     }
	   }
	 };

  return $this;
}

sub getTag {
  my ($name,$ptrAttr,$mode) = @_;
  my ($out,$key,$value);
  $mode||='';
  if ($mode eq 'stop'){
    $out="</$name>";
  }else{
    $out="<$name";
    while (($key,$value)=each(%{$ptrAttr})){
      $out .= ' '.$key;
      if (defined($value)){
		$out .= '="'.$value.'"';
      }
    }
    if($mode eq 'start'){
      $out .= ">";
    }else{
      $out .= "/>";
    }
  }
  return $out;
}

#
# end class for handle with XML
#


#
# class for handle with database
#
sub getDbConnect {
  my $ptrBin = shift;
  my $myDbType = shift;
  my $myDbAddress = shift ;
  my $myDbUser = shift ;
  my $myDbPasswd = shift ;
  my $myDbDB = shift ;
  my $myDbHost = shift ;

  #added by Honig for Sphera
  my $myDbSocket = shift;
  my $variables = shift;

  my ($dbh,$this,$sth);
  my ($shell,$args);
  my $echo = $ptrBin->{'echo'}||'echo';
  my ($sql,@result,$resultIndex,$ptrResult);

  unless(defined($myDbPasswd) && defined($myDbUser)){
    if($myDbType eq 'mysql'){
      foreach my $myCnf ('/etc/my.cnf',$ENV{'HOME'}.'/.my.cnf'){
		next unless (-T $myCnf);
		next unless open(CNF,"<$myCnf");
		binmode CNF;
		my($key,$value);
		my $state=0;
		while (<CNF>){
		  next unless $_;
		  next if (/^#/);
		  if($state){
			if(/^\s*\[/){
			  last;
			}
			if(/^\s*(\S+)\s*=\s*(\S+)/){
			  $key=$1;
			  $value=$2;
			  if($key =~ /^user$/){
				$myDbUser = $value;
				# sometimes strings are quoted
				$myDbUser =~ s/^('|")(.*)\1$/$2/;
			  }elsif($key =~ /^pass$/){
				$myDbPasswd = $value;
				# sometimes strings are quoted
				$myDbPasswd =~ s/^('|")(.*)\1$/$2/;
			  }
			}
		  }else{
			if(/^\s*\[client\]/){
			  $state=1;
			}
		  }
		}
		close (CNF);
      }
    }elsif($myDbType eq 'Pg'){
      $myDbPasswd ||= $ENV{'PGPASSWORD'};
      $myDbUser ||= $ENV{'PGUSER'};
    }
  }

  if (eval "require DBI"){
    if (eval "require DBD::$myDbType"){
      if($myDbAddress){
		if($myDbUser && $myDbPasswd){
		  $dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}elsif($myDbUser){
		  $dbh = DBI->connect($myDbAddress,$myDbUser) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}else{
		  $dbh = DBI->connect($myDbAddress) or
			die "Error: unable to connect to database: $DBI::errstr\n";
		}
      }
    }
  }
#  unless($dbh){
#    if ($myDbType eq 'Pg'){
#      if(eval "require Pg"){
#        $myDbAddress="user=$myDbUser dbname=$myDbDB";
#        if($myDbPasswd){
#          $myDbAddress.=" password=$myDbPasswd";
#        }
#        $dbh = Pg::connectdb($myDbAddress);
#      }
#    }
# }
	unless ($dbh) {

		if ($myDbType eq 'mysql') {
			$shell = $ptrBin->{'mysql'} || 'mysql';
			$args  = " --socket='$myDbSocket'" if ($myDbSocket);
			if ($myDbPasswd) {
				$args .= " -B -u '$myDbUser' -p'$myDbPasswd' '$myDbDB' ";
			} else {
				$args .= " -B -u $myDbUser $myDbDB ";
			}
			$shell = "LD_LIBRARY_PATH=$variables $shell" if ($variables);
    }elsif($myDbType eq 'Pg'){
      $shell = $ptrBin->{'psql'}||'psql';
	  if($myDbPasswd) {
		$shell = "PGPASSWORD=$myDbPasswd $shell";
	  }
	  if($myDbUser) {
		$shell = "PGUSER=$myDbUser $shell";
	  }

      $args = " -d $myDbDB -c ";
    }
#
# check connection to Confixx databnase
#
    if($myDbDB=~/confixx/){
      $sql = "select hostname from admin";
      @result = `$echo "$sql"| $shell $args`;
      unless($result[0]=~/^\w+$/){
		$args = ' -h $myDbHost'.$args;
		@result = `$echo "$sql"| $shell $args`;
		unless($result[0]=~/^\s*hostname/){
		  die "Error: Can not connect to Confixx database (unable to execute sql: $sql)\n@result\n";
		}
      }
    }
#
# end check connection to Confixx databnase
#
    $shell .= $args;
  }
  $this = {
	   'RECONNECT'=>sub{
	     if($dbh){
	       $dbh->disconnect();
		   if($myDbPasswd){
			 $dbh = DBI->connect($myDbAddress,$myDbUser,$myDbPasswd) or
			   die "Error: unable to connect to database: $DBI::errstr\n";
		   }else{
			 $dbh = DBI->connect($myDbAddress,$myDbUser) or
			   die "Error: unable to connect to database: $DBI::errstr\n";
		   }
	       return $dbh;
	     }else{
	       return 1;
	     }
	   },
	   'EXECUTE'=> sub {
	     $sql = shift;
	     if($dbh){
	       $sth = &executeDbh($dbh,$sql);
	       if($sth){
		 return $sth->rows;
	       }else{
		 return 0;
	       }
	     }else{
	       $resultIndex=0;
	       &executeDbShell($shell,\@result,$echo,$sql,$myDbType);
	       return scalar(@result);
	     }

	   },
	   'FETCHROW'=> sub {
	     if($sth){
	       if(@result = $sth->fetchrow){
		 return \@result;
	       }else{
		 return undef;
	       }
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
		 return undef;
	       }else{
		 return $result[$resultIndex];
	       }
	     }
	   },
	   'FETCHHASH'=>sub {
	     if($sth){
	       $ptrResult = $sth->fetchrow_hashref();
	     }else{
	       $resultIndex++;
	       if($resultIndex>$#result){
		 $ptrResult = undef;
	       }else{
			 my ($i,$key);
			 $ptrResult = {};

			 for($i=0; $i<@{$result[0]}; $i++){
			   $ptrResult->{$result[0]->[$i]}=$result[$resultIndex]->[$i];
			 }
	       }
	     }
	     return $ptrResult;
	   },
	   'FINISH'=>sub {
	     if($sth){
	       $sth->finish();
	     }
	     @result=();
	     $ptrResult = undef;
	   },
	   'PRINTERROR'=>sub {
	     if($dbh){
	       if(@_){
			 $dbh->{'PrintError'}=shift;
	       }
	       return $dbh->{'PrintError'};
	     }else{
	       return 0;
	     }
	   },
	   'PING'=>sub{
	     if($dbh){
	       return $dbh->ping();
	     }else{
	       return 1;
	     }
	   },
	   'DISCONNECT'=>sub{
	     if($dbh){
	       return $dbh->disconnect();
	     }else{
	       return 1;
	     }
	   },
	   'USER'=>sub{
	     return $myDbUser;
	   },
	   'PASSWORD'=>sub{
	     return $myDbPasswd;
	   },
	   'HOST'=>sub{
	     return $myDbHost;
	   },
	   'TYPE'=>sub{
	     return $myDbType;
	   },
	  };

  return $this;
}

sub executeDbh {
  my $dbh = shift;
  unless(ref($dbh)=~/DBI/){
    &printToError ("Error: DBI object is not found");
  }
  my $sql = shift;

  my $sth = $dbh->prepare($sql) or
    &printToError ("Error: prepare sql: $sql\n$DBI::errstr");
  if($sth->execute){
    return $sth;
  }else{
    &printToError ("Error: exesute sql: $sql\n$DBI::errstr");
  }
}

sub executeDbShell {
  my ($shell,$ptrRes,$echo,$sql, $type) = @_;
  @{$ptrRes}=();
  my (@out,@row,$ret);
  $ret=0;
  if($type eq 'Pg'){
    @out = `$shell "$sql"`;
  }else{
    @out = `$echo "$sql"|$shell`;
  }
  if ($out[0]=~/^error/i){
    &printToErrror ("Error: execute sql: $sql\n@out");
  }else{
    if($type eq 'Pg'){
      foreach my $line (@out){
	chomp $line;
	next if ($line=~/^-[-+]*$/); # skip delimiter
	if($line=~/^\(\s*\d+\s+\w+\)$/){
	    last;  ## (6 rows)
	}
	$line=~s/^\s+//;
	$line=~s/\s+$//;
	@row = split(/\s*\|\s*/,$line);
	push @{$ptrRes}, [@row];
	$ret++;
      }
    }else{
      if($out[0]=~/^\+[-+]*$/){
	foreach my $line (@out){
	  chomp $line;
	  next if ($line=~/^\+[-+]*$/); # skip delimiter
	  $line=~s/^\|\s*//;
	  $line=~s/\s+$//;
	  @row = split(/\s*\|\s*/,$line);
	  @row = map { $_ eq 'NULL' ? undef : $_ } @row;
	  push @{$ptrRes}, [@row];
	  $ret++;
	}
      }else{  ## batch mode
	foreach my $line (@out){
	  chomp $line;
	  @row = split(/\t/,$line);
	  @row = map { $_ eq 'NULL' ? undef : $_ } @row;
	  push @{$ptrRes}, [@row];
	  $ret++;
	}
      }
    }
  }
  return $ret;
}

#
# end class for handle with database
#

sub makeMIMEBase64 {
  my ($useModule,$this);
  if (eval "require MIME::Base64"){
    $useModule = 1;
  }else{
    $useModule = 0;
  }

  $this={
	 'ENCODE'=>sub{
	   my $src = shift;
	   my ($encoded);
	   if($useModule){
	     $encoded = MIME::Base64::encode($src);
		 chomp $encoded;
	   }else{
	     $encoded = &encodeBase64($src);
	   }
	   return $encoded;
	 },
	 'USEMODULE'=>sub{
	   return $useModule;
	 },
	};

  return $this;
}

#
# url decode (certificates in Plesk DB encoded by the url encoding)
#

sub urlDecode {
    my $url = $_[0];
    $url =~ tr/+/ /;
    $url =~ s/%([a-fA-F0-9]{2,2})/chr(hex($1))/eg;
    return $url;
}

sub encodeHtml {
  my(@ret,$src);
  if ($src=shift){
    if(ref $src){
      if(${$src}=~/[&"'><]/){
		${$src}=~s/&/&amp;/sg;
		${$src}=~s/>/&gt;/sg;
		${$src}=~s/</&lt;/sg;
		${$src}=~s/\'/&#39;/sg;
		${$src}=~s/\"/&quot;/sg;
      }
      return ${$src};
    }else{
      if($src=~/[&"'><]/){
		$src=~s/&/&amp;/sg;
		$src=~s/>/&gt;/sg;
		$src=~s/</&lt;/sg;
		$src=~s/\'/&#39;/sg;
		$src=~s/\"/&quot;/sg;
      }
    }
  }
  return $src;
}

# Normalization according to the XML 1.0 (Third Edition),
# clauses 2.3 "Common Syntactic Constructs"
# and 3.3.3 "Attribute-value normalization"

sub normalizeAttribute {
  $_ = $_[0];
  s/\r/&#xD;/g;
  s/\n/&#xA;/g;
  s/\t/&#x9;/g;
  s/\</&lt;/g;
  s/\&/&amp;/g;
  s/\"/&quot;/g;
  s/\'/&#39;/g;
  return $_;
}

############################
#
#	Non-ascii characters handling
{
	# static
	my $charset;

	sub setCharset {
	  $charset = shift;
	}

	# try to determine the local charmap
	sub _determineCharset()
	{
	  my $charset;
	  my $locale = $bin{'locale'} || 'locale';

	  if (exists $ENV{'LANG'}) {
		my @pieces = split /\./, $ENV{'LANG'};
		if (defined $pieces[1]) {
		  $charset = $pieces[1];
		  chomp $charset;
		}
	  }
	  if (not $charset && open(CM, "$locale charmap |")) {
	  	binmode CM;
		$charset = <CM>;
		close(CM);
		chomp $charset;
	  }
	  if (not $charset) {
		$charset = 'ISO-8859-1';
	  }

	  return $charset;
	}

	sub toUtf8
	{
		my ($subj, $myCharset) = @_;

		$charset = _determineCharset() if (not $charset);

		$myCharset = $charset unless ( $myCharset );

		if ( (not $subj) ||
				 (not $bin{'iconv'}) ||
				 ( $myCharset =~ /UTF-?8/i )
			 ) {
			return $subj;
		}

		if ($subj =~ /^[\x00-\x7f]*$/) { return $subj; }

		my $converted;
		my $escaped = $subj;
		$escaped =~ s/'/'"'"'/g;

		if (open(IC, "echo '$escaped' | " . $bin{'iconv'} . " -f '$myCharset' -t 'UTF-8' |")) {
			binmode IC;
			while (my $string = <IC>) { $converted .= $string; }
			close(IC);
			chomp $converted;
			if ($converted) {
			  return $converted;
			}
		}

		return "";
	}
}

#
# Based on the b64.c by Bob Trower (base64.sf.net)
#

my @cb64 = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
			'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
			'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
			'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
			'g', 'h', 'i', 'j',	'k', 'l', 'm', 'n',
			'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
			'w', 'x', 'y', 'z', '0', '1', '2', '3',
			'4', '5', '6', '7', '8', '9', '+', '/');

@cb64 = map { ord } @cb64;

sub encodeBlock {
  my ($len, @in) = @_;
  my @res;
  push @res, $cb64[$in[0] >> 2];
  push @res, $cb64[(($in[0] & 0x3)<<4)|(($in[1]&0xf0)>>4)];
  push @res, $len>1 ? $cb64[(($in[1]&0xf)<<2)|(($in[2]&0xc0)>>6)] : ord('=');
  push @res, $len>2 ? $cb64[$in[2]&0x3f] : ord('=');
  return @res;
}

sub encodeBase64 {
  my @in = unpack "C*", $_[0];

  my @out;
  while (scalar(@in)) {
	my @block = (0, 0, 0);
	my $length = 0;
	while (scalar(@in) && $length < 3) {
	  $block[$length++] = shift(@in);
	}
	push @out, encodeBlock($length, @block);
  }
  return pack "C*", @out;
}

sub getBins {
  my ($ptrBin,$path) = @_;
  my (%searchPath,$key,$prefix,$bin,$fullPath);
  my ($notFound,$i,@prefixes);

  $notFound=0;

  $i=0;
  foreach $prefix (split(/:/,$path)){
    next unless $prefix;
    unless(exists($searchPath{$prefix})){
      $searchPath{$prefix}=$i++;
      push @prefixes,$prefix;
    }
  }

  foreach $key (keys %{$ptrBin}){
	$ptrBin->{$key} = undef;
    foreach $prefix (@prefixes){
      $prefix.='/' unless $prefix=~/\/$/;
      $bin = $key;
      $bin =~ s/^\///;
      $fullPath = $prefix.$bin;
      if (-x $fullPath){
		$ptrBin->{$key}=$fullPath;
		last;
      }
    }
    $notFound++;
  }
  return $notFound;
}

sub getArguments {
  my ($ptrArgv,$ptrOpts) = @_;

  my (@keys,$firstKey,%prepKeys,$key,$value,$ptrArr,$arg,$state);
  my (%retHash,$pat,$found,$used,@rest,$fullArg,$prevKey);

  while (($key,$value)=each(%{$ptrOpts})){
    @keys = split(/\|/,$key);
    $firstKey = $keys[0];
    $firstKey =~s/^-*//;
    push @{$prepKeys{$firstKey}},[@keys];
    push @{$prepKeys{$firstKey}},$value;
  }

  $state =0;
  foreach $arg (@{$ptrArgv}){

#    print "arg: $arg\n";

    $used = 0;
    if($state==1){
      if  ($arg =~ /^-/){
	$state=0;
      }else{
	$used = 1;
	$retHash{$prevKey}=$arg;
	next;
      }
    }
    if ($state == 2){
      $retHash{$prevKey}=$arg;
      $state = 0;
      $used = 1;
    }else{
      $fullArg = $arg;
      if ($arg =~ /^(-\S+)=(\S+)/){
	$arg = $1;
	$value = $2;
      }else{
	$value = undef;
      }
      foreach $key (keys %prepKeys){

	$ptrArr = $prepKeys{$key};

#	print "key: $key, ",join('|',@{$ptrArr->[0]}),"\n";


	$found = 0;
	foreach $pat (@{$ptrArr->[0]}){
	  if ($pat eq $arg){
	    $found = 1;
	    last;
	  }
	}
	if($found){
	  $used = 1;
	  if(defined($value)){
	    $retHash{$key}=$value;
	  }else{
	    if($ptrArr->[1]){
	      if($ptrArr->[1] =~ /\?$/){
		$state = 1;
	      }else{
		$state = 2;
	      }
	      $prevKey = $key;
	    }else{
	      $retHash{$key}=undef;
	    }
	  }
	  last;
	}
      }
    }
    unless($used){
      if($value){
	push @rest,$fullArg;
      }else{
	push @rest,$arg;
      }
    }
  }
  @{$ptrArgv}=@rest;
  return \%retHash;

}

sub makeDumpStatus($) {
  my $statusFile = shift;
  my($pid,$currAccount,$currDomain,$leftAccounts,$leftDomains,$this);
  $pid=$$;
  $this = {'FILE'=>sub {
	     if(@_){
	       $statusFile = shift;
	     }
	     return $statusFile;
	   },
	   'PID'=>sub {
	     if(@_){
	       $pid = shift;
	     }
	     return $pid;
	   },
	   'ACCOUNT'=>sub{
	     if(@_){
	       $currAccount = shift;
	       if($leftAccounts){
		 $leftAccounts--;
	       }
	     }
	     return $currAccount;
	   },
	   'DOMAIN'=>sub{
	     if(@_){
	       $currDomain = shift;
	       if($leftDomains){
		 $leftDomains--;
	       }
	     }
	     return $currDomain;
	   },
	   'COUNTS'=>sub{
	     if(@_){
	       my $newValue = shift;
	       if (defined $newValue){
		 $leftAccounts=$newValue;
	       }
	       $newValue = shift;
	       if (defined $newValue){
		 $leftDomains=$newValue;
	       }
	     }
	     return ($leftAccounts,$leftDomains);
	   },
	   'PRINT'=>sub{
	     if(@_){
	       $statusFile=shift;
	     }
	     open(STATUS,">$statusFile.tmp");

		 binmode STATUS;
	     my $root = &makeXmlNode('agent-dump-status');
	     my $item = &makeXmlNode('pid',$pid);
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('account',$currAccount||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domain',$currDomain||'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('accounts-left',
				  defined($leftAccounts)?$leftAccounts:'');
	     $root->{'ADDCHILD'}->($item);
	     $item = &makeXmlNode('domains-left',
				  defined($leftDomains)?$leftDomains:'');
	     $root->{'ADDCHILD'}->($item);

	     &printXml($root,*STATUS,'agent-dump-status.dtd');

	     close(STATUS);

		 rename("$statusFile.tmp", $statusFile);
	   }

	  };
  return $this;
}

sub makeContentList {
  my ($dirDump,$xmlDump) = @_;
  my ($name,$fullPath,%files,%fromXML,$size,@cids,$root,$item,$key,$cmd,
      $gzip,$wc,$md5sum);

  if(opendir(DUMP,$dirDump)){
    while ($name=readdir(DUMP)){
      $fullPath="$dirDump/$name";
      unless(-f $fullPath && $name=~/\.\w\w$/){
	next;
      }
      $key = $fullPath;
      $key =~ s/\.\w\w$//;
      $size=(stat($fullPath))[7];
      if(defined $size){
	push @{$files{$key}},[$fullPath,$size];
      }
    }
    closedir(DUMP);
  }

	if (-T $xmlDump) {
		if (open(DUMP, "<$xmlDump")) {
			binmode DUMP;
			while (<DUMP>) {
				if (@cids = /cid(?:_\w+)?=\"([^"]+)\"/g) {
					foreach $fullPath (@cids) {
						unless ($fullPath =~ /^\//) {
							$fullPath = "$dirDump/$fullPath";
						}
						$fromXML{$fullPath} = undef;
					}
				}
			}
			close(DUMP);
		}

		foreach $key (keys %files) {
			unless (exists($fromXML{$key})) {
				delete $files{$key};
			}
		}
	}

  $gzip = $bin{'gzip'}||'gzip';
  $wc = $bin{'wc'}||'wc';
  $md5sum = $bin{'md5sum'}||'md5sum';

  $root=&makeXmlNode('contentlist');

  my($compressed,$uncompressed,$list,$count,$ptrList,@sortedList,
     $ptrItem,$out);
  my($totalUncomp,$totalSize)=(0,0);
  foreach $key (keys %files){

#
# sort by extension  .aa .ab ...
#
    @sortedList =  sort {&getExt($a->[0]) cmp &getExt($b->[0])} @{$files{$key}};

#
# output list of files
#
    $compressed=0;
    foreach $ptrItem (@sortedList){
      ($fullPath,$size)=@{$ptrItem};
      $item=&makeXmlNode('file',$fullPath);
      $item->{'ATTRIBUTE'}->('size',$size,'noencode');
      $root->{'ADDCHILD'}->($item);

#
# get md5sum
#
      if(-x $md5sum){
	$out=`$md5sum $fullPath`;
	if($out=~/\s*([0-9a-fA-F]+)\s+/){
	  $item->{'ATTRIBUTE'}->('md5sum',$1,'noencode');
	}
      }

      $compressed += $size;

    }

    $uncompressed = undef;
#
# get unpacked size of the archive
#
    if ($key=~/\.tgz$|\.gz$/){
      if(-x $gzip){
	if(scalar(@sortedList)==1){
	  ($fullPath,$size)=@{$sortedList[0]};
	  ($uncompressed,$compressed)=&getZipSize($fullPath,$gzip);
	}else{
	  $list = join(' ',map {$_->[0]} @sortedList);
	  if(-x $wc){
	    $cmd = "cat $list | $gzip | $wc -c";
	    $uncompressed = `$cmd`;
	    chomp $uncompressed;
	    if ($uncompressed =~ /^\s*(\d+)\s*$/){
	      $uncompressed = $1;
	    }else{
	      $uncompressed = undef;
	    }
	  }else{
	    $cmd = "cat $list >$key";
	    if(system($cmd)==0){
	      ($uncompressed,$compressed)=&getZipSize($key,$gzip);
	      unlink $key;
	    }
	  }
	}
      }
    }
#
# calc total size
#
    $totalSize += $compressed;
    $totalUncomp += $uncompressed if $uncompressed;
  }

  $root->{'ATTRIBUTE'}->('size',$totalSize,'noencode');
  if ($totalUncomp && $totalUncomp != $totalSize){
    $root->{'ATTRIBUTE'}->('unpacked',$totalUncomp,'noencode');
  }

  return $root;
}

sub getExt {
  my $path = shift;
  if($path=~/\.([^.]*)$/){
    return $1;
  }
  return $path;
}

sub getZipSize {
  my ($fullPath,$gzip) = @_;
  $gzip ||= 'gzip';
  my $cmd = "$gzip -l $fullPath";
  my ($uncompressed,$compressed,@out);
  if (open(OUTPUT,"$cmd|")){
    ($compressed,$uncompressed)=(0,0);
    while(<OUTPUT>){
      @out = split(' ',$_);
      if($out[0]=~/^\d+$/ && $out[1]=~/^\d+$/){
	$uncompressed+=$out[1];
	$compressed+=$out[0];
      }
    }
    close (OUTPUT);
  }
  return $uncompressed,$compressed;
}

my $outputIsFile;

sub openOutput
{
	my ($outPath) = @_;
	my ($outFh);

	if ($outPath) {
		open(OUTPUT, ">>$outPath") or die "Error: can not open '$outPath' for write: $!\n";
		binmode OUTPUT;
		$outFh = *OUTPUT;
		$outputIsFile = 1;
	} else {
		$outFh = *STDOUT;
		$outputIsFile = 0;
	}

	return $outFh;
}

sub closeOutput{
  if($outputIsFile){
    close (OUTPUT) or
      &printToError("Error: unable to close OUTPUT: $!\n");
    $outputIsFile=0;
  }
}

#=========================================================================
#
# dtd-specific (but common) subroutines
#
sub createPasswordNode {
  my ($parent, $password, $type) = @_;

  my $item = makeXmlNode('password', $password);
  $item->{'ATTRIBUTE'}->('type', $type);
  $parent->{'ADDCHILD'}->($item);
}

sub createSysuserNode {
  my ($parent, $name, $password, $type) = @_;

  my $item = makeXmlNode('sysuser');
  $item->{'ATTRIBUTE'}->('name', $name);
  createPasswordNode($item, $password, $type);
  $parent->{'ADDCHILD'}->($item);
}

sub readFile {
  my ($fileName) = @_;

  unless (-f $fileName) {
	return undef;
  }

  open(INPUT, "< $fileName");
  binmode INPUT;
  my $content = join('', <INPUT>);
  close(INPUT);

  return $content;
}

1;

#
# end common subroutines
#
#=========================================================================


$ptrArgs = &getArguments (\@ARGV,\%arg_opts);

#
# search binaries
#

&getBins(\%bin,'/usr/local/psa/python2.2/bin:/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:/usr/local/sbin:'.$ENV{'PATH'});

my ($outPath,$outFh,$outHas,$dumpFile,$statusFile,$app);

$app = $0;
if ($app=~/(.+)\/([^\/]+)$/){
  $workDir = $1;
  $app = $2;
}
$app=~s/\.pl$//;

$pwd=`$bin{'pwd'}`||'.';

chop $pwd;

if($workDir eq '.'){
  $workDir = $pwd;
}elsif($workDir !~ /^\//){
  $workDir = $pwd.'/'.$workDir;
}

$dumpFile = $workDir.'/PleskX.xml';
$statusFile = $workDir.'/PleskX.status';
$dumpDir = $workDir.'/pma';
$pleskVersion = undef;
$pleskMajorVersion = undef;
$pleskMinorVersion = undef;
$pleskMinorMinorVersion = undef;

my $objDumpStatus = &makeDumpStatus($ptrArgs->{'status-file'}||$statusFile);

if(!$ptrArgs->{'output'} &&
   (exists $ptrArgs->{'dump-all'} ||
    exists $ptrArgs->{'dump-accounts'}||
    exists $ptrArgs->{'dump-domains'}
   )
  ){
  $ptrArgs->{'output'}=$dumpFile;
}

if ($outPath = $ptrArgs->{'output'}) {

	open(OUTPUT, ">$outPath") or die "Error: can not open '$outPath' for write: $!\n";
	binmode OUTPUT;
	$dumpFile = $outPath;
	close(OUTPUT);
}

if(exists $ptrArgs->{'version'}){
  print $VERSION,"\n";
  exit 0;
}

#
# get MIME Base64 encoder
#
my $wrapBase64 = makeMIMEBase64();

my (%PsaConfig,%domainsInfo);
my $rootName = 'PleskXdump';
my $dtdName = 'PleskX.dtd';

unless($ptrArgs->{'config'} = &loadMainConfig($ptrArgs->{'config'},\%PsaConfig)){
  if(exists $ptrArgs->{'get-status'}){

    my $outFh = &openOutput($ptrArgs->{'output'});

    &printAgentStatus($ptrArgs->{'output'}||$dumpFile,
		      $objDumpStatus->{'FILE'}->(),
		      $outFh);

    &closeOutput();

    exit 0;
  }else{

    if(exists ($ptrArgs->{'dump-all'}) ||
       exists ($ptrArgs->{'dump-accounts'}) ||
       exists ($ptrArgs->{'dump-domains'})){

      my $root=&makeEmptyDump($rootName);

      if(ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/){
		my $outFh = &openOutput($ptrArgs->{'output'});
		&printXml($root,$outFh,$dtdName);
		&closeOutput();
      }
    }

    die "Error: Psa config file 'psa.conf' is not found\n";
  }
}
#
# get db connect
#
# It's better to always parse psa.conf to prevent troubles with Standard builds
my $mysqlpath = `cat /etc/psa/psa.conf | grep ^MYSQL_BIN | awk '{print \$2}'`;
chomp $mysqlpath;
if ($mysqlpath) {
	$bin{'mysql'}     = $mysqlpath . "/mysql";
	$bin{'mysqldump'} = $mysqlpath . "/mysqldump";
}

my $dbAddress = "DBI:$dbType:dbname=$dbName;host=$dbHost";
my $wrapDbh = &getDbConnect(\%bin,$dbType,$dbAddress,
			   $dbUser, $PsaConfig{'password'},
			   $dbName, $dbHost);
unless(ref($wrapDbh)=~/HASH/ && ref($wrapDbh->{'EXECUTE'})=~/CODE/){
  die "Error: can not connect to Plesk database\n";
}



if(exists $ptrArgs->{'get-status'}){

  my $outFh = &openOutput($ptrArgs->{'output'});

  &printAgentStatus($outPath=$ptrArgs->{'output'}||$dumpFile,
		    $objDumpStatus->{'FILE'}->(),
		    $outFh);

  &closeOutput();

}elsif(exists ($ptrArgs->{'dump-all'}) ||
       exists ($ptrArgs->{'dump-accounts'}) ||
       exists ($ptrArgs->{'dump-domains'})){

  my ($root,@accounts,@domains,$ptrAccounts,$ptrDomains, $value);
  if(exists $ptrArgs->{'no-content'}){
    $::dumpDir = &getDumpDir("$workDir/pma");
  }else{
    $::dumpDir = &getDumpDir("$workDir/pma",'clean');
  }

  if (exists $ptrArgs->{'no-compress'}) {
	setCompress(0);
  }

  $pleskVersion = getVersion();
  $pleskVersion =~ s/[^.0-9]//g;

  if ($pleskVersion) {
	&printToLog("Version of Plesk: $pleskVersion");
  } else {
	&printToLog("Unable to parse Plesk version into subcomponents: $pleskVersion");
  }

  if (atLeastVersion(7, 1, 0)) {
	setCharset("UTF-8");
  }

  &printToLog("Work dir: $workDir");
  &printToLog("Dump file: $dumpFile");
  &printToLog("Status file: ".$objDumpStatus->{'FILE'}->());

  if ($value = $ptrArgs->{'dump-accounts'}){
  	if ($value eq "-") {
		$value = <STDIN>;
		chomp $value;
	}
    @accounts = split(/\s*,\s*/,$value);
	$ptrAccounts = \@accounts;
  }

  if ($value = $ptrArgs->{'dump-domains'}){
  	if ($value eq "-") {
		$value = <STDIN>;
		chomp $value;
	}
    @domains = split(/\s*,\s*/,$value);
	$ptrDomains = \@domains;
  }
#
# generate a xml dump
#

  $root = &getPleskDump($dumpDir, $ptrAccounts, $ptrDomains, $rootName);

#
# print dump to output
#
  &printToLog('Print dump file ...',1);


  if(ref($root)=~/HASH/ && ref($root->{'PRINT'})=~/CODE/){
    my $outFh = &openOutput($ptrArgs->{'output'});
    &printXml($root,$outFh,$dtdName);
    &closeOutput();
  }


  &printToLog(' OK');

}elsif(exists $ptrArgs->{'get-content-list'}){

  $::dumpDir = &getDumpDir("$workDir/pma");

  my $root = &makeContentList($::dumpDir,$dumpFile);

  if(ref($root)=~/HASH/){
    my $outFh = &openOutput($ptrArgs->{'output'});

    &printXml($root,$outFh);

    &closeOutput();
  }

}else{
  &printHelp();
}


exit $returnCode;

#
# end main
#
#==============================================================
#
# subroutines
#

sub atLeastVersion {
  my ($major, $minor, $minorMinor) = @_;

  if ($pleskMajorVersion > $major) { return 1; }

  if ($pleskMajorVersion == $major && $pleskMinorVersion > $minor) { return 1; }

  if ($pleskMajorVersion == $major && $pleskMinorVersion == $minor
	  && $pleskMinorMinorVersion >= $minorMinor)
	{
	  return 1;
	}

  return 0;
}

sub moreThanVersion {
  my ($major, $minor, $minorMinor) = @_;

  if ($pleskMajorVersion > $major) { return 1; }
  if ($pleskMajorVersion == $major && $pleskMinorVersion > $minor) { return 1; }
  if ($pleskMajorVersion == $major && $pleskMinorVersion == $minor && $pleskMinorMinorVersion > $minorMinor) { return 1;}

  return 0;
}

# Condition is not necessary if getting admin parameters
sub getMaxButtonLength {
  my ($parent, $table, $condition) = @_;

  my $sql = "SELECT val FROM $table WHERE param='max_button_length'";
  $sql .= " AND $condition" if $condition;

  if ($wrapDbh->{'EXECUTE'}->($sql) and my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	if ($ptrRow->[0]) {
		$parent->{'ATTRIBUTE'}->('max-button-length', $ptrRow->[0]);
	}
  }
  $wrapDbh->{'FINISH'}->();
}

#Condition is not necessary if getting admin parameters
sub getSkin {
  my ($parent, $table, $condition) = @_;

  return unless atLeastVersion(5, 0, 0);

  my $sql = "SELECT s.name FROM Skins as s, $table as t " .
	"WHERE t.param='skin_id' AND t.val = s.id";
  $sql .= " AND $condition" if $condition;

  if ($wrapDbh->{'EXECUTE'}->($sql) and my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	$parent->{'ATTRIBUTE'}->('skin', $ptrRow->[0]);
  }
  $wrapDbh->{'FINISH'}->();
}

sub getLockScreen {
  my ($parent, $table, $condition) = @_;

  return unless atLeastVersion(7, 5, 2);

  my $sql = "SELECT val FROM $table WHERE param='disable_lock_screen'";
  $sql .= " AND $condition" if $condition;

  my $ptrRow;
  if ($wrapDbh->{'EXECUTE'}->($sql)
	  and $ptrRow = $wrapDbh->{'FETCHROW'}->()
	  and $ptrRow->[0] eq 'true') {
	$parent->{'ATTRIBUTE'}->('lock-screen', 'false');
  }
}

sub supportedVersion {
  return !moreThanVersion(7, 5, 3);
}

my $supportedVersionStr = "7.5.3";

sub getPleskDump {
  my ($dumpDir, $ptrAccounts, $ptrDomains, $rootName) = @_;
  $rootName = 'PleskXdump' unless $rootName;

  my ($sql,$list,$ptrRow,$name,$domain,$item,
      %ipMap,%domains,%clients,$ptrClients,$fakeClient);
  my ($domainsCount, $clientsCount);

  my $root = &makeXmlNode($rootName);

  unless(ref($wrapDbh)=~/HASH/ && ref($wrapDbh->{'EXECUTE'})=~/CODE/){
    $root->{'CONTENT'}->("Error: there is no connect to database");
    &printToError("Error: unable to connect to SQL database");
    return $root;
  }


#
# prepare list of domains
#
  if (ref($ptrDomains)=~/ARRAY/ && (@{$ptrDomains}>0)){
    $fakeClient= &makeXmlNode('client');

	my $domainNameColumn = 'name';
	if (atLeastVersion(7, 1, 0)) {
	  $domainNameColumn = 'displayName';
	}

    $list = join(',',map {"'$_'"} @{$ptrDomains});
    $sql = "SELECT d.id,d.name,d.cl_id,c.login FROM domains d, clients c ".
      "WHERE d.$domainNameColumn in ($list) AND d.cl_id = c.id";
    if($wrapDbh->{'EXECUTE'}->($sql)){
      while( $ptrRow = $wrapDbh->{'FETCHROW'}->()){
		$domains{$ptrRow->[0]}=[$ptrRow->[1],$ptrRow->[2]];
		$clients{$ptrRow->[2]}=[$ptrRow->[3],0];
      }
    }
    $wrapDbh->{'FINISH'}->();
  }

  if (ref($ptrAccounts)=~/ARRAY/ && (@{$ptrAccounts}>0)) {
	$list = join (',', map {"'$_'"} @{$ptrAccounts});

	$sql = "SELECT c.id, c.login, d.id, d.name FROM clients c LEFT JOIN domains d ".
	  "ON c.id = d.cl_id WHERE c.login in ($list)";

	if ($wrapDbh->{'EXECUTE'}->($sql)) {
	  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		$clients{$ptrRow->[0]} = [$ptrRow->[1], 0];
		$domains{$ptrRow->[2]} = [$ptrRow->[3], $ptrRow->[0]];
	  }
	}
  }

  unless (keys %clients) {
	$sql = "SELECT id, pname FROM clients";
	$wrapDbh->{'EXECUTE'}->($sql);
	while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  $clients{$ptrRow->[0]} = [$ptrRow->[1],1];
	}
	$wrapDbh->{'FINISH'}->();
  }
#
# end prepare list of domains
#

#
# prepare list of clients
#

  if(ref($ptrAccounts)=~/ARRAY/ && (@{$ptrAccounts}>0)){

    $fakeClient = undef;

    $list = join(',',map {"'$_'"} @{$ptrAccounts});
    $sql = "SELECT id, login FROM clients WHERE login in ($list)";
    if($wrapDbh->{'EXECUTE'}->($sql)){
      while( $ptrRow = $wrapDbh->{'FETCHROW'}->()){
		$clients{$ptrRow->[0]}=[$ptrRow->[1],1];
      }
    }
    $wrapDbh->{'FINISH'}->();
  }
#
# end prepare list of clients
#

#
# domains/clients count
#
  if (ref($ptrDomains) =~ /ARRAY/ && (@{$ptrDomains}>0)) {
	$domainsCount = scalar(@{$ptrDomains});
	$clientsCount = 1;
  } elsif (ref($ptrAccounts) =~ /ARRAY/ && (@{$ptrAccounts} > 0)) {
	my $list = join(",", keys %clients);
	$sql = "SELECT COUNT(*) FROM domains WHERE cl_id in ($list)";
	$wrapDbh->{'EXECUTE'}->($sql);
	$ptrRow = $wrapDbh->{'FETCHROW'}->();
	$wrapDbh->{'FINISH'}->();

	$domainsCount = $ptrRow->[0];
	$clientsCount = scalar(@{$ptrAccounts});
  } else {
	$sql = "SELECT COUNT(*) FROM domains";
	$wrapDbh->{'EXECUTE'}->($sql);
	$ptrRow = $wrapDbh->{'FETCHROW'}->();
	$wrapDbh->{'FINISH'}->();

	$domainsCount = $ptrRow->[0];

	$sql = "SELECT COUNT(*) FROM clients";
	$wrapDbh->{'EXECUTE'}->($sql);
	$ptrRow = $wrapDbh->{'FETCHROW'}->();
	$wrapDbh->{'FINISH'}->();

	$clientsCount = $ptrRow->[0];
  }
#
# end domains/clients count
#
  $objDumpStatus->{'COUNTS'}->($clientsCount, $domainsCount);


#
# prepare ip
#
  &printToLog('Dump ip map ...',1);

#
# %ipMap - hash 'ip' => 'id'
#
  if($pleskMajorVersion <= 5){
    $item = &makeIpNode5(\%ipMap,$ptrClients);
  }else{
    $item = &makeIpNode(\%ipMap,$ptrClients);
  }
  $root->{'ADDCHILD'}->($item);

  &printToLog(' OK');
#
# end prepare ip
#



#
# loop by client
#
  my($clientName,$full,$clientID);
  foreach $clientID (sort {$a <=> $b} keys %clients){

    ($clientName,$full)=@{$clients{$clientID}};

    &printToLog("Client '$clientName' is started");

    $objDumpStatus->{'ACCOUNT'}->($clientName);
    $objDumpStatus->{'PRINT'}->();

    $item = &makeClientNode($clientID,\%ipMap,$full,\%domains,$fakeClient);

    if(ref($item)=~/HASH/ && ref($item->{'PRINT'})=~/CODE/){
      $root->{'ADDCHILD'}->($item);
      &printToLog("Client '$clientName' is successfully dumped");
    }else{
      &printToLog("Client '$clientName' is not dumped");
    }
  }
#
# end loop by client
#

  $objDumpStatus->{'ACCOUNT'}->('');
  $objDumpStatus->{'DOMAIN'}->('');
  $objDumpStatus->{'COUNTS'}->(0,0);
  $objDumpStatus->{'PRINT'}->();


  return $root;
}

#
# Converts the old country codes to the recent ones
#
sub normalizeCountry {
  my ($country) = @_;
  if ($country eq 'SU') {
	return 'RU';
  }

  if ($country eq 'UK') {
  	return 'GB';
  }

  return $country;
}

#
# sub makeClientNode
#
#  arguments:
#         $clientID - ID of client
#         $ptrIphash
#         $full - dump all domains
#         $ptrDomainHash
#         $fakeClient - if all domains are in one fake client
#

sub makeClientNode {
  my($clientID,$ptrIpHash,$full,$ptrDomainsHash,$fakeClient)=@_;
  $ptrIpHash ||= {};
  $full = 1 unless defined($full) ;
  $ptrDomainsHash ||= {};

  my ($item,$root,$sql,$ptrHash,$value,%client,@domains,$ptrRow,$id,%clientParams);

#
# get client info
#
  $sql = "select * from clients where id = $clientID";

  unless($wrapDbh->{'EXECUTE'}->($sql)){
    $wrapDbh->{'FINISH'}->();
    return undef;
  }
  unless($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
    $wrapDbh->{'FINISH'}->();
    return undef;
  }
  %client=%{$ptrHash};

  $client{'country'} = normalizeCountry($client{'country'});

  $wrapDbh->{'FINISH'}->();

#
# end get client info
#

  $root = &makeXmlNode('client');
  my($attribute,$field,$name);
  while (($attribute,$field)=each(%clientsAttribute)){
    if(exists($client{$field})){
	  unless (defined($client{$field})) {
		$client{$field} = '';
	  }
      $root->{'ATTRIBUTE'}->($attribute,$client{$field});
    }
  }

  if (atLeastVersion(7, 1, 5)) {
	$sql = "SELECT password, type FROM accounts WHERE id = '"
	  . $client{'account_id'} . "'";

	$wrapDbh->{'EXECUTE'}->($sql);
	$ptrHash = $wrapDbh->{'FETCHHASH'}->();
	$item = makePasswordNode($ptrHash->{'password'}, pwdType($ptrHash->{'type'}));
  } else {
	$item = makePasswordNode($client{'passwd'},'plain');
  }

  $root->{'ADDCHILD'}->($item);

  unless (exists $ptrArgs->{'no-content'}) {
	if($client{'status'}){
	  if($pleskMajorVersion <= 5){
		$item = &makeXmlNode('status',$client{'status'});
		$root->{'ADDCHILD'}->($item);
	  }else{
		$item = &makeXmlNode('status','false');
		$root->{'ADDCHILD'}->($item);
	  }
	}

	while (($name, $field) = each(%clientsInfo)) {
	  if (exists($client{$field}) && $client{$field}) {
		$item = makeXmlNode('pinfo', $client{$field});
		$item->{'ATTRIBUTE'}->('name', $name);
		$root->{'ADDCHILD'}->($item);
	  }
	}

	# Locale changed format since 7.5.0, and we
	# don't bother converting older formats
	if (atLeastVersion(7, 5, 0)) {
	  if (exists($client{'locale'}) && $client{'locale'}) {
		my $localeNode = makeXmlNode('pinfo', $client{'locale'});
		$localeNode->{'ATTRIBUTE'}->('name', 'locale');
		$root->{'ADDCHILD'}->($localeNode);
	  }
	}

	if($pleskMajorVersion <= 5){
	  $sql = "SELECT param,val FROM cl_param WHERE cl_id = $clientID";
	  if($wrapDbh->{'EXECUTE'}->($sql)){
		while ($ptrRow=$wrapDbh->{'FETCHROW'}->()){
		  $clientParams{$ptrRow->[0]}=$ptrRow->[1];
		}
	  }
	  $wrapDbh->{'FINISH'}->();
	}

	# button length
	getMaxButtonLength($root, "cl_param", "cl_id=$clientID");
	getSkin($root, "cl_param", "cl_id=$clientID");
	getLockScreen($root, "cl_param", "cl_id=$clientID");

#
# limits
#
	if($pleskMajorVersion <= 2){
	  addClientLimits25($root, $clientID);
	}elsif($pleskMajorVersion <= 5){
	  addClientLimits($root, $clientParams{'lim_id'});
	}else{
	  addClientLimits($root, $client{'limits_id'});
	}

#
# end limits
#


	if($pleskMajorVersion <= 2){
	  addPermissions25($root, $clientID);
	} elsif($pleskMajorVersion <= 5){
	  addPermissions($root, $clientParams{'perm_id'});
	}else{
	  addPermissions($root, $client{'perm_id'});
	}
	addClientMulitpleLoginPermission($root, $clientID);
  }

#
# ip pool
#
  if($pleskMajorVersion <= 5){
    my ($domain,$ip);
    my $nodePool = &makeXmlNode('ip_pool');
    $root->{'ADDCHILD'}->($nodePool);

    foreach $domain (sort keys %domainsInfo){
      $ptrRow = $domainsInfo{$domain};
      if($ptrRow->[3]==$clientID){
		$ip = $ptrRow->[1];
		$id = $ptrIpHash->{$ip};
		if($id){
		  $item = &makeXmlNode('client_ip');
		  $item->{'ATTRIBUTE'}->('ip_ref','ip_'.$id);
		  $nodePool->{'ADDCHILD'}->($item);
		}
      }
    }
  }else{
    unless($id = $client{'pool_id'}){

      # client may not have a IP pool
      unless (0 == $id) {
		&printToError("Error: makeClientNode: pool ID is not found ($client{'pname'})");
      }

    }else{
      my($ip_ref);
      my $nodePool = &makeXmlNode('ip_pool');
      $root->{'ADDCHILD'}->($nodePool);

      if($full){
		$sql="SELECT r.component_id, i.ip_address FROM Repository r, IP_Addresses i".
		  " WHERE r.component_id=i.id AND r.rep_id = $id";
		if($wrapDbh->{'EXECUTE'}->($sql)){
		  while($ptrRow = $wrapDbh->{'FETCHROW'}->()){
			$ip_ref = $ptrRow->[0];
			$item = &makeXmlNode('client_ip');
			$item->{'ATTRIBUTE'}->('ip_ref','ip_'.$ip_ref);
			$nodePool->{'ADDCHILD'}->($item);
		  }
		}
		$wrapDbh->{'FINISH'}->();
      }else{
		#
		# get ip pool only for selected domains
		#
		my (%ids,$list,$ip);
		$list = join(',',keys %{$ptrDomainsHash});

		$sql = "SELECT DISTINCT val FROM dns_recs WHERE dom_id in ($list) AND type='A'";
		if($wrapDbh->{'EXECUTE'}->($sql)){
		  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
			$ip = $ptrRow->[0];
			if($id = $ptrIpHash->{$ip}){
			  $ids{$id}=$ip;
			}
		  }
		}
		$wrapDbh->{'FINISH'}->();

		$sql = "SELECT DISTINCT ip_address_id FROM forwarding WHERE dom_id in ($list)";
		if($wrapDbh->{'EXECUTE'}->($sql)){
		  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
			$id = $ptrRow->[0];;
			$ids{$id}=1;
		  }
		}
		$wrapDbh->{'FINISH'}->();

		$sql = "SELECT DISTINCT ip_address_id FROM hosting WHERE dom_id in ($list)";
		if($wrapDbh->{'EXECUTE'}->($sql)){
		  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
			$id = $ptrRow->[0];;
			$ids{$id}=1;
		  }
		}
		$wrapDbh->{'FINISH'}->();
		foreach $id (sort {$a <=> $b} keys %ids){
		  $item = &makeXmlNode('client_ip');
		  $item->{'ATTRIBUTE'}->('ip_ref','ip_'.$id);
		  $nodePool->{'ADDCHILD'}->($item);
		}
      }
    }
  }
#
# end ip pool
#

#
# prepare domain's list
#
  my($domainID,$domainName);
  if($full){
    $sql = "SELECT id,name FROM domains WHERE cl_id=$clientID ORDER BY id";
    if($wrapDbh->{'EXECUTE'}->($sql)){
      while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
		($domainID,$domainName)=@{$ptrRow};
		push @domains,[$domainID,$domainName];
      }
    }
	$wrapDbh->{'FINISH'}->();
  }else{
    foreach $domainID (sort {$a <=> $b} keys %{$ptrDomainsHash}){
      $ptrRow = $ptrDomainsHash->{$domainID};
      next unless $ptrRow->[1]==$clientID;
      push @domains,[$domainID,$ptrRow->[0]];
    }
  }
#
# end prepare domain's list
#

#
# loop by domain of the client
#
  foreach $ptrRow (@domains){
    ($domainID,$domainName)=@{$ptrRow};

    &printToLog("Domain '$domainName' is started");

    $objDumpStatus->{'DOMAIN'}->($domainName);
    $objDumpStatus->{'PRINT'}->();

    $item = &makeDomainNode($domainID,$ptrIpHash);
    if(ref($item)=~/HASH/){
      $root->{'ADDCHILD'}->($item);
      &printToLog("Domain '$domainName' is successfully dumped");
    }else{
      &printToLog("Domain '$domainName' is not dumped");
    }
  }
#
# end loop by domain of the client
#


  return $root;
}

sub addPermissions25 {
  my ($root, $id) = @_;

  # [42098] migration from 7.14 to 7.50 fails
  return unless ($id);

  my ($perm_name, $sql, $ptrRow, $value, $item);
  my @perm_names = ('create_domains', 'manage_dns', 'manage_log');

  foreach $perm_name (@perm_names) {
	$sql = "SELECT DISTINCT val FROM cl_param WHERE cl_id = '$id' AND param = '$perm_name' AND val = 'true'";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
	  $item = makeXmlNode('permission');
	  $item->{'ATTRIBUTE'}->('name', $perm_name);
	  $root->{'ADDCHILD'}->($item);
	}
	$wrapDbh->{'FINISH'}->();
  }

  $sql = "SELECT DISTINCT val FROM cl_param WHERE cl_id='$id' AND (param='ip_based_allow' OR param='nb_allow') AND val='true'";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	$item = makeXmlNode('permission');
	$item->{'ATTRIBUTE'}->('name', 'manage_phosting');
	$root->{'ADDCHILD'}->($item);
  }
  $wrapDbh->{'FINISH'}->();
}

sub addPermissions {
  my ($root, $id) = @_;
  my ($sql, $ptrRow, $item, $name);

  # [42098] migration from 7.14 to 7.50 fails
  return unless ($id);

  $sql = "SELECT permission FROM Permissions WHERE id=$id AND value='true'";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  $name = $ptrRow->[0];
	  next unless exists $xmlPermissions{$name};
	  $item = makeXmlNode('permission');
	  $item->{'ATTRIBUTE'}->('name', $ptrRow->[0]);
	  $root->{'ADDCHILD'}->($item);
	}
  }
  $wrapDbh->{'FINISH'}->();

  $sql = "SELECT value FROM Permissions WHERE id = $id and ("
	. "permission = 'ipb_allow' or permission = 'nb_allow' and value = 'true')";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	$item = makeXmlNode('permission');
	$item->{'ATTRIBUTE'}->('name', 'manage_phosting');
	$root->{'ADDCHILD'}->($item);
  }
  $wrapDbh->{'FINISH'}->();
}

sub addMultipleLoginPermission {
  my ($root, $name) = @_;
  my $item = makeXmlNode($name);
  $item->{'ATTRIBUTE'}->('name', 'multiple-sessions');
  $root->{'ADDCHILD'}->($item);
}

sub addDomainUserMultipleLoginPermission {
  my ($root, $id) = @_;

  my $sql = "SELECT val FROM dom_param WHERE dom_id=$id AND param='multiply_login' AND val='true'";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	addMultipleLoginPermission($root, 'permission');
  }
  $wrapDbh->{'FINISH'}->();
}

sub addClientMulitpleLoginPermission {
  my ($root, $id) = @_;

  my $sql = "SELECT val FROM cl_param WHERE cl_id =$id AND param='multiply_login' AND val='true'";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	addMultipleLoginPermission($root, 'permission');
  }
  $wrapDbh->{'FINISH'}->();
}

sub addMailUserMultipleLoginPermission {
  my ($root, $id) = @_;

  return unless atLeastVersion(6, 0, 0);

  my $sql = "SELECT val FROM mn_param WHERE mn_id=$id AND param='multiply_login' AND val='true'";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	addMultipleLoginPermission($root, 'mailuser-permission');
  }
}

sub insertLimitNode {
  my($root,$name,$value) = @_;
  my $item = &makeXmlNode('limit', $value);
  $item->{'ATTRIBUTE'}->('name', $name);
  $root->{'ADDCHILD'}->($item);
}

sub insertLimit {
  my($root,$name,$value) = @_;
  my($item);

  if($value eq '' || $value eq '-1') {
	insertLimitNode($root, $name, '-1');
	return;
  }

  if($name eq 'expiration'){
	my($mday,$mon,$year)=(localtime($value))[3..5];
	$mon++;
	$year+=1900;
	$value = sprintf('%04d-%02d-%02d',$year,$mon,$mday);
  }

  if($pleskMajorVersion <= 6 && $name eq 'mbox_quota') {
	$value = $value * 1024;
  }

  insertLimitNode($root, $name, $value);
}

sub addClientLimits {
  my ($root, $clientId) = @_;
  return addLimits($root, $clientId, \%clientLimits);
}

sub addClientLimits25 {
  my ($root, $clientId) = @_;
  return addLimits25($root, $clientId, \%clientLimits25, 'cl_id', 'cl_param');
}

sub addDomainLimits {
  my ($root, $clientId) = @_;
  return addLimits($root, $clientId, \%domainLimits);
}

sub addDomainLimits25 {
  my ($root,$domainId) = @_;
  addLimits25($root, $domainId, \%domainLimits25, 'dom_id', 'dom_param');
  #fix bug 50361
  my $ptrRow;
  my $sql ="SELECT traffic FROM hosting WHERE dom_id='$domainId'";
  if ($wrapDbh->{'EXECUTE'}->($sql) && ($ptrRow = $wrapDbh->{'FETCHROW'}->())){
  	insertLimit($root, 'max_traffic', $ptrRow->[0]);
	}
	$wrapDbh->{'FINISH'}->();
      $sql = "SELECT size FROM domains WHERE id='$domainId'";
  if ($wrapDbh->{'EXECUTE'}->($sql) && ($ptrRow = $wrapDbh->{'FETCHROW'}->())){

  	insertLimit($root, 'disk_space', $ptrRow->[0]);

	}
	$wrapDbh->{'FINISH'}->();
}

sub addLimits25 {
  my($root, $limitId, $limitNames, $idName, $paramName)=@_;

  if(!$limitId) {
	return;
  }

  my($value, $name, $count, $sql, $ptrRow);

  $sql = "SELECT val, param FROM $paramName WHERE $idName='$limitId'";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  ($name, $value) = @{$ptrRow};
	}
  }

  foreach $name (%{$limitNames}) {
	$sql = "SELECT val FROM $paramName WHERE $idName = '$limitId' AND param = '$name'";

	if(($wrapDbh->{'EXECUTE'}->($sql)) && ($ptrRow = $wrapDbh->{'FETCHROW'}->())) {
	  ($value) = @{$ptrRow};
	  if($name eq 'max_pbox') {
		$name = 'max_box';
	  }
	  insertLimit($root, $name, $value);
	}
	$wrapDbh->{'FINISH'}->();
  }
}

#
# sub addLimits
#
# arguments:
#           $root    - XML node to add 'limit' node
#           $limitID - limit id
#
# return:
#           $count - number of the added nodes
#
sub addLimits {
  my ($root, $limitId, $limitNames) = @_;

  if (!$limitId) {
	return;
  }

  my ($sql, $count, $ptrRow, $value, $name);
  $sql = "SELECT limit_name,value FROM Limits WHERE id=$limitId";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
    $count=0;
    while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
      ($name, $value) = @{$ptrRow};
      next unless exists $limitNames->{$name};
	  insertLimit($root, $name, $value);
      $count++;
    }
  }
  $wrapDbh->{'FINISH'}->();
  return $count;
}

#
# makeDomainMode
#
# arguments:
#           $domainID - domain id
#           $ptrIpHash - pointer to hash 'ip' => 'id'
#
# return:
#           $root - pointer to XML node
#
sub makeDomainNode {
  my($domainID,$ptrIpHash)=@_;

  my($sql,$item,$root,%domain,%domParams,$ptrRow,$ptrHash,
     $domainName,$ip,$id);

#
# get domain's info
#
  printToLog("Getting domain info");
  $sql = "SELECT * FROM domains WHERE id = $domainID";

  unless($wrapDbh->{'EXECUTE'}->($sql)){
    $wrapDbh->{'FINISH'}->();
    return undef;
  }
  unless($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
    $wrapDbh->{'FINISH'}->();
    return undef;
  }
  %domain=%{$ptrHash};
  $wrapDbh->{'FINISH'}->();

  $sql = "SELECT param,val FROM dom_param WHERE dom_id=$domainID";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
      $domParams{$ptrRow->[0]}=$ptrRow->[1];
    }
  }
  $wrapDbh->{'FINISH'}->();

#
# end get domain's info
#

  $root=&makeXmlNode('domain');

  if (atLeastVersion(7, 1, 0)) {
	$domainName = $domain{'displayName'};
  } else {
	$domainName = $domain{'name'};
  }

  addWwwStatus($root, $domainID, $domain{'name'});

  $root->{'ATTRIBUTE'}->('name',$domainName);

  my $domainType = $domain{'htype'};

  printToLog("Getting domain IP");
#
# domain's ip
#
  $id=undef;
  if ($pleskMajorVersion <= 5) {
    if ($ip = $domainsInfo{$domainName}->[1]) {
      $id = $ptrIpHash->{$ip};
    }
  } else {
    if($domainType=~/_fwd$/){
      $sql = "SELECT ip_address_id FROM forwarding WHERE dom_id = $domainID";
      if($wrapDbh->{'EXECUTE'}->($sql)){
		($id) = @{$wrapDbh->{'FETCHROW'}->()};
      }
    }else{
      $sql = "SELECT ip_address_id FROM hosting WHERE dom_id = $domainID";
      if($wrapDbh->{'EXECUTE'}->($sql)){
		($id) = @{$wrapDbh->{'FETCHROW'}->()};
      }
    }
    $wrapDbh->{'FINISH'}->();

    unless ($id) {
	  printToLog("Unable to get ip address from hosting/forwarding. Trying to " .
				 "get it from DNS.");

      $sql = "SELECT DISTINCT val FROM dns_recs WHERE dom_id=$domainID AND type='A'";
      if ($wrapDbh->{'EXECUTE'}->($sql)) {
		($ip) = @{$wrapDbh->{'FETCHROW'}->()};
		$id = $ptrIpHash->{$ip};
	  }
	  $wrapDbh->{'FINISH'}->();

	  unless(defined $id) {
		$id = $defaultIpId;
      }
    }
  }
  if($id){
    $item = &makeXmlNode('domain_ip');
    $item->{'ATTRIBUTE'}->('ip_ref','ip_'.$id);
    $root->{'ADDCHILD'}->($item);
  }

#
# end domain's ip
#

  # No further info required if 'no-content' specified
  if (exists $ptrArgs->{'no-content'}) {
	return $root;
  }

  printToLog("Getting domain limits");
#
# limits
#
#  $id = ($pleskMajorVersion <= 5) ? $domParams{'lim_id'} : $domain{'limits_id'};
#  if($id){
#    addDomainLimits($root,$id);
#  }
	if($pleskMajorVersion <= 2){
	addDomainLimits25($root,$domainID);
	} elsif($pleskMajorVersion <= 5){
	addDomainLimits($root,$domParams{'lim_id'});
	} else {
	addDomainLimits($root,$domain{'limits_id'});
	}

#
# end limits
#

  printToLog("Getting domain permissions");
#
# permissions
#
  if ($id = $domParams{'perm_id'}) {
    addPermissions($root,$id);
  }

#
# end permissions
#

  printToLog("Dumping domain mailsystem");
#
# mail system
#

  my $rootMail = &makeXmlNode('mailsystem');

  if (atLeastVersion(6, 0, 0)) {
	$sql = "SELECT status FROM DomainServices WHERE type='mail' "
	  . " AND dom_id=$domainID";

	if ($wrapDbh->{'EXECUTE'}->($sql) and my $ptrRow=$wrapDbh->{'FETCHROW'}->()) {
	  if ($ptrRow->[0] eq '0') {
		$rootMail->{'ATTRIBUTE'}->('enabled', 'on');
	  } else {
		$rootMail->{'ATTRIBUTE'}->('enabled', 'off');
	  }
	}
	$wrapDbh->{'FINISH'}->();
  }

  if($pleskMajorVersion <= 5){
    $sql = "SELECT id, password, 'plain' FROM mail ".
      " WHERE dom_id=$domainID ORDER BY mail_name";
  }else{
    $sql = "SELECT m.id, a.password, a.type FROM mail m".
      " LEFT JOIN accounts a ON m.account_id=a.id ".
      " WHERE m.dom_id=$domainID ORDER BY m.mail_name";
  }

  if ($wrapDbh->{'EXECUTE'}->($sql)){
    my(@mails);
    while($ptrRow = $wrapDbh->{'FETCHROW'}->()){
      push @mails,[@{$ptrRow}];
    }
    $wrapDbh->{'FINISH'}->();

    foreach $ptrRow (@mails){
	  $ptrRow->[2] = pwdType($ptrRow->[2]);
      $item = &makeMailUserNode(@{$ptrRow},$domainName);
      if(ref($item)=~/HASH/){
		$rootMail->{'ADDCHILD'}->($item);
      }
    }
  }

  getCatchAllAddress($rootMail, $domainID);

  $root->{'ADDCHILD'}->($rootMail);
#
# end mail system
#

  printToLog("Dumping domain DNS");
#
# dns
#
  $sql = "SELECT * FROM dns_recs WHERE dom_id=$domainID";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
    while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
      next if ($ptrHash->{'type'} =~ /none/);
	  if ($ptrHash->{'type'} eq 'TXT') {
		$ptrHash->{'val'} =~ s/"(.*)"/$1/;
	  }
      $item = &makeXmlNode('dnsrec');
      $item->{'ATTRIBUTE'}->('type', $ptrHash->{'type'});
      $item->{'ATTRIBUTE'}->('src', $ptrHash->{'host'});
      $item->{'ATTRIBUTE'}->('dst', $ptrHash->{'val'});
	  $item->{'ATTRIBUTE'}->('opt', $ptrHash->{'opt'}) if ($ptrHash->{'opt'});
      $root->{'ADDCHILD'}->($item);
    }
  }
  $wrapDbh->{'FINISH'}->();
#
# end dns
#

  printToLog("Dumping domain databases");
#
# database
#
  $sql = "SELECT id,name,type FROM data_bases WHERE dom_id=$domainID ORDER BY id";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    my (@databases);
    while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
      push @databases,[@{$ptrRow}];
    }
    $wrapDbh->{'FINISH'}->();
    foreach $ptrRow (@databases){
      $item=&makeDatabaseNode(@{$ptrRow});
      if(ref($item)=~/HASH/){
		$root->{'ADDCHILD'}->($item);
      }
    }
  }
  $wrapDbh->{'FINISH'}->();
#
# end database
#

  printToLog("Dumping domain maillists");
#
# maillist
#
  unless($pleskMajorVersion <= 5){
    $sql="SELECT id,name,status FROM MailLists WHERE dom_id=$domainID";
    if($wrapDbh->{'EXECUTE'}->($sql)){
      my (@mlists);
      while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
		push @mlists,[@{$ptrRow}];
      }
      $wrapDbh->{'FINISH'}->();
      foreach $ptrRow (@mlists){
		$item=&makeMailListNode(@{$ptrRow});
		if(ref($item)=~/HASH/){
		  $root->{'ADDCHILD'}->($item);
		}
      }
    }
	$wrapDbh->{'FINISH'}->();
  }
#
# end maillist
#

  printToLog("Dumping domain statistics");
#
# traffic
#
  &addCurrentTraffic($root,$domainID);
#
# end traffic
#

  printToLog("Dupming domain status");
  my $status;

  if ($pleskMajorVersion < 5 || ($pleskMajorVersion == 5 && $pleskMinorVersion == 0)) {
	if ($domain{'admin_status'} eq 'false') {
	  $status = 'false';
	} elsif ($domain{'status'} eq 'false') {
	  $status = 'disabled-by-user';
	}
  } else {
	if ($domain{'status'} eq '16') {
	  $status = 'disabled-by-user';
	} elsif ($domain{'status'} eq '64') {
	  $status = 'false';
	} elsif (($domain{'status'} eq '2')) {
	  $status = 'disabled-by-parent';
	}
  }

  if (defined $status) {
    $item = &makeXmlNode('status', $status);
    $root->{'ADDCHILD'}->($item);
  }

  printToLog("Dumping domain certificates");
#
# certificates
#
  if($pleskMajorVersion <= 5){
    $sql = "SELECT id FROM certificates WHERE dom_id=$domainID";
  }else{
    $sql = "SELECT c.id FROM certificates c, Repository r ".
      "WHERE c.id=r.component_id AND r.rep_id='".$domain{'cert_rep_id'}.
      "' ORDER BY c.id";
  }
  if($wrapDbh->{'EXECUTE'}->($sql)){
    my (@ids);
    while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
      push @ids,$ptrRow->[0];
    }
    $wrapDbh->{'FINISH'}->();

    foreach $id (@ids){
      $item = &makeCertificateNode($id);
      if(ref($item)=~/HASH/){
		$root->{'ADDCHILD'}->($item);
      }
    }
  }
#
# end certificates
#

  printToLog("Duming domain user");
#
# domainuser
#
  $item=&makeDomainUserNode($domainID);
  if(ref($item)=~/HASH/){
    $root->{'ADDCHILD'}->($item);
  }
#
# end domainuser
#

  printToLog("Dumping domain hosting");
#
# hosting
#
  $item=undef;
  if($domainType=~/vrt_hst/){
    $item = &makePhostingNode(\%domain);
  }elsif($domainType=~/std_fwd/){
    $item = &makeShostingNode(\%domain);
  }elsif($domainType=~/frm_fwd/){
    $item = &makeFhostingNode(\%domain);
  }
  if(ref($item)=~/HASH/){
    $root->{'ADDCHILD'}->($item);
  }

#
# end hosting
#

  return $root;
}

sub addWwwStatus {
  my ($parent, $domainID, $domainName) = @_;

  my $sql = "SELECT * FROM dns_recs WHERE dom_id = $domainID AND type = 'CNAME' " .
	"AND host = 'www.$domainName.'";

  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	$parent->{'ATTRIBUTE'}->('www', 'true');
  } else {
	$parent->{'ATTRIBUTE'}->('www', 'false');
  }
  $wrapDbh->{'FINISH'}->();
}

sub makeDomainUserNode {
  my($domainId)=@_;

  my($sql,$root,$item,$ptrHash,$xmlName,$fieldName);

  $sql="SELECT * FROM dom_level_usrs WHERE dom_id=$domainId";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    return undef;
  }
  my %domainUser=%{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();
  $root = &makeXmlNode('domainuser');

  {
	  # [Bug 44920] - domain user table changes
	  my $du_status_field = atLeastVersion(8,0,0) ? 'status' : 'state';
	  if ($domainUser{$du_status_field}) {
		  $root->{'ATTRIBUTE'}->('status', $domainUser{$du_status_field});
	  }
  }

  if (atLeastVersion(7, 1, 5)) {
	$root->{'ADDCHILD'}->(makeAccountPasswordNode($domainUser{'account_id'}));
  } else {
	$root->{'ADDCHILD'}->(makePasswordNode($domainUser{'passwd'}, 'plain'));
  }

  if($pleskMajorVersion >= 5 && defined($domainUser{'card_id'})) {
	$sql = "SELECT * from Cards WHERE id=$domainUser{'card_id'}";
	if($wrapDbh->{'EXECUTE'}->($sql)){
	  $ptrHash = $wrapDbh->{'FETCHHASH'}->();
	  while (($xmlName,$fieldName)=each %domainUserAttribute){
		if($ptrHash->{$fieldName}){
		  $root->{'ATTRIBUTE'}->($xmlName,$ptrHash->{$fieldName});
		}
	  }

	  while (($xmlName, $fieldName) = each %domainUserInfo) {
		my $value = $ptrHash->{$fieldName};
		if ($value) {
		  $value = normalizeCountry($value) if $fieldName eq 'country';
		  $item = makeXmlNode('pinfo', $value);
		  $item->{'ATTRIBUTE'}->('name', $xmlName);
		  $root->{'ADDCHILD'}->($item);
		}
	  }
	}
	$wrapDbh->{'FINISH'}->();
  }

  if (atLeastVersion(7, 1, 0)) {
	my ($ptrRow);
	$sql = "SELECT val FROM dom_param WHERE dom_id = $domainId AND param = 'locale'";
	if ($wrapDbh->{'EXECUTE'}->($sql) && ($ptrRow = $wrapDbh->{'FETCHROW'}->())) {
	  $item = makeXmlNode('pinfo', $ptrRow->[0]);
	  $item->{'ATTRIBUTE'}->('name', 'locale');
	  $root->{'ADDCHILD'}->($item);
	}
  }

  if (atLeastVersion(7, 1, 0)) {
	addPermissions($root, $domainUser{'perm_id'});
  }
  addDomainUserMultipleLoginPermission($root, $domainId);

  # max button length
  getMaxButtonLength($root, "dom_param", "dom_id=$domainId");
  getSkin($root, "dom_param", "dom_id=$domainId");
  getLockScreen($root, "dom_param", "dom_id=$domainId");

  return $root;
}


#
# addCurrentTraffic - add current traffic
#
#     arguments:
#                 $root - XML node to add traffic's nodes
#                 $domainId - ID of domain
#
sub addCurrentTraffic {
  my ($root, $domainId) = @_;
  my ($sql, $ptrRow, $item);

  my $trafficValue = '';

  if ($pleskMajorVersion <= 5) {
	$sql = "SELECT transfer, date FROM stat WHERE dom_id=$domainId";
    if ($wrapDbh->{'EXECUTE'}->($sql)) {
      while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		if ($ptrRow->[0]) {
		  $trafficValue .= $ptrRow->[1];
		  $trafficValue .= ' http in ';
		  $trafficValue .= $ptrRow->[0];
		  $trafficValue .= "\n";
		}
	  }
    }
	$wrapDbh->{'FINISH'}->();
  } else {
	$sql = "SELECT http_in, http_out, ftp_in, ftp_out, smtp_in, smtp_out, ".
	  "pop3_imap_in, pop3_imap_out, date FROM DomainsTraffic WHERE dom_id=$domainId";

    if ($wrapDbh->{'EXECUTE'}->($sql)) {
      my @keys = ('http', 'ftp', 'smtp', 'pop3-imap');
      my ($key,$i);
      while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		for($i = 0; $i < @keys*2; ++$i) {
		  if($ptrRow->[$i]) {
			$trafficValue .= $ptrRow->[8];
			$trafficValue .= ' ';
			$trafficValue .= $keys[$i/2];
			$trafficValue .= ' ';
			$trafficValue .= ($i%2 ? 'out' : 'in');
			$trafficValue .= ' ';
			$trafficValue .= $ptrRow->[$i];
			$trafficValue .= "\n";
		  }
		}
      }
    }
	$wrapDbh->{'FINISH'}->();
  }

  $root->{'ADDCHILD'}->(makeXmlNode('traffic', $trafficValue));
}


# There are 3 types of nonexistent user mail handling:
# Bounce with message (bounce:message text)
# Catch to address (email@address)
# SMTP reject (reject)
# Returns bounce|catch|reject or empty string
sub getNonexistentMode
{
	my ($domainId) = @_;

	my $sql = "SELECT p.value FROM Parameters p, DomainServices ds " .
	"WHERE ds.parameters_id = p.id AND ds.dom_id = $domainId AND p.parameter = 'nonexist_mail'";
	my $mode;
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
		$mode = @{$wrapDbh->{'FETCHROW'}->()}->[0];
	}
	$wrapDbh->{'FINISH'}->();

	return $mode;
}

sub getCatchAllAddress
{
	my ($parent, $domainId) = @_;

	my $sql;
	if (atLeastVersion(6, 0, 0)) {
		my $mode = getNonexistentMode($domainId);

		if ($mode =~ /^catch$/) {
			$sql = "SELECT p.value FROM Parameters p, DomainServices ds " .
			"WHERE ds.parameters_id = p.id AND ds.dom_id = $domainId AND p.parameter = 'catch_addr'";

		} elsif ($mode =~ /^bounce$/) {
			$sql = "SELECT CONCAT('bounce:', p.value) FROM Parameters p, DomainServices ds " .
			"WHERE ds.parameters_id = p.id AND ds.dom_id = $domainId AND p.parameter = 'bounce_mess'";

		} elsif ($mode =~ /^reject$/) {
			$sql = "SELECT 'reject'";
		}

	} elsif (atLeastVersion(5, 0, 0)) {
		$sql = "SELECT p.value FROM Parameters p, dom_param dp WHERE " .
		"dp.dom_id = $domainId AND dp.param='param_id' AND dp.val = p.id AND p.parameter = 'catch_addr'";

	} else {
		$sql = "SELECT dp.val FROM dom_param dp WHERE dp.dom_id = $domainId AND dp.param = 'catch_addr'";
	}

	# some not supported mode or default parameter [bug 43901]
	return unless ($sql);

	if ($wrapDbh->{'EXECUTE'}->($sql)) {
		my $catchAllAddr = @{$wrapDbh->{'FETCHROW'}->()}->[0];
		if ($catchAllAddr) {
			$parent->{'ADDCHILD'}->(makeXmlNode('catch-all', $catchAllAddr));
		}
	}
	$wrapDbh->{'FINISH'}->();
}


sub getMaillistPassword {
  my $mlistName = shift;
  my $script =<<EOS
import sys
import string

from Mailman import mm_cfg
from Mailman import MailList
from Mailman import Errors

def main():
    listname = '$mlistName'
    listname = string.lower(listname)
    try:
        mlist = MailList.MailList(listname, lock=0)
	print mlist.password
    except Errors.MMListError, e:
	print '$mlistName is not found'
	sys.exit(1)

if __name__ == '__main__':
    main()

EOS
;

 $ENV{'PYTHONPATH'}=$PsaConfig{'MAILMAN_ROOT_D'};
 my $python = $bin{'python'} || 'python';
 if ($PsaConfig{'PYTHON_ROOT_D'} ne '')  {
     $python = $PsaConfig{'PYTHON_ROOT_D'} . '/bin/python';
  }

  unless(open(PWD,"$python -c \"$script\" |")) {
       return '';
  }
  binmode PWD;
  my $line = <PWD>;
  chomp $line;
  close (PWD);
  return $line;

}


#
# &makeMailListNode
#
#   arguments:
#              $mlistId - ID of the mail list
#              $mlistName - name
#              $mllistState - state of maillist
#
#   return:
#              $root - XMl node
#
sub makeMailListNode {
  my($mlistId,$mlistName,$mlistState)=@_;
  my($cmd,$item, $root);

  unless (-d $PsaConfig{'MAILMAN_ROOT_D'}){
    &printToLog("Warning: makeMailListNode: directory of mailman '$PsaConfig{'MAILMAN_ROOT_D'}' is not found");
    return undef;
  }

  $root=&makeXmlNode('maillist');
  $root->{'ATTRIBUTE'}->('name',$mlistName);
  if ($mlistState){
    $root->{'ATTRIBUTE'}->('state','off');
  }
    # BUGFIX 48718. Maillists are not transferring from Plesk with mailman < 2.1
    my $script =<<EOS

import sys
import getopt
import string

from Mailman import Utils
from Mailman.MailList import MailList

PROGRAM = sys.argv[0]

try:
    True, False
except NameError:
    True = 1
    False = 0

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'wmh',
                                   ['with-listnames', 'moderators', 'help'])
    except getopt.error, msg:
        usage(1, msg)

    withnames = moderators = False
    for opt, arg in opts:
        if opt in ('-h', '--help'):
            usage(0)
        elif opt in ('-m', '--moderators'):
            moderators = True
        elif opt in ('-w', '--with-listnames'):
            withnames = True

    listnames = args or Utils.list_names()
    bylist = {}

    for listname in listnames:
	listname = string.lower(listname)
        mlist = MailList(listname, lock=0)
        addrs = mlist.owner[:]
        if moderators:
            addrs.extend(mlist.moderator)
        bylist[listname] = addrs

    if withnames:
        for listname in listnames:
	    listname = string.lower(listname)
            unique = {}
            for addr in bylist[listname]:
                unique[addr] = 1
            keys = unique.keys()
            keys.sort()
            print listname
            for k in keys:
                print '\t', k
    else:
        unique = {}
        for listname in listnames:
	    listname = string.lower(listname)
            for addr in bylist[listname]:
                unique[addr] = 1
        keys = unique.keys()
        keys.sort()
        for k in keys:
            print k



if __name__ == '__main__':
    main()

EOS
;
    $ENV{'PYTHONPATH'}=$PsaConfig{'MAILMAN_ROOT_D'};
    my $python = $bin{'python'} || 'python';
    if ($PsaConfig{'PYTHON_ROOT_D'} ne '') {
        $python = $PsaConfig{'PYTHON_ROOT_D'} . '/bin/python';
    }
    if (open(LIST,"$python -c \"$script\" " . $mlistName . " |")) {
    	binmode LIST;
	while (my $line = <LIST>) {
	  chomp $line;
	  $item = &makeXmlNode('owner', $line);
	  $root->{'ADDCHILD'}->($item);
	}
	close(LIST);
  }
  # BUGFIX #39286: maillist's admin password after migration is lost
  # Getting cyrpted password from mailman for every list and put password
  # to dump file

  my $passwd = &getMaillistPassword($mlistName);
  $item = &makePasswordNode($passwd, 'encrypted');
  $root->{'ADDCHILD'}->($item);

  $cmd = $PsaConfig{'MAILMAN_ROOT_D'} . '/bin/list_members';
  if (-x $cmd) {
	$cmd .= ' ' . $mlistName;

	if (open(LIST,"$cmd |")) {
	  binmode LIST;
	  while (my $line = <LIST>) {
		chomp $line;
		if ($line ne $mlistName) {
		  $item = &makeXmlNode('recipient', $line);
		  $root->{'ADDCHILD'}->($item);
		}
	  }
	  close(LIST);
	}
  }


  return $root;
}


sub addUrlDecodedTextNode {
  my ($parent, $name, $value) = @_;

  $parent->{'ADDCHILD'}->(makeXmlNode($name, urlDecode($value))) if ($value);
}

sub makeCertificateNode {
  my($certId)=@_;

  my($sql,$root,%cert,$item);

  $sql = "SELECT * FROM certificates WHERE id=$certId";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    &printToError("Error: makeCertificateNode: certificate ID '$certId' is not found");
    return undef;
  }

  %cert= %{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();

  $root = makeXmlNode('certificate');

  if($pleskMajorVersion <= 5){
	addUrlDecodedTextNode($root, 'certificate-data', $cert{'pub_key'});
  }else{
	addUrlDecodedTextNode($root, 'certificate-data', $cert{'cert'});
	addUrlDecodedTextNode($root, 'signing-request', $cert{'csr'});
	addUrlDecodedTextNode($root, 'ca-certificate', $cert{'ca_cert'});
  }
  addUrlDecodedTextNode($root, 'private-key', $cert{'pvt_key'});
  $root->{'ATTRIBUTE'}->('name', $cert{'name'});

  return $root;
}

my(%PostgreSqlAdmin);

sub makeDatabaseNode {
  my ($dbId, $dbName, $dbTypePlesk) = @_;
  my ($root, $sql, $ptrHash, $ptrRow, $dbHost, $dbType, $item);

  printToLog(">Dumping " . $dbName);

  if ($dbTypePlesk =~ /postgres/) {
	$dbTypePlesk = "postgresql";
  }

  $root = makeXmlNode('database');
  $root->{'ATTRIBUTE'}->('name', $dbName);
  $root->{'ATTRIBUTE'}->('type', $dbTypePlesk);

  printToLog("Adding db users");

  if (atLeastVersion(7, 5, 0)) {
	$sql = "SELECT login, password, type FROM db_users, accounts WHERE "
	  . "db_users.db_id = $dbId AND db_users.account_id = accounts.id";

	$wrapDbh->{'EXECUTE'}->($sql);
	while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
      my $dbUserRoot = makeXmlNode('dbuser');
      $dbUserRoot->{'ATTRIBUTE'}->('name',$ptrHash->{'login'});

	  $item = makePasswordNode($ptrHash->{'password'}, pwdType($ptrHash->{'type'}));

	  $dbUserRoot->{'ADDCHILD'}->($item);
      $root->{'ADDCHILD'}->($dbUserRoot);
	}
  } else {
	$wrapDbh->{'EXECUTE'}->("SELECT login, passwd FROM db_users WHERE db_id = $dbId");

	while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
	  my $dbPasswdType;

	  # In Plesk 7.1.5 DB password type wasn't described in the DB. We
	  # can distinguish the type only using length of password.
	  if (length($ptrHash->{'passwd'}) > 14) {
		$dbPasswdType = 'encrypted';
	  } else {
		$dbPasswdType = 'plain';
	  }

      my $dbUserRoot = makeXmlNode('dbuser');
      $dbUserRoot->{'ATTRIBUTE'}->('name', $ptrHash->{'login'});

	  $item = makePasswordNode($ptrHash->{'passwd'}, $dbPasswdType);

	  $dbUserRoot->{'ADDCHILD'}->($item);
      $root->{'ADDCHILD'}->($dbUserRoot);
	}
  }
  $wrapDbh->{'FINISH'}->();

  printToLog("Determining DB connection settings");
#
# dump
#
  my $dbPasswd;
  if($dbTypePlesk=~/mysql/){

    $dbType = 'mysql';
    $dbUser = $wrapDbh->{'USER'}->();
    $dbPasswd = $wrapDbh->{'PASSWORD'}->();
    $dbHost = $wrapDbh->{'HOST'}->();

  }elsif($dbTypePlesk=~/postgresql/){

    $dbType = 'Pg';
    unless($dbUser && defined($dbPasswd)){ ## if is not any user try to work with admin-account
      unless(keys %PostgreSqlAdmin){
		$sql = "SELECT param, val FROM misc WHERE param like 'postgresql_admin_\%'";
		if($wrapDbh->{'EXECUTE'}->($sql)){
		  while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
			if($ptrRow->[0]=~/login$/){
			  $PostgreSqlAdmin{'user'}=$ptrRow->[1];
			}elsif($ptrRow->[0]=~/passwd$/){
			  $PostgreSqlAdmin{'password'}=$ptrRow->[1];
			}
		  }
		}
		$wrapDbh->{'FINISH'}->();
      }
      $dbUser = $PostgreSqlAdmin{'user'};
      $dbPasswd = $PostgreSqlAdmin{'password'};
    }

  }

  printToLog("Dumping");

  if($dumpFile = &makeDumpDb($dbName,$dbType,$dbUser,$dbPasswd,$dbHost)){
    $root->{'ATTRIBUTE'}->('cid',$dumpFile);
  }
#
# end dump
#


  return $root;
}

#
# makeMailUserNode
#
# argumets:
#           $mailId - mail ID
#           $passwd - password
#           $typePasswd - type of password
#           $domainName - name of domain
#
# return:
#           $root = XML node 'mailuser'
#
sub makeMailUserNode {
  my($mailId,$passwd,$typePasswd,$domainName)=@_;

  my($sql,%mail,$item,$mailName,$ptrRow,$ptrHash,$dir,$id, $mbox_quota);

  $sql = "SELECT * FROM mail WHERE id = $mailId";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    &printToError("Error: makeMailUserNode: mail ID '$mailId' is not found");
    return undef;
  }
  %mail = %{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();
  my $root = &makeXmlNode('mailuser');
  $mailName = $mail{'mail_name'};
  $root->{'ATTRIBUTE'}->('name',$mailName);

  if(defined($passwd) && $passwd){
	if($passwd =~ /NULL/) {
	  $item = &makePasswordNode('', 'empty');
	} else {
	  $item = &makePasswordNode($passwd,$typePasswd);
	}

    $root->{'ADDCHILD'}->($item);
  }

  if ($mail{'cp_access'} eq 'true') {
	makePermissionNode($root, 'mailuser', 'cp-access');
  }

  # multiple login
  addMailUserMultipleLoginPermission($root, $mailId);

  $root->{'ATTRIBUTE'}->('autoresponders-enabled', $mail{'autoresponder'});

  my $enable_mailbox = $mail{'postbox'} =~ /true/;

  # If mbox quota set up, we want to create mailbox node
  # even if mailbox is disabled
  $enable_mailbox ||= $mail{'mbox_quota'};

  # Check whether there autoresponder with attach
  # On 'Olde Pleskes' there was bug allowing attaches
  # when mailbox in turned off, so we have to explicitly
  # turn mailbox on if there is attach.
  $sql = "SELECT COUNT(ra.filename) FROM mail_resp AS mr, resp_attach as ra
            WHERE ra.rn_id = mr.id AND mr.mn_id = $mailId";
  if ($wrapDbh->{'EXECUTE'}->($sql)
	  and ($ptrRow = $wrapDbh->{'FETCHROW'}->())
	  and $ptrRow->[0] != 0) {
	$enable_mailbox = 1;
  }

  if ($enable_mailbox) {
    $item = &makeXmlNode('mailbox');
    $item->{'ATTRIBUTE'}->('type','mdir');
    $root->{'ADDCHILD'}->($item);

	if ($mail{'postbox'} =~ /true/) {
	  $item->{'ATTRIBUTE'}->('enabled', 'true');
	} else {
	  $item->{'ATTRIBUTE'}->('enabled', 'false');
	}

    $dir=$PsaConfig{'QMAIL_MAILNAMES_D'}."/$domainName/$mailName/Maildir";
    $dumpFile = &makeDumpFile("$dumpDir/$mailName\@$domainName.mdir",
			     $dir);
    if($dumpFile){
      $item->{'ATTRIBUTE'}->('cid',$dumpFile);
    }

	$dir=$PsaConfig{'QMAIL_MAILNAMES_D'}."/$domainName/$mailName/\@attachments";
	$dumpFile = &makeDumpFile("$dumpDir/$mailName\@$domainName.attach",
							  $dir);
	if($dumpFile){
	  $root->{'ATTRIBUTE'}->('cid_attach',$dumpFile);
	}

    if($mail{'mbox_quota'}){
	  $mbox_quota = $mail{'mbox_quota'};

	  if($pleskMajorVersion <= 6 && $mbox_quota ne "-1") {
		$mbox_quota = $mbox_quota * 1024;
	  }

      $item->{'ATTRIBUTE'}->('quota',$mbox_quota);
    }
  }

#
# aliases
#
  if($pleskMajorVersion >= 5) {
	$sql="SELECT alias FROM mail_aliases WHERE mn_id=$mailId";
	if($wrapDbh->{'EXECUTE'}->($sql)){
	  while ($ptrRow=  $wrapDbh->{'FETCHROW'}->()){
		$item=&makeXmlNode('alias',$ptrRow->[0]);
		$root->{'ADDCHILD'}->($item);
	  }
	}
	$wrapDbh->{'FINISH'}->();
  }
#
# end aliases
#

#
# redirect
#
  if ($mail{'redir_addr'}) {
    $item = makeXmlNode('redirect',$mail{'redir_addr'});
	$item->{'ATTRIBUTE'}->('enabled', $mail{'redirect'});
    $root->{'ADDCHILD'}->($item);
  }
#
# end redirect
#

#
# mail group
#
  $root->{'ATTRIBUTE'}->('mailgroup-enabled', $mail{'mail_group'});

  $sql="SELECT address FROM mail_redir WHERE mn_id=$mailId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
	while ($ptrRow=  $wrapDbh->{'FETCHROW'}->()){
	  $item=&makeXmlNode('mailgroup-member',$ptrRow->[0]);
	  $root->{'ADDCHILD'}->($item);
	}
  }
  $wrapDbh->{'FINISH'}->();

#
# end mail group
#

#
# autoresponders
#
# ALL autoresponders should be included in dump - not just if they are enabled.
#  if($mail{'autoresponder'}=~/true/)
  {
    my(@autos);
    $sql = "SELECT id FROM mail_resp WHERE mn_id=$mailId ORDER BY id";
    if($wrapDbh->{'EXECUTE'}->($sql)){
      while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
		push @autos,$ptrRow->[0];
	  }
    }
    $wrapDbh->{'FINISH'}->();
    foreach $id (@autos){
      $item = &makeAutoresponderNode($id, $mailName . "@" . $domainName);
      if(ref($item)=~/HASH/){
		$root->{'ADDCHILD'}->($item);
      }
    }
  }

#
# end autoresponders
#

#
# Addressbook
#
  if (atLeastVersion(6, 0, 0)) {
	makeAddressbookNode($root, $mailName . '@' . $domainName);
  }

#
# Spamassassin
#
  if ($pleskMajorVersion == 7) {
  	makeSpamassassinNode($root,$mail{'mail_name'}.'@'.$domainName);
  }
#
# End Spamassassin
#

#
# Drweb
#
if (-e "/etc/drweb/users.conf"){
  	$item=&makeDrwebNode($root,$mailName.'@'.$domainName);
  }
#
# End Drweb
#

  if (atLeastVersion(6, 0, 0)) {
	getMaxButtonLength($root, "mn_param", "mn_id=$mailId");
	getSkin($root, "mn_param", "mn_id=$mailId");
	getLockScreen($root, "mn_param", "mn_id=$mailId");
  }

  return $root;
}

#
# makeAutoresponderNode
#
#  arguments:
#             $id - ID of autoreponder
#
#  return:
#             $root - XML node
#
sub makeAutoresponderNode {
  my($autoId, $mailName)=@_;

  my($name,$value,$sql,%auto,$ptrRow,$item);

  $sql = "SELECT * FROM mail_resp WHERE id=$autoId";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    &printToError("Error: makeAutoresponderNode: autoresponder # $autoId is not found");
    return undef;
  }

  %auto=%{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();

  my $root = &makeXmlNode('autoresponder_ext');

  $value=$wrapBase64->{'ENCODE'}->($auto{'text'});
  $item=&makeXmlNode('text',$value);
  if (exists $auto{'charset'} && $auto{'charset'} ne '') {
	$item->{'ATTRIBUTE'}->('charset', $auto{'charset'});
  }
  $root->{'ADDCHILD'}->($item);

  if (exists $auto{'content_type'} && $auto{'content_type'}) {
	$root->{'ATTRIBUTE'}->('content-type', $auto{'content_type'});
  }

  $name = $auto{'resp_name'};
  $root->{'ATTRIBUTE'}->('name',$name);

  if ($auto{'resp_on'}=~/false/){
    $root->{'ATTRIBUTE'}->('status','off');
  }

  unless ($auto{'key_where'}=~/no/){
    $value=$auto{'key_where'}.':'.$auto{'keystr'};
    $root->{'ATTRIBUTE'}->('require',$value);
  }

  if($auto{'subject'}){
    $value = $wrapBase64->{'ENCODE'}->($auto{'subject'});
    chomp $value;
    $root->{'ATTRIBUTE'}->('subject',$value);
  }

  if ($auto{'reply_to'} and $auto{'reply_to'} ne $mailName) {
	$root->{'ATTRIBUTE'}->('replyto',$auto{'reply_to'});
  }

#
# forward
#
  $sql="SELECT address FROM resp_forward WHERE rn_id=$autoId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    my (@list);
    while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
      push @list,$ptrRow->[0];
    }
    $root->{'ATTRIBUTE'}->('redirect',join(',',@list));
  }
  $wrapDbh->{'FINISH'}->();
#
# end forward
#
  $item = &makeXmlNode('limit',$auto{'ans_freq'});
  $item->{'ATTRIBUTE'}->('name','ans_freq');
  $root->{'ADDCHILD'}->($item);

  $item = &makeXmlNode('limit',$auto{'mem_limit'});
  $item->{'ATTRIBUTE'}->('name','ans_count');
  $root->{'ADDCHILD'}->($item);


#
# attachment
#
  $sql = "SELECT filename FROM resp_attach WHERE rn_id=$autoId";
  if($wrapDbh->{'EXECUTE'}->($sql)){

    while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
      $item=&makeXmlNode('attach');
      $item->{'ATTRIBUTE'}->('file',$ptrRow->[0]);
      $root->{'ADDCHILD'}->($item);
    }
  }
  $wrapDbh->{'FINISH'}->();
#
# end attachment
#

  return $root;
}


sub makeAddressbookNode {
  my ($parent, $mailName) = @_;

  my $sql = "SHOW TABLES FROM horde LIKE 'turba_objects'";
  return unless $wrapDbh->{'EXECUTE'}->($sql);
  $wrapDbh->{'FINISH'}->();

  my $node = makeXmlNode('addressbook');

  my %objectFields = ('id'=>'id',
					  'name'=>'name',
					  'alias'=>'alias',
					  'email'=>'email',
					  'title'=>'title',
					  'company'=>'company',
					  'home-phone'=>'homephone',
					  'work-phone'=>'workphone',
					  'mobile-phone'=>'cellphone',
					  'fax'=>'fax',
					  'home-address'=>'homeaddress',
					  'work-address'=>'workaddress',
					  'notes'=>'notes');

  my $ptrHash;
  $sql = "SELECT * FROM horde.turba_objects WHERE owner_id = '$mailName'";
  $wrapDbh->{'EXECUTE'}->($sql);

  while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()) {
	if ($ptrHash->{'object_type'} eq 'Group') {
	  my $group = makeXmlNode('addressbook-group');
	  $group->{'ATTRIBUTE'}->('id', $ptrHash->{'object_id'});

	  my $members = $ptrHash->{'object_members'};
	  while ($members =~ /"(.*?)"/g) {
		my $member = $1;
		chomp $member;

		my $memberNode = makeXmlNode('addressbook-member');
		$memberNode->{'ATTRIBUTE'}->('id', $member);
		$group->{'ADDCHILD'}->($memberNode);
	  }

	  $node->{'ADDCHILD'}->($group);
	} else {
	  my $user = makeXmlNode('addressbook-contact');
	  my ($field, $attr);
	  while (($field, $attr) = each %objectFields) {
		if (defined $ptrHash->{'object_' . $attr}) {
          # Here 'encoded' attribute means there is no meaning to
          # convert data to UTF as it was already converted manually
		  my $value = $ptrHash->{'object_' . $attr};
		  $value = $wrapBase64->{'ENCODE'}->($value) if ($attr ne 'id');
		  $user->{'ATTRIBUTE'}->($field, $value);
		}
	  }
	  $node->{'ADDCHILD'}->($user);
	}
  }

  $parent->{'ADDCHILD'}->($node);
}

sub makePhostingNode {
  my ($ptrDomain)=@_;
  unless(ref($ptrDomain)=~/HASH/){
    &printToError("Error: makePhostNode: bag arguments");
    return undef;
  }
  my $root=&makeXmlNode('phosting');

  my($domainName,$domainRoot,$path,$fileName,$sql,%hosting,$domainId,
    $xmlName,$fieldName,$id,$item,$ptrRow,$ptrHash);
  $domainName = $ptrDomain->{'name'};
  $domainId = $ptrDomain->{'id'};

  my $shared_content;

  # shared http/https content
  if (atLeastVersion(7, 5, 3)) {
	my $sql = "SELECT same_ssl FROM hosting WHERE dom_id=$domainId";
	if ($wrapDbh->{'EXECUTE'}->($sql) and my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  if ($ptrRow->[0] eq 'true') {
		$shared_content = 1;
		$root->{'ATTRIBUTE'}->('shared-content', 'true');
	  }
	}
  }

  $domainRoot=$PsaConfig{'HTTPD_VHOSTS_D'}.'/'.$domainName;
  if (-d $domainRoot) {
    if($fileName = &makeDumpFile("$dumpDir/$domainName.html",
				 "$domainRoot/httpdocs")){
      $root->{'ATTRIBUTE'}->('cid_docroot',$fileName);
    }
    if(!$shared_content and $fileName = &makeDumpFile("$dumpDir/$domainName.shtml",
				 "$domainRoot/httpsdocs")){
      $root->{'ATTRIBUTE'}->('cid_docroot_ssl',$fileName);
    }
    if($fileName = &makeDumpFile("$dumpDir/$domainName.cgi",
				 "$domainRoot/cgi-bin")){
      $root->{'ATTRIBUTE'}->('cid_cgi',$fileName);
    }

	if (-d $domainRoot . "/statistics/webstat") {
	  if ($fileName = makeDumpFile("$dumpDir/$domainName.webstat",
								   "$domainRoot/statistics/webstat")) {
		$root->{'ATTRIBUTE'}->('cid_webstat', $fileName);
	  }
	}

	if (-d $domainRoot . "/statistics/webstat-ssl") {
	  if ($fileName = makeDumpFile("$dumpDir/$domainName.webstat-ssl",
								   "$domainRoot/statistics/webstat-ssl")) {
		$root->{'ATTRIBUTE'}->('cid_webstat_ssl', $fileName);
	  }
	}

	if (-d $domainRoot . "/private") {
	  if ($fileName = makeDumpFile("$dumpDir/$domainName.private",
								   "$domainRoot/private")) {
		$root->{'ATTRIBUTE'}->('cid_private', $fileName);
	  }
	}
  }

  $sql = "SELECT * FROM hosting WHERE dom_id=$domainId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    %hosting = %{$wrapDbh->{'FETCHHASH'}->()};
  }
  $wrapDbh->{'FINISH'}->();

  while (($xmlName,$fieldName)=each(%hostingAttribute)){
    if($hosting{$fieldName}=~/true/){
      $root->{'ATTRIBUTE'}->($xmlName,'true');
    }
  }

  $sql = "SELECT val FROM dom_param WHERE dom_id='$domainId' "
	. "AND param='apacheErrorDocs'";
  if ($wrapDbh->{'EXECUTE'}->($sql)) {
	if ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
	  my $errdocs = (@{$ptrRow})[0];
	  if ($errdocs eq 'true') {
		$root->{'ATTRIBUTE'}->('errdocs', 'true');
	  }
	}
  }

#
# sysuser
#
  if($pleskMajorVersion >= 5) {
	if($id = $hosting{'sys_user_id'}){
	  $item= &makePleskSysUserNode($id);
	  if(ref($item)=~/HASH/){
		$root->{'ADDCHILD'}->($item);
	  }
	}
  } else {
	$item = &makeSyntheticSysUserNode($hosting{'login'}, $hosting{'passwd'}, 'plain');
	if(ref($item) =~ /HASH/) {
	  $root->{'ADDCHILD'}->($item);
	}
  }
#
# end sysuser
#

#
# scripting
#
  $item = &makeXmlNode('scripting');
  while(($xmlName,$fieldName)=each(%hostingScripting)){
    if($hosting{$fieldName}=~/true/){
      $item->{'ATTRIBUTE'}->($xmlName,'true');
    }
  }
  $root->{'ADDCHILD'}->($item);
#
# end scripting
#

#
# frontpage user
#
  if($hosting{'fp_adm'}){
    $item = &makeXmlNode('fpuser');
    $item->{'ATTRIBUTE'}->('name',$hosting{'fp_adm'});
    if($hosting{'fp_pass'}){
      my $passNode = &makePasswordNode($hosting{'fp_pass'},'plain');
      $item->{'ADDCHILD'}->($passNode);
    }
    $root->{'ADDCHILD'}->($item);
  }
#
# end frontpage user
#

#
# logrotation
#
  &makeLogrotationNode($root, $domainId);
#
# end logrotation
#

#
# anonftp
#
  &addAnonFtp($root,$domainId,$domainName);
#
# end anonftp
#

#
# protected dirs
#
	if (moreThanVersion(5.0.0)) {
		$sql = "SELECT id,path,realm,non_ssl,`ssl` FROM protected_dirs WHERE dom_id=$domainId ORDER BY id";
	} else {
		$sql = "SELECT id,path,realm FROM protected_dirs WHERE dom_id=$domainId ORDER BY id";
	}

  if($wrapDbh->{'EXECUTE'}->($sql)){
    my (@dirs);
    while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
      push @dirs,[@{$ptrRow}];
    }
    $wrapDbh->{'FINISH'}->();
    foreach $ptrRow (@dirs){
      $item = &makeProtDirNode(@{$ptrRow});
      if(ref($item)=~/HASH/){
		$root->{'ADDCHILD'}->($item);
      }
    }
  }
  $wrapDbh->{'FINISH'}->();
#
# end protected dirs
#

#
# web users
#
  $sql = "SELECT * FROM web_users WHERE dom_id=$domainId ORDER BY id";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    my (@webs);
    while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
      push @webs,{%{$ptrHash}};
    }
    $wrapDbh->{'FINISH'}->();
    foreach $ptrHash (@webs){
      $item = &makeWebUserNode($ptrHash,$domainName);
      if(ref($item)=~/HASH/){
		$root->{'ADDCHILD'}->($item);
      }
    }
  }
#
# end web users
#

#
# subdomains
#
  if($pleskMajorVersion >= 7){
    $sql = "SELECT * FROM subdomains WHERE dom_id=$domainId ORDER BY id";

    if($wrapDbh->{'EXECUTE'}->($sql)){
      my (@subdoms);
      while ($ptrHash = $wrapDbh->{'FETCHHASH'}->()){
		push @subdoms,{%{$ptrHash}};
      }
      $wrapDbh->{'FINISH'}->();
      foreach $ptrHash (@subdoms){
		$item = &makeSubDomainNode($ptrHash,$domainName);
		if(ref($item)=~/HASH/){
		  $root->{'ADDCHILD'}->($item);
		}
      }
    }
  }
#
# end subdomains
#

  #
  # configuration files (conf/vhost[_ssl].conf)
  #
  my $fileName;
  my $domainRoot = $PsaConfig{'HTTPD_VHOSTS_D'}.'/'.$domainName;
  my $vhostConf = $domainRoot .'/conf/vhost.conf';
  my $vhostSslConf = $domainRoot .'/conf/vhost_ssl.conf';
  if (-f $vhostConf || -f $vhostSslConf) {
	if($fileName = makeDumpFile("$dumpDir/$domainName.conf", "$domainRoot/conf", ["vhost.conf","vhost_ssl.conf"])) {
	  $root->{'ATTRIBUTE'}->('cid_conf', $fileName);
	}
  }
  #
  # end configuration files (conf/vhost[_ssl].conf)
  #

  #
  # Webalizer configuration
  #
  if (atLeastVersion(7, 5, 3)) {
	my $node = makeXmlNode("webalizer");

	my $sql = "SELECT referrer FROM webalizer_hidden_referrer "
	  . " WHERE dom_id=$domainId";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
	  while (my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		if ($ptrRow->[0] eq "Direct Request") {
		  $node->{'ATTRIBUTE'}->('hide-direct-referrer', 'true');
		} else {
		  my $ref = makeXmlNode("webalizer-hidden-referrer", $ptrRow->[0]);
		  $node->{'ADDCHILD'}->($ref);
		}
	  }
	}
	$wrapDbh->{'FINISH'}->();

	$sql = "SELECT referrer, group_name FROM webalizer_group_referrer "
	  . " WHERE dom_id=$domainId";
	if ($wrapDbh->{'EXECUTE'}->($sql)) {
	  while (my $ptrRow = $wrapDbh->{'FETCHROW'}->()) {
		my $ref = makeXmlNode("webalizer-group-referrer", $ptrRow->[0]);
		$ref->{'ATTRIBUTE'}->('group-name', $ptrRow->[1]);
		$node->{'ADDCHILD'}->($ref);
	  }
	}
	$wrapDbh->{'FINISH'}->();

	$root->{'ADDCHILD'}->($node);
  }

  # Cron


  return $root;
}

sub makeLogrotationNode {
  my ($parent, $domainId) = @_;

  if($pleskMajorVersion >= 5) {
	my $sql = "SELECT * FROM dom_param dp LEFT JOIN log_rotation lr ".
	  "ON dp.val = lr.id ".
		" WHERE param = 'logrotation_id' AND dom_id = $domainId";

	my $hashPtr;
	if($wrapDbh->{'EXECUTE'}->($sql) && ($hashPtr = $wrapDbh->{'FETCHHASH'}->())) {
	  my %logRotation = %{$hashPtr};

	  if ($logRotation{'id'}) {
		my $item = &makeXmlNode('logrotation');
		$item->{'ATTRIBUTE'}->('max-number-of-logfiles',
							   $logRotation{'max_number_of_logfiles'});
		$item->{'ATTRIBUTE'}->('compress', $logRotation{'compress_enable'});
		unless (defined $logRotation{'email'}) {
		  $logRotation{'email'} = '';
		}
		$item->{'ATTRIBUTE'}->('email', $logRotation{'email'});
		$item->{'ATTRIBUTE'}->('enabled', $logRotation{'turned_on'});

		my $rottype;
		if($logRotation{'period_type'} eq 'by_time') {
		  $rottype = &makeXmlNode('logrotation-period');
		  $rottype->{'ATTRIBUTE'}->('period', $logRotation{'period'});
		} else {
		  if($pleskMajorVersion <= 6) {
			$rottype = &makeXmlNode('logrotation-maxsize', $logRotation{'period'}*1024);
		  } else {
			$rottype = &makeXmlNode('logrotation-maxsize', $logRotation{'period'});
		  }
		}
		$item->{'ADDCHILD'}->($rottype);
		$parent->{'ADDCHILD'}->($item);
	  }
	}
	$wrapDbh->{'FINISH'}->();
  }
}

#
# makeSubDomainNode
#
#  argumenets:
#       $ptrSubDomain - pointer to hash with subdomain's info (row of table 'subdomains')
#       $domainName - name of domain
#
#  return:
#       $root = XMl node
#
sub makeSubDomainNode {
  my($ptrSubDomain,$domainName)=@_;
  unless(ref($ptrSubDomain)=~/HASH/){
    &printToError("Error: makeSubDomainNode: bad argumets");
    return undef;
  }
  my($root,$subDomainRoot,$fileName,$subDomainName,$item,$xmlName,$fieldName);
  $root = &makeXmlNode('subdomain');
  if (atLeastVersion(7, 1, 0)) {
	$subDomainName = $ptrSubDomain->{'displayName'};
  } else {
	$subDomainName = $ptrSubDomain->{'name'};
  }
  $root->{'ATTRIBUTE'}->('name',$subDomainName);

#
# content
#
  $subDomainRoot=$PsaConfig{'HTTPD_VHOSTS_D'}."/$domainName/subdomains/$subDomainName";
  if(-d $subDomainRoot){
    if($fileName = &makeDumpFile("$dumpDir/$subDomainName.$domainName.html",
			      "$subDomainRoot/httpdocs")){
      $root->{'ATTRIBUTE'}->('cid_docroot',$fileName);
    }
    if($fileName = &makeDumpFile("$dumpDir/$subDomainName.$domainName.cgi",
			      "$subDomainRoot/cgi-bin")){
      $root->{'ATTRIBUTE'}->('cid_cgi',$fileName);
    }
  }
#
# end content
#
  if($ptrSubDomain->{'sys_user_type'}=~/native/){
    $item = &makePleskSysUserNode($ptrSubDomain->{'sys_user_id'});
    if(ref($item)=~/HASH/){
      $root->{'ADDCHILD'}->($item);
    }
  }

#
# scripting
#
  $item = &makeXmlNode('scripting');
  while(($xmlName,$fieldName)=each(%subDomainScripting)){
    if($ptrSubDomain->{$fieldName}=~/true/){
      $item->{'ATTRIBUTE'}->($xmlName,'true');
    }
  }
  $root->{'ADDCHILD'}->($item);
#
# end scripting
#

  return $root;
}


#
# makeWebUserNode
#
#  argumenets:
#       $ptrWebUser - pointer to hash with web user's info (row of table 'web_users')
#       $domainName - name of domain
#
#  return:
#       $root = XML node
#
sub makeWebUserNode {
  my($ptrWebUser,$domainName)=@_;
  unless(ref($ptrWebUser)=~/HASH/){
    &printToError("Error: makeWebUserNode: bad argumets");
    return undef;
  }
  my($root,$home,$fileName,$userName,$item,$xmlName,$fieldName);
  $root = &makeXmlNode('webuser');


  if($pleskMajorVersion >= 5) {
	$item = &makePleskSysUserNode($ptrWebUser->{'sys_user_id'});
	if(ref($item)=~/HASH/){
	  $root->{'ADDCHILD'}->($item);
	  $userName = $item->{'ATTRIBUTE'}->('name');
	  $root->{'ATTRIBUTE'}->('name',$userName);

	  my $subDomainRoot=$PsaConfig{'HTTPD_VHOSTS_D'}."/$domainName/web_users/$userName";

	  if(-d $subDomainRoot){
		if($fileName = &makeDumpFile("$dumpDir/$domainName\_$userName",
									 $subDomainRoot)){
		  $root->{'ATTRIBUTE'}->('cid',$fileName);
		}
	  }
	  my $privateData = $PsaConfig{'HTTPD_VHOSTS_D'}."/$domainName/private/$userName";
	  if (-d $privateData) {
		if ($fileName = makeDumpFile("$dumpDir/$domainName\_$userName.private",
									 $privateData)) {
		  $root->{'ATTRIBUTE'}->('cid_private', $fileName);
		}
	  }
	}
  } else {
	my $username = $ptrWebUser->{'username'};
	my @pwnam = getpwnam($username);
	my $password = $pwnam[1];
	$item = &makeSyntheticSysUserNode($username, $password, 'encrypted');
	$root->{'ADDCHILD'}->($item);
	$root->{'ATTRIBUTE'}->('name',$username);
  }

#
# scripting
#
  $item = &makeXmlNode('scripting');
  while(($xmlName,$fieldName)=each(%webUserScripting)){
    if($ptrWebUser->{$fieldName}=~/true/){
      $item->{'ATTRIBUTE'}->($xmlName,'true');
    }
  }
  $root->{'ADDCHILD'}->($item);
#
# end scripting
#

  return $root;
}

#
# makeProtDirNode - make protected directory node
#
# arguments:
#        $pdirId - ID
#        $pdirPath - path of protected directory
#        $pdirTitle - header text
#        $pdirNonSSL - whether is directory in http
#        $pdirSSL - or in https
#
# return:
#        $root - XML node
#
sub makeProtDirNode {
  my($pdirId, $pdirPath, $pdirTitle, $pdirNonSSL, $pdirSSL) = @_;

  my($root, $sql, $userNode, $item, $ptrRow);

  $root = makeXmlNode('pdir');
  # workaround of CLI inabliity to create '' directory.
  $pdirPath = '/' . $pdirPath;

  $root->{'ATTRIBUTE'}->('name', $pdirPath);
  if($pdirTitle) {
    $root->{'ATTRIBUTE'}->('title', $pdirTitle);
  }
  if ($pdirNonSSL) {
    $root->{'ATTRIBUTE'}->('nonssl', $pdirNonSSL);
  }
  if ($pdirSSL) {
    $root->{'ATTRIBUTE'}->('ssl', $pdirSSL);
  }

  if (atLeastVersion(7, 1, 0)) {
    $sql = "SELECT p.login, a.password, a.type FROM pd_users p ".
      " LEFT JOIN accounts a ON p.account_id = a.id ".
	  " WHERE p.pd_id=$pdirId ORDER BY p.id";
  } else {
    $sql = "SELECT p.login, p.passwd, 'plain' FROM pd_users p ".
      " WHERE p.pd_id=$pdirId ORDER BY p.id"
  }

  if ($wrapDbh->{'EXECUTE'}->($sql)) {
    while ($ptrRow=$wrapDbh->{'FETCHROW'}->()) {
      $userNode = makeXmlNode('pduser');
      $root->{'ADDCHILD'}->($userNode);
      $userNode->{'ATTRIBUTE'}->('name', $ptrRow->[0]);
      $item = makePasswordNode($ptrRow->[1], pwdType($ptrRow->[2]));
      $userNode->{'ADDCHILD'}->($item);
    }
  }
  $wrapDbh->{'FINISH'}->();
  return $root;
}

sub makePermissionNode {
  my ($parent, $prefix, $name) = @_;
  my $node = makeXmlNode($prefix . '-permission');
  $node->{'ATTRIBUTE'}->('name', $name);
  $parent->{'ADDCHILD'}->($node);
}

sub makeAnonftpLimitNode {
  my ($parent, $name, $value) = @_;
  if($value != 0) {
	my $node = makeXmlNode('anonftp-limit', $value);
	$node->{'ATTRIBUTE'}->('name', $name);
	$parent->{'ADDCHILD'}->($node);
  }
}

sub addAnonFtp {
  my($root,$domainId,$domainName)=@_;
  unless(ref($root)=~/HASH/){
    &printToError("Error: addAnonFtp: bad argumets");
    return undef;
  }

  my($anonRoot,$ptrHash,$count,$fileName,$domainRoot,$sql);
  $count=0;
  $sql="SELECT * FROM anon_ftp WHERE dom_id=$domainId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    while($ptrHash=$wrapDbh->{'FETCHHASH'}->()){

      $anonRoot = &makeXmlNode('anonftp');
      $root->{'ADDCHILD'}->($anonRoot);
      if($ptrHash->{'status'}=~/true/){
		$anonRoot->{'ATTRIBUTE'}->('pub','true');
      }
      if($ptrHash->{'incoming'}=~/true/){
		$anonRoot->{'ATTRIBUTE'}->('incoming','true');
      }
	  if (defined($ptrHash->{'max_conn'})) {
		makeAnonftpLimitNode($anonRoot,'max-connections', $ptrHash->{'max_conn'});
	  }
	  if (defined($ptrHash->{'bandwidth'})) {
		makeAnonftpLimitNode($anonRoot, 'bandwidth', $ptrHash->{'bandwidth'});
	  }
	  if (defined($ptrHash->{'quota'})) {
		makeAnonftpLimitNode($anonRoot, 'incoming-disk-quota', $ptrHash->{'quota'});
	  }
	  if (defined($ptrHash->{'display_login'})) {
		$anonRoot->{'ATTRIBUTE'}->('display-login', $ptrHash->{'display_login'});
	  }
	  if ($ptrHash->{'incoming_readable'} =~ /true/) {
		makePermissionNode($anonRoot, 'anonftp', 'incoming-download');
	  }
	  if ($ptrHash->{'incoming_subdirs'} =~ /true/) {
		makePermissionNode($anonRoot, 'anonftp', 'incoming-mkdir');
	  }
	  if (defined($ptrHash->{'login_text'})) {
		my $loginMessageNode = makeXmlNode('login-message', $ptrHash->{'login_text'});
		$anonRoot->{'ADDCHILD'}->($loginMessageNode);
	  }
      $domainRoot=$PsaConfig{'HTTPD_VHOSTS_D'}.'/'.$domainName;
      if(-d $domainRoot){

		if($fileName = &makeDumpFile("$dumpDir/$domainName.anonftp.pub",
									 "$domainRoot/anon_ftp/pub")){
		  $anonRoot->{'ATTRIBUTE'}->('cid',$fileName);
		}
		if($fileName = &makeDumpFile("$dumpDir/$domainName.anonftp.incoming",
									 "$domainRoot/anon_ftp/incoming")){
		  $anonRoot->{'ATTRIBUTE'}->('cid_incoming',$fileName);
		}
      }
      $count++;
    }
  }
  $wrapDbh->{'FINISH'}->();
  return $count;
}

sub getSysUserInfo {
  my ($sysUserId) = @_;

  my $sql = "SELECT * FROM sys_users WHERE id=$sysUserId";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    printToError("Error: makeSysUser: sys user #$sysUserId  is not found");
    return undef;
  }
  my %sysuser = %{$wrapDbh->{'FETCHHASH'}->()};
  $wrapDbh->{'FINISH'}->();

  return %sysuser;
}

sub makePleskSysUserNode {
  my %sysuser = getSysUserInfo(@_);

  my $root = &makeXmlNode('sysuser');
#
# attributes
#
  $root->{'ATTRIBUTE'}->('name',lc($sysuser{'login'}));
  $root->{'ATTRIBUTE'}->('shell',$sysuser{'shell'}) if $sysuser{'shell'};

  my $quota;
  if($sysuser{'quota'}){
	if($pleskMajorVersion <= 6){
	  unless($sysuser{'quota'} eq "-1") {
		$quota = $sysuser{'quota'}*1024*1024;
	  }
	} else{
	  $quota = $sysuser{'quota'};
	}
	$root->{'ATTRIBUTE'}->('quota',$quota) if $quota;
  }

#
# end attributes
#

#
# password node
#
  if ($pleskMajorVersion <= 5) {
	$root->{'ADDCHILD'}->(makePasswordNode($sysuser{'passwd'}, 'plain'));
  } else {
	$root->{'ADDCHILD'}->(makeAccountPasswordNode($sysuser{'account_id'}));
  }
#
# end password node
#

#
# crontab node
#
  my $crontabmng = "$PsaConfig{'PRODUCT_ROOT_D'}/admin/sbin/crontabmng";
  if (-x $crontabmng) {
	open (CRONTAB, "$crontabmng get $sysuser{'login'} |");
	my $crontab = join "", <CRONTAB>;
	close (CRONTAB);
	if ($crontab ne "\n") {
	  $root->{'ADDCHILD'}->(makeXmlNode('cron', $crontab));
	}
  }
#
# end crontab node
#

  return $root;

}

sub makeSyntheticSysUserNode {
  my ($name, $password, $passtype) = @_;

  my $root = &makeXmlNode('sysuser');
  $root->{'ATTRIBUTE'}->('name', $name);
  my $passwdNode = &makePasswordNode($password, $passtype);
  $root->{'ADDCHILD'}->($passwdNode);
  return $root;
}

sub makeFhostingNode {
  my ($ptrDomain)=@_;
  unless(ref($ptrDomain)=~/HASH/){
    &printToError("Error: makeFhostingNode: bag arguments");
    return undef;
  }
  my($sql,$domainId,$root,$forward);
  $domainId=$ptrDomain->{'id'};
  $sql="SELECT redirect FROM forwarding WHERE dom_id=$domainId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    ($forward)=@{$wrapDbh->{'FETCHROW'}->()};
    $root=&makeXmlNode('fhosting',$forward);
  }else{
    &printToError("Error: makeFhostingNode: forward for domain '".$ptrDomain->{'name'}."' is not found");
  }
  $wrapDbh->{'FINISH'}->();

  return $root;
}

sub makeShostingNode {
  my ($ptrDomain)=@_;
  unless(ref($ptrDomain)=~/HASH/){
    &printToError("Error: makeShostingNode: bag arguments");
    return undef;
  }

  my($sql,$domainId,$root,$forward);
  $domainId=$ptrDomain->{'id'};
  $sql="SELECT redirect FROM forwarding WHERE dom_id=$domainId";
  if($wrapDbh->{'EXECUTE'}->($sql)){
    ($forward)=@{$wrapDbh->{'FETCHROW'}->()};
    $root=&makeXmlNode('shosting',$forward);
  }else{
    &printToError("Error: makeShostingNode: forward for domain '".$ptrDomain->{'name'}."' is not found");
  }
  $wrapDbh->{'FINISH'}->();

  return $root;
}

sub makeIpNode5 {
  my($ptrIpHash,$ptrClients) = @_;
  $ptrIpHash||={};
  my($root,$item,$ptrRow,$sharedNode,$exclusiveNode,$ip,$ipNode,$id,
     $ptrHash,$sql,$list,@ids,@names,$domain,$domainId);

  $root=&makeXmlNode('map');

  $sharedNode = &makeXmlNode('shared');
  $exclusiveNode = &makeXmlNode('exclusive');
  $root->{'ADDCHILD'}->($sharedNode);
  $root->{'ADDCHILD'}->($exclusiveNode);

#
# get domains
#
  if(ref($ptrClients) =~ /HASH/) {
    $list = join(',', sort {$a <=> $b} keys %{$ptrClients});
  }

  if ($pleskMajorVersion <= 2) {
	# Sometimes host have A record not on the domain name itself
	# but on some alias. I've see A record pointing www2.domain to IP address
	if ($list) {
	  $sql = "SELECT d.name, d.id, r.val, a.status, d.cl_id FROM dns_recs r, domains d ".
		"LEFT JOIN hosting h ON h.dom_id=d.id ".
		  "LEFT JOIN anon_ftp a ON a.dom_id=d.id ".
			"WHERE r.dom_id=d.id AND r.type='A' AND cl_id IN ($list)";
	} else {
	  $sql = "SELECT d.name, d.id, r.val, a.status, d.cl_id FROM dns_recs r, domains d ".
		"LEFT JOIN hosting h ON h.dom_id=d.id ".
		  "LEFT JOIN anon_ftp a ON a.dom_id=d.id ".
			"WHERE r.dom_id=d.id AND r.type='A'";
	}
  } else {
	if ($list) {
	  $sql = "SELECT d.name, d.id, h.ip_address, a.status, d.cl_id FROM domains d ".
		"LEFT JOIN hosting h ON h.dom_id=d.id ".
		  "LEFT JOIN anon_ftp a ON a.dom_id=d.id WHERE cl_id IN ($list)";
	} else {
	  $sql = "SELECT d.name, d.id, h.ip_address, a.status, d.cl_id FROM domains d ".
		"LEFT JOIN hosting h ON h.dom_id=d.id ".
		  "LEFT JOIN anon_ftp a ON a.dom_id=d.id";
	}
  }

  if ($wrapDbh->{'EXECUTE'}->($sql)) {
    while ($ptrRow = $wrapDbh->{'FETCHROW'}->()) {
      $domainsInfo{$ptrRow->[0]} = [@{$ptrRow}[1..4]];
    }
  }
  $wrapDbh->{'FINISH'}->();
#
# end get domains
#

#
# get ip from DNS
#
  while (($domain,$ptrRow) = each (%domainsInfo)){
    unless($ptrRow->[1]){
      push @names,$domain;
    }
  }
  if(@names){
    $list = join(',',map {"'$_'"} @names);
    $sql = "SELECT host,val FROM dns_recs WHERE type='A' AND host in ($list)";
    if($wrapDbh->{'EXECUTE'}->($sql)){
      while($ptrRow=$wrapDbh->{'FETCHROW'}->()){
		if($ptrRow->[1]=~/^\d+\.\d+\.\d+\.\d+$/){
		  $domainsInfo{$ptrRow->[0]}->[1]=$ptrRow->[1];
		}
      }
    }
    $wrapDbh->{'FINISH'}->();
  }
#
# end get ip from DNS
#

  $sql = "SELECT val FROM misc WHERE param LIKE 'NameVirtualHost'";
  ($wrapDbh->{'EXECUTE'}->($sql) && ($ptrRow = $wrapDbh->{'FETCHROW'}->()))
	or die "Cannot retrieve name-based hosts IP";

  my $defaultIp = (@{$ptrRow})[0];

  $wrapDbh->{'FINISH'}->();

  while (($domain, $ptrRow) = each (%domainsInfo)) {
	unless ($ptrRow->[1] && $ptrRow->[1] !~ /NULL/) {
	  $domainsInfo{$domain}->[1] = $defaultIp;
	}
  }

#
# sort by shared or exclusive
#
  my %excl;
  if($pleskMajorVersion <= 2) {
	foreach $domain (keys %domainsInfo) {
	  my ($exclusive);
	  ($domainId, $ip, $exclusive) = @{$domainsInfo{$domain}};
	  $excl{$ip} = $defaultIp ne $ip;
	}
  } else {
	my ($exclusive);
	foreach $domain (keys %domainsInfo){
	  ($domainId,$ip,$exclusive)=@{$domainsInfo{$domain}};
	  if($exclusive){
		$excl{$ip}=1;
	  }
	}
	foreach $domain (keys %domainsInfo){
	  ($domainId,$ip,$exclusive)=@{$domainsInfo{$domain}};
	  unless($exclusive){
		$excl{$ip}=0;
	  }
	}
  }
#
# end sort by shared or exclusive
#

#
# generate xml
#
  foreach $ip (sort keys %excl){
    $id = join('',unpack('n2',pack('C4', split(/\./,$ip))));

    $ptrIpHash->{$ip}=$id; ## store to map ip => id

    $ipNode = &makeXmlNode('ip');
    $ipNode->{'ATTRIBUTE'}->('id','ip_'.$id);
      $item = &makeXmlNode('old',$ip);
    $ipNode->{'ADDCHILD'}->($item);
    $item = &makeXmlNode('new');
    $ipNode->{'ADDCHILD'}->($item);

    if ($excl{$ip}){
      $exclusiveNode->{'ADDCHILD'}->($ipNode);
    }else{
      $sharedNode->{'ADDCHILD'}->($ipNode);
    }
  }
  return $root;
}

sub makeIpNode {
  my($ptrIpHash,$ptrClients) = @_;
  $ptrIpHash||={};
  my($root,$item,$ptrRow,$id,$shared,$exclusive,$ip,$ipNode,
     $ptrHash,$sql,$list,@ids);

  $root=&makeXmlNode('map');

  $shared = &makeXmlNode('shared');
  $exclusive = &makeXmlNode('exclusive');
  $root->{'ADDCHILD'}->($shared);
  $root->{'ADDCHILD'}->($exclusive);

  if(ref($ptrClients)=~/HASH/){
    $list = join(',',sort{$a<=>$b}keys %{$ptrClients});
  }
  if ($list){
    $sql = "SELECT DISTINCT r.component_id FROM Repository r, clients c ".
      "WHERE r.rep_id=c.pool_id AND c.id IN ($list)";
    if($wrapDbh->{'EXECUTE'}->($sql)){
      while($ptrHash=$wrapDbh->{'FETCHROW'}->()){
	push @ids,$ptrHash->[0];
      }
    }
    $wrapDbh->{'FINISH'}->();
  }

  if(@ids){
    $list=join(',',@ids);
    $sql = "SELECT * FROM IP_Addresses WHERE id IN ($list) ORDER BY id";
  }else{
    $sql = "SELECT * FROM IP_Addresses ORDER BY id";
  }

  my $firstIpId;

  if($wrapDbh->{'EXECUTE'}->($sql)){
    while($ptrHash=$wrapDbh->{'FETCHHASH'}->()){
      $id = 'ip_'.$ptrHash->{'id'};
      $ip = $ptrHash->{'ip_address'};

      $ptrIpHash->{$ip}=$ptrHash->{'id'}; ## store to map ip => id

	  if(!(defined($firstIpId))) {
		$firstIpId = $ptrHash->{'id'};
	  }
	  if(!(defined($defaultIpId)) and $ptrHash->{'type'} eq 'shared') {
		$defaultIpId = $ptrHash->{'id'};
	  }

      $ipNode = &makeXmlNode('ip');
      $ipNode->{'ATTRIBUTE'}->('id',$id);
      $item = &makeXmlNode('old',$ip);
      $ipNode->{'ADDCHILD'}->($item);
      $item = &makeXmlNode('new');
      $ipNode->{'ADDCHILD'}->($item);

      if ($ptrHash->{'type'}=~/shared/){
		$shared->{'ADDCHILD'}->($ipNode);
	  }else{
		$exclusive->{'ADDCHILD'}->($ipNode);
      }
    }
  }

  if(!(defined($defaultIpId))) {
	$defaultIpId = $firstIpId;
  }

  $wrapDbh->{'FINISH'}->();

  return $root;
}

#
# Returns password node for given account ID
#
sub makeAccountPasswordNode
{
	my $account_id = shift;
	my $node;

	if ($wrapDbh->{'EXECUTE'}->("SELECT password, type FROM accounts WHERE id='$account_id'")) {
		my ($passwd, $type) = @{$wrapDbh->{'FETCHROW'}->()};
		$node = &makePasswordNode($passwd, pwdType($type));
    } else {
		# generates a stub node
		$node = &makePasswordNode();
	}
	$wrapDbh->{'FINISH'}->();

	return $node;
}

#
# get version of Plesk
#
sub getVersion {
  my $psaRoot = $PsaConfig{'PRODUCT_ROOT_D'};
  my $versionFile = (-d $psaRoot)?"$psaRoot/version":'/usr/local/psa/version';
  my ($version, $out);
  unless (open(VERSION,"< $versionFile")) {
	return;
  }
  binmode VERSION;

  $out = <VERSION>;
  close(VERSION);
  chomp $out;

  if ($out =~ /^(\d+)\.(\d+)(.*)$/) {
	$pleskMajorVersion = $1;
	$pleskMinorVersion = $2;

	if($3 =~ /^\.(\d+)/) {
	  $pleskMinorMinorVersion = $1;
	} else {
	  $pleskMinorMinorVersion = 0;
	}

	if ($pleskMajorVersion < 2 || ($pleskMajorVersion == 2 && $pleskMinorVersion < 5)) {
	  return;
	}

	if ($pleskMinorMinorVersion != 0) {
	  return "Plesk $pleskMajorVersion.$pleskMinorVersion.$pleskMinorMinorVersion";
	} else {
	  return "Plesk $pleskMajorVersion.$pleskMinorVersion";
	}
  }
}

sub printAgentStatus {
  my $dumpFile = shift;
  my $statusFile = shift;
  my $outFh = shift || *STDOUT;
  my $rootName = shift || 'agentstatus';

  my ($root,$util,$out,$item,$engineRoot,$version);

  $root = &makeXmlNode($rootName);

  $version = getVersion();
  unless ($version && (defined $bin{'iconv'}) && supportedVersion()) {
	if (!supportedVersion()) {
	  $item = makeXmlNode('wrong-platform', "Migration from Plesk versions above $supportedVersionStr is not supported");
	} elsif (not defined $bin{'iconv'}) {
	  $item = &makeXmlNode('wrong-platform', 'no iconv found on the source host');
	} else {
	  $item = &makeXmlNode('wrong-platform', '');
	}

    $root->{'ADDCHILD'}->($item);
    &printXml($root,$outFh,$rootName.'.dtd');
    return;
  }

  $item = &makeXmlNode('dumpfile',$dumpFile);
  $root->{'ADDCHILD'}->($item);
  $item = &makeXmlNode('statusfile',$statusFile);
  $root->{'ADDCHILD'}->($item);

  &printXml($root,$outFh,$rootName.'.dtd');
}

sub loadMainConfig
{
	my $pathToConf = shift || '/etc/psa/psa.conf';
	my $ptrHash    = shift || {};
	unless (-T $pathToConf) {
		return undef;
	}

	my $fileParser = &makeFileParser($pathToConf);
	my $ptrRet = $fileParser->{'PARSE'}->('KEYSEPARATOR' => qr/\s+/, 'COMMENT' => '#');
	unless (ref($ptrRet) =~ /HASH/) {
		return undef;
	}
	%{$ptrHash} = %{$ptrRet};

	my $dirConf = $pathToConf;
	$dirConf =~ s/\/[^\/]+$//;
	my ($dir, $path, $password);

	foreach $dir ($dirConf, '/etc/psa') {
		$path = "$dir/.psa.shadow";
		if (-T $path) {
			if (open(OUTPUT, "< $path")) {
				binmode OUTPUT;
				$password = <OUTPUT>;
				close (OUTPUT);

				chomp $password;
				$password =~ s/^\s+//;
				$password =~ s/\s+$//;
				$ptrHash->{'password'} = $password;
				last;
			}
		}
	}

  return $pathToConf;
}

sub makeEmptyDump {
  my($rootName)=@_;
  my $root=&makeXmlNode($rootName);

  my $mapRoot = &makeXmlNode('map');
  $root->{'ADDCHILD'}->($mapRoot);

  my $item = &makeXmlNode('shared');
  $mapRoot->{'ADDCHILD'}->($item);

  $item = &makeXmlNode('exclusive');
  $mapRoot->{'ADDCHILD'}->($item);

  return $root;
}

# Converts 'crypt' -> 'encrypted'
# It's a workaround for psa.acounts.type
# bug 47030
sub pwdType
{
	my ($type) = @_;
	$type =~ s/^crypt$/encrypted/;
	return $type;
}
sub makeSpamassassinNode {
	my ($root,$mailName)=@_ ;

  my  $sql = "SELECT count(*) FROM sa_conf WHERE mailname='$mailName'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
  &printToError("Error during getting information about spam filter for $mailName ");

  return undef;

  }
  my $ptrRow = $wrapDbh->{'FETCHROW'}->();

  if ($ptrRow->[0]==0) {

        return;
  }
  $wrapDbh->{'FINISH'}->();

  $sql = "SELECT * FROM sa_conf WHERE mailname='$mailName'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
    &printToError("Error during getting information about spam filter for $mailName ");

  return undef;
  }
  my %spam = %{$wrapDbh->{'FETCHHASH'}->()};

	my $item;
    if ($spam{'flt_enabled'} eq 'both'){
		$item=&makeXmlNode('spamassassin');
		$item->{'ATTRIBUTE'}->('status','on');
		$item->{'ATTRIBUTE'}->('server-conf','true');
	} elsif ($spam{'flt_enabled'} eq 'user'){
		$item=&makeXmlNode('spamassassin');
		$item->{'ATTRIBUTE'}->('status','on');
		$item->{'ATTRIBUTE'}->('server-conf','false');
	} else {
		#skip creation
		return ;
	}
	$item->{'ATTRIBUTE'}->('hits',$spam{'hits_required'});
		if($spam{'spam_action'} eq 'save'){
			$item->{'ATTRIBUTE'}->('action','mark');
			if($spam{'rw_subject_tag'})
			{
				$item->{'ATTRIBUTE'}->('modify-subj','true');
				$item->{'ATTRIBUTE'}->('subj-text',$spam{'rw_subject_tag'});
			}else{
				$item->{'ATTRIBUTE'}->('modify-subj','false');
			}
		}else{
			$item->{'ATTRIBUTE'}->('action','del');
		}
	&makeBlackListNode($item,$spam{'mailname'});
	&makeWhiteListNode($item,$spam{'mailname'});

	$root->{'ADDCHILD'}->($item);
	return $root;

}

sub makeWhiteListNode {
	my ($root,$mailname) =@_;
	my ($sql,@patterns,$pattern,$item,$ptrRow);
   $sql = "SELECT count(pattern) FROM sa_list WHERE mailname='$mailname' AND color='white'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
  &printToError("Error during getting information about white list for $mailname ");
  return undef;
  }
$ptrRow = $wrapDbh->{'FETCHROW'}->();
                                                                                if ($ptrRow->[0]==0){
                                                                                        return;
}

$wrapDbh->{'FINISH'}->();

  $sql = "SELECT pattern FROM sa_list WHERE mailname='$mailname' AND color='white'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
  &printToError("Error during getting information about white list patterns for $mailname ");
  return undef;
  }
  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
		push @patterns,$ptrRow->[0];
	  }
	  foreach $pattern (@patterns){
	$item = makeXmlNode('whitelist-member',$pattern);
	$root->{'ADDCHILD'}->($item);
	}
	return $root;
}

sub makeBlackListNode {
	my ($root,$mailname)=@_ ;
	my ($sql,@patterns,$pattern,$item,$ptrRow);
   $sql = "SELECT count(pattern) FROM sa_list WHERE mailname='$mailname' AND color='black'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
  &printToError("Error during getting information about black list for $mailname ");
  return undef;
  }
$ptrRow = $wrapDbh->{'FETCHROW'}->();
   if ($ptrRow->[0]==0){

   	return;
   }

$wrapDbh->{'FINISH'}->();

   $sql = "SELECT pattern FROM sa_list WHERE mailname='$mailname' AND color='black'";
  unless($wrapDbh->{'EXECUTE'}->($sql)){
  &printToError("Error during getting information about black list patterns for $mailname ");
  return undef;
  }
  while ($ptrRow = $wrapDbh->{'FETCHROW'}->()){
		push @patterns,$ptrRow->[0];
	  }
	  foreach $pattern (@patterns){
	$item = makeXmlNode('blacklist-member',$pattern);
	$root->{'ADDCHILD'}->($item);
	}
	return $root;

}

sub makeDrwebNode {
  my ($root, $mailname)=@_;
  my $drwebFile = "/etc/drweb/users.conf";
  my $item;
  unless (open(DRWEBFILE, "<$drwebFile")) {
	printToLog("Error: unable to open $drwebFile");
	return;
  }
  while (<DRWEBFILE>) {
	chomp;
	next if /^#/; #skip comments
	next if /^\[/;
	my @parts = split /\s+/, $_, 4;
	if ($parts[3] =~ /$mailname/) {
	  $item = &makeXmlNode('drweb');
	  if ($parts[1] =~ /to/) {
		$item->{'ATTRIBUTE'}->('state', 'in');
	  } elsif ($parts[1] =~ /from/) {
		$item->{'ATTRIBUTE'}->('state', 'out');
	  } else {
		$item->{'ATTRIBUTE'}->('state', 'inout');
	  }
	}
  }
  close(DRWEBFILE);
  $root->{'ADDCHILD'}->($item);
  return $root;
}

sub printHelp {

  print <<"HELP";

Agent of migration from 'Plesk x' to 'Plesk 7'

Usage:
  $0 <options>

Options:
  -c |--config=<path>        path to 'psa.conf'

  -s |--get-status           get status of the agent
  -dc|--dump-accounts=<list> a coma separated list of resellers to dump
  -dd|--dump-domains=<list>  a coma separated list of customers to dump
  -da|--dump-all             make a full dump

  -lc|--get-content-list     get list of content files
  -nc|--no-content           do not make content files
  -nz|--no-compress          do not compress content files

  -o |--output=<path>        redirect output to file

  -h |--help                 this help
  -d |--debug=<level>        level of debug messages
  -v |--version

HELP
}
