<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
  <!-- @desc Cobalt RaQ3 (NG, experimental) -->

  <xsl:output indent="yes" encoding="UTF-8"/>

  <xsl:template match="RaQ34dump">
	<xsl:text disable-output-escaping="yes">&lt;!DOCTYPE plesk7dump SYSTEM 'Plesk70.dtd'&gt;</xsl:text>
	<plesk7dump>
	  <xsl:call-template name="create-ip-map"/>
	  <xsl:apply-templates select="client"/>
	</plesk7dump>
  </xsl:template>

  <xsl:key name="ip-shared" match="domain[not(anonftp)]" use="@ip"/>

  <xsl:template name="create-ip-map">
    <map>
      <shared>
        <xsl:for-each select="client/domain[generate-id(.) = generate-id(key('ip-shared', @ip))]">
          <ip id="{generate-id(.)}">
            <old><xsl:value-of select="@ip"/></old>
            <new><xsl:value-of select="@ip"/></new>
          </ip>
        </xsl:for-each>
      </shared>
      <exclusive>
        <xsl:for-each select="client/domain[anonftp]">
          <ip id="{generate-id(.)}">
            <old><xsl:value-of select="@ip"/></old>
            <new><xsl:value-of select="@ip"/></new>
          </ip>
        </xsl:for-each>
      </exclusive>
    </map>
  </xsl:template>

  <xsl:template match="client">
	<client name="{@name}">
	  <xsl:copy-of select="password"/>
	  <xsl:copy-of select="@email"/>
	  <xsl:copy-of select="@contact"/>
	  <xsl:if test="@enabled">
		<status><xsl:value-of select="@enabled"/></status>
	  </xsl:if>

      <ip_pool>
        <xsl:for-each select="domain">
          <xsl:choose>
            <xsl:when test="anonftp">
              <client_ip ip_ref="{generate-id(.)}"/>
            </xsl:when>
            <xsl:otherwise>
              <client_ip ip_ref="{generate-id(key('ip-shared', @ip))}"/>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:for-each>
      </ip_pool>

	  <xsl:apply-templates select="domain"/>
	</client>
  </xsl:template>

  <xsl:template match="domain">
	<domain name="{@name}" www="{@www}">

	  <!-- IP -->
	  <xsl:choose>
		<xsl:when test="anonftp">
		  <domain_ip ip_ref="{generate-id(.)}"/>
		</xsl:when>
		<xsl:otherwise>
		  <domain_ip ip_ref="{generate-id(key('ip-shared', @ip))}"/>
		</xsl:otherwise>
	  </xsl:choose>

	  <!-- LIMITS -->
	  <xsl:for-each select="domain-limit">
		<xsl:if test="@name='quota'">
		  <xsl:call-template name="limit">
			<xsl:with-param name="limit" select="'disk_space'"/>
			<xsl:with-param name="value" select="text()"/>
		  </xsl:call-template>
		</xsl:if>
		<xsl:if test="@name='users'">
		  <xsl:call-template name="limit">
			<xsl:with-param name="limit" select="'max_wu'"/>
			<xsl:with-param name="value" select="text()"/>
		  </xsl:call-template>
		</xsl:if>
		<xsl:if test="@name='traffic'">
		  <xsl:call-template name="limit">
			<xsl:with-param name="limit" select="'max_traffic'"/>
			<xsl:with-param name="value" select="text()"/>
		  </xsl:call-template>
		</xsl:if>
	  </xsl:for-each>

	  <!-- PERMISSION -->

	  <!-- MAILSYSTEM -->
	  <xsl:call-template name="create-mailsystem"/>

	  <!-- MAILLIST -->
	  <xsl:apply-templates select="maillist"/>

	  <!-- STATUS -->
	  <xsl:if test="@enabled">
		<status><xsl:value-of select="@enabled/text()"/></status>
	  </xsl:if>

	  <!-- CERTIFICATES -->
	  <xsl:copy-of select="certificate"/>


	  <!-- 'no-content' does not provide admin/sysuser, so there's no sense to create phosting -->
	  <xsl:if test="admin/sysuser">

		<!-- PHOSTING -->
		<phosting fpauth="false">
		  <xsl:if test="@cid">
			<xsl:attribute name="cid_docroot"><xsl:value-of select="@cid"/></xsl:attribute>
		  </xsl:if>

		  <xsl:if test="admin/@cid_private">
			<xsl:attribute name="cid_admin_private"><xsl:value-of select="admin/@cid_private"/></xsl:attribute>
		  </xsl:if>

		  <xsl:if test="domain-permission[@name='ssl']">
			<xsl:attribute name="https">true</xsl:attribute>
			<xsl:if test="domain-permission[@name='frontpage']">
			  <xsl:attribute name="fpssl">true</xsl:attribute>
			</xsl:if>
		  </xsl:if>

		  <xsl:call-template name="convert-boolean-permission">
			<xsl:with-param name="raqname" select="'frontpage'"/>
			<xsl:with-param name="pleskname" select="'fp'"/>
		  </xsl:call-template>

		  <!-- SYSUSER -->
		  <xsl:copy-of select="admin/sysuser"/>

		  <!-- SCRIPTING -->
		  <xsl:call-template name="create-scripting"/>

		  <!-- FPUSER -->
		  <xsl:if test="domain-permission[@name='frontpage']">
			<fpuser>
			  <xsl:attribute name="name">
				<xsl:call-template name="mangle-sysuser-name">
				  <xsl:with-param name="name" select="admin/sysuser/@name"/>
				</xsl:call-template>
			  </xsl:attribute>
			  <xsl:copy-of select="admin/sysuser/password"/>
			</fpuser>
		  </xsl:if>

		  <!-- LOGROTATION -->
		  <!-- ANONFTP -->
		  <anonftp>
			<xsl:attribute name="pub">
			  <xsl:if test="anonftp/@cid">true</xsl:if>
			  <xsl:if test="not(anonftp/@cid)">false</xsl:if>
			</xsl:attribute>
			<xsl:attribute name="incoming">
			  <xsl:if test="anonftp/@cid_incoming">true</xsl:if>
			  <xsl:if test="not(anonftp/@cid_incoming)">false</xsl:if>
			</xsl:attribute>
			<xsl:copy-of select="anonftp/@cid"/>
			<xsl:copy-of select="anonftp/@cid_incoming"/>

			<xsl:if test="anonftp/@cid_incoming and anonftp/anonftp-limit[@name='incoming-quota']">
			  <anonftp-limit name="incoming-disk-quota">
				<xsl:value-of select="anonftp/anonftp-limit[@name='incoming-quota']"/>
			  </anonftp-limit>
			</xsl:if>
			<xsl:if test="anonftp/limit[@name='users']">
			  <anonftp-limit name="max-connections">
				<xsl:value-of select="anonftp/limit[@name='users']"/>
			  </anonftp-limit>
			</xsl:if>
			<anonftp-permission name="incoming-mkdir"/>
			<anonftp-permission name="incoming-download"/>
		  </anonftp>
		  <!-- PDIR -->
		  <!-- WEBUSER -->
		  <xsl:apply-templates select="user" mode="webuser"/>
		</phosting>
	  </xsl:if>
	</domain>
  </xsl:template>

  <xsl:template match="maillist">
	<maillist name="{@name}" state="on">
	  <owner><xsl:value-of select="../admin/mail/@mailname"/>@<xsl:value-of select="../@name"/></owner>
	  <xsl:copy-of select="recipient"/>
	</maillist>
  </xsl:template>

  <xsl:template name="create-mailsystem">
	<mailsystem>
	  <xsl:attribute name="enabled">
		<xsl:choose>
		  <xsl:when test="domain-permission[@name='email']">on</xsl:when>
		  <xsl:otherwise>off</xsl:otherwise>
		</xsl:choose>
	  </xsl:attribute>

	  <xsl:for-each select="admin|user">
		<xsl:call-template name="create-mailuser"/>
	  </xsl:for-each>
	  <xsl:if test="@catch-all">
		<catch-all><xsl:value-of select="@catch-all/text()"/></catch-all>
	  </xsl:if>
	</mailsystem>
  </xsl:template>

  <xsl:template match="user" mode="webuser">
	<webuser>
	  <xsl:copy-of select="@cid"/>
	  <xsl:copy-of select="@cid_private"/>
	  <sysuser name="{sysuser/@name}">
		<xsl:if test="sysuser/@quota">
		  <xsl:attribute name="quota">
			<xsl:value-of select="sysuser/@quota"/>
		  </xsl:attribute>
		</xsl:if>
		<xsl:if test="../domain-permission[@name='shell'] and user-permission[@name='shell']">
		  <xsl:attribute name="shell">/bin/bash</xsl:attribute>
		</xsl:if>

		<xsl:copy-of select="sysuser/password"/>
	  </sysuser>
	</webuser>
  </xsl:template>

  <xsl:template name="create-mailuser">
	<mailuser name="{mail/@mailname}" mailgroup-enabled="false">
	  <xsl:copy-of select="sysuser/password"/>
	  <mailbox type="mbox">
		<xsl:if test="user-limit[@name='quota']">
		  <xsl:attribute name="quota">
			<xsl:value-of select="user-limit[@name='quota']"/>
		  </xsl:attribute>
		</xsl:if>
		<xsl:copy-of select="@cid"/>
	  </mailbox>

	  <xsl:copy-of select="mail/alias"/>

	  <xsl:if test="mail/forward">
		<redirect><xsl:value-of select="mail/forward/text()"/></redirect>
	  </xsl:if>

	  <xsl:if test="mail/autoresponder">
		<autoresponder_ext name="vacation">
		  <xsl:copy-of select="mail/autoresponder/@status"/>
		  <text><xsl:value-of select="mail/autoresponder/text()"/></text>
		</autoresponder_ext>
	  </xsl:if>
	</mailuser>
  </xsl:template>

  <xsl:template name="create-scripting">
    <scripting>
      <xsl:call-template name="convert-boolean-permission">
        <xsl:with-param name="raqname" select="'cgi'"/>
        <xsl:with-param name="pleskname" select="'cgi'"/>
      </xsl:call-template>

      <xsl:call-template name="convert-boolean-permission">
        <xsl:with-param name="raqname" select="'ssi'"/>
        <xsl:with-param name="pleskname" select="'ssi'"/>
      </xsl:call-template>
    </scripting>
  </xsl:template>

  <xsl:template name="convert-boolean-permission">
    <xsl:param name="raqname"/>
    <xsl:param name="pleskname"/>
    <xsl:choose>
      <xsl:when test="domain-permission[@name = $raqname]">
        <xsl:attribute name="{$pleskname}">true</xsl:attribute>
      </xsl:when>
      <xsl:otherwise>
        <xsl:attribute name="{$pleskname}">false</xsl:attribute>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template name="mangle-sysuser-name">
    <xsl:param name="name"/>
    <xsl:choose>
      <xsl:when test="$name = 'admin' or $name = 'ftp' or $name = 'anonymous' or $name = 'root' or $name = 'mailman' or $name = 'tomcat4'"><xsl:value-of select="concat($name, '1')"/></xsl:when>
      <xsl:otherwise><xsl:value-of select="$name"/></xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template name="mangle-sysuser">
    <xsl:param name="sysuser"/>
    <sysuser>
      <xsl:attribute name="name">
        <xsl:call-template name="mangle-sysuser-name">
          <xsl:with-param name="name" select="$sysuser/@name"/>
        </xsl:call-template>
      </xsl:attribute>
      <xsl:copy-of select="$sysuser/@quota"/>
      <xsl:copy-of select="$sysuser/@shell"/>
      <xsl:copy-of select="$sysuser/password"/>
    </sysuser>
  </xsl:template>

  <xsl:template name="create-sysuser">
    <xsl:param name="user"/>
    <sysuser>
      <xsl:attribute name="name">
        <xsl:call-template name="mangle-sysuser-name">
          <xsl:with-param name="name" select="$user/sysuser/@name"/>
        </xsl:call-template>
      </xsl:attribute>
      <xsl:if test="$user/permission[@name = 'shell']">
        <xsl:attribute name="shell">/bin/bash</xsl:attribute>
      </xsl:if>
      <xsl:copy-of select="$user/sysuser/password"/>
    </sysuser>
  </xsl:template>

  <!-- - - - - - - - - - - - - - - - - - - - - - -
       Outputs Plesk's limit (used for domains
       and clients accounts)
       Parameters:
       limit - limit name
       value - limit value
       - - - - - - - - - - - - - - - - - - - - - - - -->
  <xsl:template name="limit">
    <xsl:param name="limit"/>
    <xsl:param name="value"/>
    <xsl:element name="limit">
      <xsl:attribute name="name">
        <xsl:value-of select="$limit"/>
      </xsl:attribute>
      <xsl:value-of select="format-number($value, '#')"/>
    </xsl:element>
  </xsl:template>

</xsl:stylesheet>
