#!/usr/bin/perl

# checkXXX function synopsis:
#
# ($agent, $name, $notes) = checkXXX();
#
# Here $agent contains name of agent, ie "PleskX"
# $name contains name of the panel, ie "Plesk 7.1.5"
# $notes contain text which will be shown to the user
#
# If undef is returned from the checker, no platform was detected.
#

use IO::File;

my @checkers;

# Cgp

sub checkCgp {
}
push @checkers, \&checkCgp;

# ConfixxX

# Let woodoo begins...

sub getConfixxConfig {
  my $confName = '/confixx_main.conf';
  if (exists ($ENV{'CONFIXX_ROOT'}) and -d $ENV{'CONFIXX_ROOT'}) {
	if (-T $ENV{'CONFIXX_ROOT'} . $confName) {
	  return $ENV{'CONFIXX_ROOT'} . $confName;
	}
  }

  unless (-T $confFile) {
	foreach my $prefix ('/root/confixx', '/usr/local/confixx', '/confixx') {
	  if (-T $prefix . '/confixx_main.conf') {
		return $prefix . '/confixx_main.conf';
	  }
	}
  }

  unless (-T $confFile) {
	my @out = `grep -hr -e confixx /etc/crontab /etc/cron.d 2>/dev/null`;
	foreach my $line (@out) {
	  chomp $line;
	  my $command = (split(' ', $line))[6];
	  $command =~ s/\/[^\/]*$//; #dirname
	  if (-T $command . '/confixx_main.conf') {
		return $command . '/confixx_main.conf';
	  }
	}
  }

  unless (-T $confFile) {
	my @out  = `crontab -l 2>/dev/null`;
	foreach my $line (@out) {
	  chomp $line;
	  next if ($line =~ /^#/);
	  my $command = (split(' ', $line))[6];
	  $command =~ s/\/[^\/]*$//;
	  if (-T $command . '/confixx_main.conf') {
		return $command . '/confixx_main.conf';
	  }
	}
  }
}

sub checkConfixxX {
  my $conf = getConfixxConfig();

  if ($conf) {
	do $conf;					#loads '$installDir'

	unless (-x "$installDir/confixx_updatescript.pl") {
	  return;
	}

	my $out = `$installDir/confixx_updatescript.pl VERSION 2>/dev/null`;
	chomp $out;
	if ($out =~ /^(?:premium\s+)?(\d+)\.(\d+)\.?(\d+)?/i) {
	  return ["ConfixxX", "Confixx " . $1 . "." . $2 . "." . ($3 || 0)];
	} else {
	  $out = `$installDir/confixx_updatescript.pl --version 2>/dev/null`;
	  chomp $out;
	  if ($out =~ /^(\d+)\.(\d+)\.?(\d+)?/) {
		return ["ConfixxX", "Confixx " . $1 . "." . $2 . "." . ($3 || 0)];
	  }
	}
  }
}
push @checkers, \&checkConfixxX;

# Let woodoo ends...

# cPanel9

sub checkcPanel9 {
  my $confFile = '/var/cpanel/cpanel.config';

  my ($cf, $line) = (new IO::File);
  if ((-T $confFile) && $cf->open("<$confFile")) {
	while ($line = $cf->getline()) {
	  chomp $line;
	  next unless $line;
	  next if ($line =~ /^#/);
	  ($key, $value) = split(/\s*=\s*/, $line, 2);
	  $key =~ s/^\s+//;

	  if ($key =~ /engineroot/) {
		my ($cpanelTool, @res) = ("$value/cpanel");
		if (-x $cpanelTool && (@res = `$cpanelTool 2>/dev/null`)) {
		  foreach my $line (@res) {
			if ($line =~ /cPanel\s+\[(\d+)\.(\d+).\d+/) {
			  if ($1 eq '9' or $1 eq '10') {
				return ["cPanel", "cPanel $1.$2"];
			  }
			}
		  }
		}
	  } elsif (-f "$confFile") {
		return ["cPanel", "cPanel 9"];
	  }
	}
  }
}
push @checkers, \&checkcPanel9;

# EnsimX

sub checkEnsimX {
  my $ensimVersion = `rpm -q webppliance-version 2>/dev/null`;
  if ($ensimVersion =~ /^webppliance-version-([^-]+)-[^-]+$/) {
	return ["EnsimX", "Ensim " . $1];
  }
}
push @checkers, \&checkEnsimX;

# PleskX

sub moreThanVersion {
  my ($pmajor, $pminor, $pminorMinor, $major, $minor, $minorMinor) = @_;

  if ($pmajor > $major) {
	return 1;
  }
  if ($pmajor == $major && $pminor > $minor) {
	return 1;
  }
  if ($pmajor == $major && $pminor == $minor && $pminorminor > $minorMinor) {
	return 1;
  }

  return 0;
}

sub checkPleskX {
  my $confFile = "/etc/psa/psa.conf";

  my ($cf, $productRoot) = (new IO::File, "/usr/local/psa");
  if ((-T $confFile) && $cf->open("<$confFile")) {
	while (my $line = $cf->getline()) {
	  chomp $line;
	  next if ($line =~ /^#/);
	  my ($param, $val) = split /\s+/, $line;
	  if ($param =~ /PRODUCT_ROOT_D/) {
		$productRoot = $val;
		last;
	  }
	}

	if (-T $productRoot . "/version" && $cf->open("<$productRoot/version")) {
	  my ($version, $major, $minor, $minorMinor) = $cf->getline();
	  chomp $version;

	  if ($version =~ /^(\d+)\.(\d+)(.*)$/) {
		$major = $1;
		$minor = $2;

		if ($3 =~ /^\.(\d+)/) {
		  $minorMinor = $1;
		} else {
		  $minorMinor = 0;
		}

		if ($major < 2 || ($major == 2 && $minor < 5)) {
		  return;
		}

		if (moreThanVersion($major, $minor, $minorMinor, 7, 5, 3)) {
		  return;
		}

		return ["PleskX", "Plesk $version"];
	  }
	}
  }
}
push @checkers, \&checkPleskX;

# RaQ2

sub checkRaQ2 {
  my $buildFile = "/etc/build";

  my ($bf, $line) = (new IO::File);
  if ((-T $buildFile) && $bf->open("<$buildFile")) {
	while ($line = $bf->getline()) {
	  return ["RaQ2", "RaQ2"] if $line =~ /2799/ or $line =~ /2800/;
	}
  }
}
push @checkers, \&checkRaQ2;

# RaQ34

sub checkRaQ34 {
  my $buildFile = "/etc/build";

  my ($bf, $line) = (new IO::File);
  if ((-T $buildFile) && $bf->open("<$buildFile")) {
	while ($line = $bf->getline()) {
	  return ["RaQ34", "RaQ3"] if $line =~ /3000/;
	  return ["RaQ34", "RaQ4"] if $line =~ /3\d\d\d/;
	}
  }
}
push @checkers, \&checkRaQ34;

# RaQx

sub checkRaQx {
  my $buildFile = "/etc/build";

  my ($bf, $line) = (new IO::File);
  if ((-T $buildFile) && $bf->open("<$buildFile")) {
	while ($line = $bf->getline()) {
	  return ["RaQx", "RaQ3"] if $line =~ /3000/;
	  return ["RaQx", "RaQ4"] if $line =~ /3\d\d\d/;
	}
  }
}
push @checkers, \&checkRaQx;

# RaQ550

sub checkRaQ550 {
  if (eval 'use lib "/usr/sausalito/perl"; require CCE' and $cce = new CCE) {
	return ["RaQ550", "RaQ550"];
  }
}
push @checkers, \&checkRaQ550;

# Slash

sub checkSlash {
  my $confFile = '/usr/local/slash/admin/conf/admin.conf';

  if (-T $confFile) {
	return ["Slash", "Slash"];
  }
}
push @checkers, \&checkSlash;

# Hsphere

sub checkHsphere {
  my $propFile = '/home/hsphere/local/home/cpanel/shiva/psoft_config/hsphere.properties';

  if (-T $propFile) {
  	return ["HsphereX","HsphereX"];
  }
}
push @checkers, \&checkHsphere;
# ------------------------------------------------------------------------------

# Normalization according to the XML 1.0 (Third Edition),
# clauses 2.3 "Common Syntactic Constructs"
# and 3.3.3 "Attribute-value normalization"

sub normalizeAttr {
  $_ = $_[0];
  s/\r/&#xD;/g;
  s/\n/&#xA;/g;
  s/\t/&#x9;/g;
  s/\</&lt;/g;
  s/\&/&amp;/g;
  s/\"/&quot;/g;
  s/\'/&#39;/g;
  return $_;
}

sub printHostParams {
  my $osname = `uname -s`;
  chomp $osname;

  my $loadavg;
  my $uptime = `uptime`;
  chomp $uptime;
  if ($uptime =~ /(\d+\.\d+),\s+(\d+\.\d+),\s+(\d+\.\d+)/) {
	$loadavg = "$1 $2 $3";
  } else {
	my $rawloadavg = `cat /proc/loadavg`;
	chomp $rawloadavg;
	if ($rawloadavg =~ /(\d+\.\d+)\s+(\d+\.\d+)\s+(\d+\.\d+)/) {
	  $loadavg = "$1 $2 $3";
	}
  }

  print '<os-detection os="' . normalizeAttr($osname) .
	'" loadavg="' . normalizeAttr($loadavg) . '"/>';
}

sub printDetectedPanel {
  print '<panel-detection>';

  my $checker;
  foreach $checker (@checkers) {
	my $res = &{$checker}();
	if ($res) {
	  print '<panel agent="' . normalizeAttr($res->[0]) . '" name="' .
		normalizeAttr($res->[1]) . '" notes="' . normalizeAttr($res->[2]) . '"/>';
	}
  }

  print '</panel-detection>';
}

sub printDetectedFs {
  my %pseudofs = (
				  'autofs' => 1,
				  'binfmt_misc' => 1,
				  'cd9660' => 1,
				  'devfs' => 1,
				  'devpts' => 1,
				  'fdescfs' => 1,
				  'iso9660' => 1,
				  'linprocfs' => 1,
				  'proc' => 1,
				  'procfs' => 1,
				  'romfs' => 1,
				  'sysfs' => 1,
				  'tmpfs' => 1,
				  'usbdevfs' => 1,
				  'usbfs' => 1,
				 );

  my %partitions;

  my $osname = `uname -s`;
  chomp $osname;

  if ($osname eq 'FreeBSD') {
	foreach my $mountinfo (`mount -p`) {
	  chomp $mountinfo;
	  my ($device, $mountpoint, $type, $options) = split /\s+/, $mountinfo;
	  my $mode = 'rw';
	  $mode = 'ro' if ($options =~ /(^|,)ro(,|$)/);

	  unless (defined $pseudofs{$type}) {
		$partitions{$mountpoint} = ();
		$partitions{$mountpoint}->{'device'} = $device;
		$partitions{$mountpoint}->{'mode'} = $mode;
		$partitions{$mountpoint}->{'type'} = $type;
	  }
	}
  } elsif ($osname eq 'Linux') {
	foreach my $mountinfo (`mount`) {
	  chomp $mountinfo;
	  my ($device, undef, $mountpoint, undef, $type, $options) = split /\s+/, $mountinfo;
	  my $mode = 'rw';
	  $mode = 'ro' if ($options =~ /[(,]ro[,)]/);

	  unless (defined $pseudofs{$type}) {
		$partitions{$mountpoint} = ();
		$partitions{$mountpoint}->{'device'} = $device;
		$partitions{$mountpoint}->{'mode'} = $mode;
		$partitions{$mountpoint}->{'type'} = $type;
	  }
	}
  } else {
	die "Unknown OS type";
  }

  my $header = `POSIXLY_CORRECT= df -P | head -n 1`;
  unless ($header =~ /(\d+)-blocks/) {
	die "Unable to determine block size in df output. First line looks like '$header'";
  }
  my $blocksize = $1;

  foreach my $dfinfo (`POSIXLY_CORRECT= df -P | tail -n +2`) {
	chomp $dfinfo;
	my (undef, $size, undef, $free, undef, $mountpoint) = split /\s+/, $dfinfo;
	if (exists $partitions{$mountpoint}) {

	  # for brain-dead NFS shares:
	  $free = $size if $free > $size;

	  $partitions{$mountpoint}->{'size'} = $size * $blocksize;
	  $partitions{$mountpoint}->{'free'} = $free * $blocksize;
	}
  }

  print "<fs-detection>";

  foreach my $partition (keys %partitions) {
	print '<fs mountpoint="' . normalizeAttr($partition) . '" '.
	  'device="' . normalizeAttr($partitions{$partition}->{'device'}) . '" ' .
	  'mode="' . normalizeAttr($partitions{$partition}->{'mode'}) . '" ' .
	  'size="' . normalizeAttr($partitions{$partition}->{'size'}) . '" ' .
	  'free="' . normalizeAttr($partitions{$partition}->{'free'}) . '" ' .
	  'type="' . normalizeAttr($partitions{$partition}->{'type'}) . '"/>';
  }

  print "</fs-detection>";
}

if (scalar(@ARGV) > 1 || (scalar(@ARGV) == 1 && $ARGV[0] ne '--fs-only')) {
  die "Usage: scout.pl [--fs-only]";
}

print '<?xml version="1.0"?>';

if (scalar(@ARGV) == 1) {
  printDetectedFs();
} else {
  print '<scout-result>';
  printHostParams();
  printDetectedPanel();
  printDetectedFs();
  print "</scout-result>";
}

